@@ -13,7 +13,7 @@ use strum::{Display, EnumMessage, IntoEnumIterator};
 use strum_macros::EnumIter;
 
 use crate::input::{InputType, Int};
-use crate::tools::{extract_i64, py_err, py_error_type};
+use crate::tools::{py_err, py_error_type};
 
 use super::PydanticCustomError;
 
@@ -42,21 +42,23 @@ pub fn list_all_errors(py: Python<'_>) -> PyResult<Bound<'_, PyList>> {
     PyList::new(py, errors)
 }
 
-fn field_from_context<'py, T: FromPyObject<'py>>(
+fn field_from_context<'py, T: FromPyObjectOwned<'py>>(
     context: Option<&Bound<'py, PyDict>>,
     field_name: &str,
     enum_name: &str,
     type_name_fn: fn() -> &'static str,
 ) -> PyResult<T> {
     context
-        .ok_or_else(|| py_error_type!(PyTypeError; "{}: '{}' required in context", enum_name, field_name))?
+        .ok_or_else(|| py_error_type!(PyTypeError; "{enum_name}: '{field_name}' required in context"))?
         .get_item(field_name)?
-        .ok_or_else(|| py_error_type!(PyTypeError; "{}: '{}' required in context", enum_name, field_name))?
+        .ok_or_else(|| py_error_type!(PyTypeError; "{enum_name}: '{field_name}' required in context"))?
         .extract::<T>()
-        .map_err(|_| py_error_type!(PyTypeError; "{}: '{}' context value must be a {}", enum_name, field_name, type_name_fn()))
+        .map_err(
+            |_| py_error_type!(PyTypeError; "{enum_name}: '{field_name}' context value must be a {}", type_name_fn()),
+        )
 }
 
-fn cow_field_from_context<'py, T: FromPyObject<'py>, B: ToOwned<Owned = T> + ?Sized + 'static>(
+fn cow_field_from_context<'py, T: FromPyObjectOwned<'py>, B: ToOwned<Owned = T> + ?Sized + 'static>(
     context: Option<&Bound<'py, PyDict>>,
     field_name: &str,
     enum_name: &str,
@@ -102,7 +104,7 @@ macro_rules! error_types {
                 let lookup = ERROR_TYPE_LOOKUP.get_or_init(py, Self::build_lookup);
                 let error_type = match lookup.get(value) {
                     Some(error_type) => error_type.clone(),
-                    None => return py_err!(PyKeyError; "Invalid error type: '{}'", value),
+                    None => return py_err!(PyKeyError; "Invalid error type: '{value}'"),
                 };
                 match error_type {
                     $(
@@ -127,7 +129,7 @@ macro_rules! error_types {
                                 dict.set_item(stringify!($key), $key)?;
                             )*
                             if let Some(ctx) = context {
-                                dict.update(ctx.bind(py).downcast::<PyMapping>()?)?;
+                                dict.update(ctx.bind(py).cast::<PyMapping>()?)?;
                                 Ok(true)
                             } else {
                                 Ok(false)
@@ -461,11 +463,7 @@ macro_rules! to_string_render {
 }
 
 fn plural_s<T: From<u8> + PartialEq>(value: T) -> &'static str {
-    if value == 1.into() {
-        ""
-    } else {
-        "s"
-    }
+    if value == 1.into() { "" } else { "s" }
 }
 
 static ERROR_TYPE_LOOKUP: PyOnceLock<AHashMap<String, ErrorType>> = PyOnceLock::new();
@@ -480,111 +478,136 @@ impl ErrorType {
     }
 
     pub fn message_template_python(&self) -> &'static str {
+        #[allow(clippy::match_same_arms)] // much nicer to have the messages explicitly listed
         match self {
-            Self::NoSuchAttribute {..} => "Object has no attribute '{attribute}'",
-            Self::JsonInvalid {..} => "Invalid JSON: {error}",
-            Self::JsonType {..} => "JSON input should be string, bytes or bytearray",
-            Self::NeedsPythonObject {..} => "Cannot check `{method_name}` when validating from json, use a JsonOrPython validator instead",
-            Self::RecursionLoop {..} => "Recursion error - cyclic reference detected",
-            Self::Missing {..} => "Field required",
-            Self::FrozenField {..} => "Field is frozen",
-            Self::FrozenInstance {..} => "Instance is frozen",
-            Self::ExtraForbidden {..} => "Extra inputs are not permitted",
-            Self::InvalidKey {..} => "Keys should be strings",
-            Self::GetAttributeError {..} => "Error extracting attribute: {error}",
-            Self::ModelType {..} => "Input should be a valid dictionary or instance of {class_name}",
-            Self::ModelAttributesType {..} => "Input should be a valid dictionary or object to extract fields from",
-            Self::DataclassType {..} => "Input should be a dictionary or an instance of {class_name}",
-            Self::DataclassExactType {..} => "Input should be an instance of {class_name}",
-            Self::DefaultFactoryNotCalled {..} => "The default factory uses validated data, but at least one validation error occurred",
-            Self::NoneRequired {..} => "Input should be None",
-            Self::GreaterThan {..} => "Input should be greater than {gt}",
-            Self::GreaterThanEqual {..} => "Input should be greater than or equal to {ge}",
-            Self::LessThan {..} => "Input should be less than {lt}",
-            Self::LessThanEqual {..} => "Input should be less than or equal to {le}",
-            Self::MultipleOf {..} => "Input should be a multiple of {multiple_of}",
-            Self::FiniteNumber {..} => "Input should be a finite number",
-            Self::TooShort {..} => "{field_type} should have at least {min_length} item{expected_plural} after validation, not {actual_length}",
-            Self::TooLong {..} => "{field_type} should have at most {max_length} item{expected_plural} after validation, not {actual_length}",
-            Self::IterableType {..} => "Input should be iterable",
-            Self::IterationError {..} => "Error iterating over object, error: {error}",
-            Self::StringType {..} => "Input should be a valid string",
-            Self::StringSubType {..} => "Input should be a string, not an instance of a subclass of str",
-            Self::StringUnicode {..} => "Input should be a valid string, unable to parse raw data as a unicode string",
-            Self::StringTooShort {..} => "String should have at least {min_length} character{expected_plural}",
-            Self::StringTooLong {..} => "String should have at most {max_length} character{expected_plural}",
-            Self::StringPatternMismatch {..} => "String should match pattern '{pattern}'",
-            Self::Enum {..} => "Input should be {expected}",
-            Self::DictType {..} => "Input should be a valid dictionary",
-            Self::MappingType {..} => "Input should be a valid mapping, error: {error}",
-            Self::ListType {..} => "Input should be a valid list",
-            Self::TupleType {..} => "Input should be a valid tuple",
-            Self::SetType {..} => "Input should be a valid set",
-            Self::SetItemNotHashable {..} => "Set items should be hashable",
-            Self::BoolType {..} => "Input should be a valid boolean",
-            Self::BoolParsing {..} => "Input should be a valid boolean, unable to interpret input",
-            Self::IntType {..} => "Input should be a valid integer",
-            Self::IntParsing {..} => "Input should be a valid integer, unable to parse string as an integer",
-            Self::IntFromFloat {..} => "Input should be a valid integer, got a number with a fractional part",
-            Self::IntParsingSize {..} => "Unable to parse input string as an integer, exceeded maximum size",
-            Self::FloatType {..} => "Input should be a valid number",
-            Self::FloatParsing {..} => "Input should be a valid number, unable to parse string as a number",
-            Self::BytesType {..} => "Input should be a valid bytes",
-            Self::BytesTooShort {..} => "Data should have at least {min_length} byte{expected_plural}",
-            Self::BytesTooLong {..} => "Data should have at most {max_length} byte{expected_plural}",
+            Self::NoSuchAttribute { .. } => "Object has no attribute '{attribute}'",
+            Self::JsonInvalid { .. } => "Invalid JSON: {error}",
+            Self::JsonType { .. } => "JSON input should be string, bytes or bytearray",
+            Self::NeedsPythonObject { .. } => {
+                "Cannot check `{method_name}` when validating from json, use a JsonOrPython validator instead"
+            }
+            Self::RecursionLoop { .. } => "Recursion error - cyclic reference detected",
+            Self::Missing { .. } => "Field required",
+            Self::FrozenField { .. } => "Field is frozen",
+            Self::FrozenInstance { .. } => "Instance is frozen",
+            Self::ExtraForbidden { .. } => "Extra inputs are not permitted",
+            Self::InvalidKey { .. } => "Keys should be strings",
+            Self::GetAttributeError { .. } => "Error extracting attribute: {error}",
+            Self::ModelType { .. } => "Input should be a valid dictionary or instance of {class_name}",
+            Self::ModelAttributesType { .. } => "Input should be a valid dictionary or object to extract fields from",
+            Self::DataclassType { .. } => "Input should be a dictionary or an instance of {class_name}",
+            Self::DataclassExactType { .. } => "Input should be an instance of {class_name}",
+            Self::DefaultFactoryNotCalled { .. } => {
+                "The default factory uses validated data, but at least one validation error occurred"
+            }
+            Self::NoneRequired { .. } => "Input should be None",
+            Self::GreaterThan { .. } => "Input should be greater than {gt}",
+            Self::GreaterThanEqual { .. } => "Input should be greater than or equal to {ge}",
+            Self::LessThan { .. } => "Input should be less than {lt}",
+            Self::LessThanEqual { .. } => "Input should be less than or equal to {le}",
+            Self::MultipleOf { .. } => "Input should be a multiple of {multiple_of}",
+            Self::FiniteNumber { .. } => "Input should be a finite number",
+            Self::TooShort { .. } => {
+                "{field_type} should have at least {min_length} item{expected_plural} after validation, not {actual_length}"
+            }
+            Self::TooLong { .. } => {
+                "{field_type} should have at most {max_length} item{expected_plural} after validation, not {actual_length}"
+            }
+            Self::IterableType { .. } => "Input should be iterable",
+            Self::IterationError { .. } => "Error iterating over object, error: {error}",
+            Self::StringType { .. } => "Input should be a valid string",
+            Self::StringSubType { .. } => "Input should be a string, not an instance of a subclass of str",
+            Self::StringUnicode { .. } => {
+                "Input should be a valid string, unable to parse raw data as a unicode string"
+            }
+            Self::StringTooShort { .. } => "String should have at least {min_length} character{expected_plural}",
+            Self::StringTooLong { .. } => "String should have at most {max_length} character{expected_plural}",
+            Self::StringPatternMismatch { .. } => "String should match pattern '{pattern}'",
+            Self::Enum { .. } => "Input should be {expected}",
+            Self::DictType { .. } => "Input should be a valid dictionary",
+            Self::MappingType { .. } => "Input should be a valid mapping, error: {error}",
+            Self::ListType { .. } => "Input should be a valid list",
+            Self::TupleType { .. } => "Input should be a valid tuple",
+            Self::SetType { .. } => "Input should be a valid set",
+            Self::SetItemNotHashable { .. } => "Set items should be hashable",
+            Self::BoolType { .. } => "Input should be a valid boolean",
+            Self::BoolParsing { .. } => "Input should be a valid boolean, unable to interpret input",
+            Self::IntType { .. } => "Input should be a valid integer",
+            Self::IntParsing { .. } => "Input should be a valid integer, unable to parse string as an integer",
+            Self::IntFromFloat { .. } => "Input should be a valid integer, got a number with a fractional part",
+            Self::IntParsingSize { .. } => "Unable to parse input string as an integer, exceeded maximum size",
+            Self::FloatType { .. } => "Input should be a valid number",
+            Self::FloatParsing { .. } => "Input should be a valid number, unable to parse string as a number",
+            Self::BytesType { .. } => "Input should be a valid bytes",
+            Self::BytesTooShort { .. } => "Data should have at least {min_length} byte{expected_plural}",
+            Self::BytesTooLong { .. } => "Data should have at most {max_length} byte{expected_plural}",
             Self::BytesInvalidEncoding { .. } => "Data should be valid {encoding}: {encoding_error}",
-            Self::ValueError {..} => "Value error, {error}",
-            Self::AssertionError {..} => "Assertion failed, {error}",
-            Self::CustomError {..} => "",  // custom errors are handled separately
-            Self::LiteralError {..} => "Input should be {expected}",
+            Self::ValueError { .. } => "Value error, {error}",
+            Self::AssertionError { .. } => "Assertion failed, {error}",
+            Self::CustomError { .. } => "", // custom errors are handled separately
+            Self::LiteralError { .. } => "Input should be {expected}",
             Self::MissingSentinelError { .. } => "Input should be the 'MISSING' sentinel",
-            Self::DateType {..} => "Input should be a valid date",
-            Self::DateParsing {..} => "Input should be a valid date in the format YYYY-MM-DD, {error}",
-            Self::DateFromDatetimeParsing {..} => "Input should be a valid date or datetime, {error}",
-            Self::DateFromDatetimeInexact {..} => "Datetimes provided to dates should have zero time - e.g. be exact dates",
-            Self::DatePast {..} => "Date should be in the past",
-            Self::DateFuture {..} => "Date should be in the future",
-            Self::TimeType {..} => "Input should be a valid time",
-            Self::TimeParsing {..} => "Input should be in a valid time format, {error}",
-            Self::DatetimeType {..} => "Input should be a valid datetime",
-            Self::DatetimeParsing {..} => "Input should be a valid datetime, {error}",
-            Self::DatetimeObjectInvalid {..} => "Invalid datetime object, got {error}",
-            Self::DatetimeFromDateParsing {..} => "Input should be a valid datetime or date, {error}",
-            Self::DatetimePast {..} => "Input should be in the past",
-            Self::DatetimeFuture {..} => "Input should be in the future",
-            Self::TimezoneNaive {..} => "Input should not have timezone info",
-            Self::TimezoneAware {..} => "Input should have timezone info",
-            Self::TimezoneOffset {..} => "Timezone offset of {tz_expected} required, got {tz_actual}",
-            Self::TimeDeltaType {..} => "Input should be a valid timedelta",
-            Self::TimeDeltaParsing {..} => "Input should be a valid timedelta, {error}",
-            Self::FrozenSetType {..} => "Input should be a valid frozenset",
-            Self::IsInstanceOf {..} => "Input should be an instance of {class}",
-            Self::IsSubclassOf {..} => "Input should be a subclass of {class}",
-            Self::CallableType {..} => "Input should be callable",
-            Self::UnionTagInvalid {..} => "Input tag '{tag}' found using {discriminator} does not match any of the expected tags: {expected_tags}",
-            Self::UnionTagNotFound {..} => "Unable to extract tag using discriminator {discriminator}",
-            Self::ArgumentsType {..} => "Arguments must be a tuple, list or a dictionary",
-            Self::MissingArgument {..} => "Missing required argument",
-            Self::UnexpectedKeywordArgument {..} => "Unexpected keyword argument",
-            Self::MissingKeywordOnlyArgument {..} => "Missing required keyword only argument",
-            Self::UnexpectedPositionalArgument {..} => "Unexpected positional argument",
-            Self::MissingPositionalOnlyArgument {..} => "Missing required positional only argument",
-            Self::MultipleArgumentValues {..} => "Got multiple values for argument",
-            Self::UrlType {..} => "URL input should be a string or URL",
-            Self::UrlParsing {..} => "Input should be a valid URL, {error}",
-            Self::UrlSyntaxViolation {..} => "Input violated strict URL syntax rules, {error}",
-            Self::UrlTooLong {..} => "URL should have at most {max_length} character{expected_plural}",
-            Self::UrlScheme {..} => "URL scheme should be {expected_schemes}",
-            Self::UuidType {..} => "UUID input should be a string, bytes or UUID object",
-            Self::UuidParsing {..} => "Input should be a valid UUID, {error}",
-            Self::UuidVersion {..} => "UUID version {expected_version} expected",
-            Self::DecimalType {..} => "Decimal input should be an integer, float, string or Decimal object",
-            Self::DecimalParsing {..} => "Input should be a valid decimal",
-            Self::DecimalMaxDigits {..} => "Decimal input should have no more than {max_digits} digit{expected_plural} in total",
-            Self::DecimalMaxPlaces {..} => "Decimal input should have no more than {decimal_places} decimal place{expected_plural}",
-            Self::DecimalWholeDigits {..} => "Decimal input should have no more than {whole_digits} digit{expected_plural} before the decimal point",
-            Self::ComplexType {..} => "Input should be a valid python complex object, a number, or a valid complex string following the rules at https://docs.python.org/3/library/functions.html#complex",
-            Self::ComplexStrParsing {..} => "Input should be a valid complex string following the rules at https://docs.python.org/3/library/functions.html#complex",
+            Self::DateType { .. } => "Input should be a valid date",
+            Self::DateParsing { .. } => "Input should be a valid date in the format YYYY-MM-DD, {error}",
+            Self::DateFromDatetimeParsing { .. } => "Input should be a valid date or datetime, {error}",
+            Self::DateFromDatetimeInexact { .. } => {
+                "Datetimes provided to dates should have zero time - e.g. be exact dates"
+            }
+            Self::DatePast { .. } => "Date should be in the past",
+            Self::DateFuture { .. } => "Date should be in the future",
+            Self::TimeType { .. } => "Input should be a valid time",
+            Self::TimeParsing { .. } => "Input should be in a valid time format, {error}",
+            Self::DatetimeType { .. } => "Input should be a valid datetime",
+            Self::DatetimeParsing { .. } => "Input should be a valid datetime, {error}",
+            Self::DatetimeObjectInvalid { .. } => "Invalid datetime object, got {error}",
+            Self::DatetimeFromDateParsing { .. } => "Input should be a valid datetime or date, {error}",
+            Self::DatetimePast { .. } => "Input should be in the past",
+            Self::DatetimeFuture { .. } => "Input should be in the future",
+            Self::TimezoneNaive { .. } => "Input should not have timezone info",
+            Self::TimezoneAware { .. } => "Input should have timezone info",
+            Self::TimezoneOffset { .. } => "Timezone offset of {tz_expected} required, got {tz_actual}",
+            Self::TimeDeltaType { .. } => "Input should be a valid timedelta",
+            Self::TimeDeltaParsing { .. } => "Input should be a valid timedelta, {error}",
+            Self::FrozenSetType { .. } => "Input should be a valid frozenset",
+            Self::IsInstanceOf { .. } => "Input should be an instance of {class}",
+            Self::IsSubclassOf { .. } => "Input should be a subclass of {class}",
+            Self::CallableType { .. } => "Input should be callable",
+            Self::UnionTagInvalid { .. } => {
+                "Input tag '{tag}' found using {discriminator} does not match any of the expected tags: {expected_tags}"
+            }
+            Self::UnionTagNotFound { .. } => "Unable to extract tag using discriminator {discriminator}",
+            Self::ArgumentsType { .. } => "Arguments must be a tuple, list or a dictionary",
+            Self::MissingArgument { .. } => "Missing required argument",
+            Self::UnexpectedKeywordArgument { .. } => "Unexpected keyword argument",
+            Self::MissingKeywordOnlyArgument { .. } => "Missing required keyword only argument",
+            Self::UnexpectedPositionalArgument { .. } => "Unexpected positional argument",
+            Self::MissingPositionalOnlyArgument { .. } => "Missing required positional only argument",
+            Self::MultipleArgumentValues { .. } => "Got multiple values for argument",
+            Self::UrlType { .. } => "URL input should be a string or URL",
+            Self::UrlParsing { .. } => "Input should be a valid URL, {error}",
+            Self::UrlSyntaxViolation { .. } => "Input violated strict URL syntax rules, {error}",
+            Self::UrlTooLong { .. } => "URL should have at most {max_length} character{expected_plural}",
+            Self::UrlScheme { .. } => "URL scheme should be {expected_schemes}",
+            Self::UuidType { .. } => "UUID input should be a string, bytes or UUID object",
+            Self::UuidParsing { .. } => "Input should be a valid UUID, {error}",
+            Self::UuidVersion { .. } => "UUID version {expected_version} expected",
+            Self::DecimalType { .. } => "Decimal input should be an integer, float, string or Decimal object",
+            Self::DecimalParsing { .. } => "Input should be a valid decimal",
+            Self::DecimalMaxDigits { .. } => {
+                "Decimal input should have no more than {max_digits} digit{expected_plural} in total"
+            }
+            Self::DecimalMaxPlaces { .. } => {
+                "Decimal input should have no more than {decimal_places} decimal place{expected_plural}"
+            }
+            Self::DecimalWholeDigits { .. } => {
+                "Decimal input should have no more than {whole_digits} digit{expected_plural} before the decimal point"
+            }
+            Self::ComplexType { .. } => {
+                "Input should be a valid python complex object, a number, or a valid complex string following the rules at https://docs.python.org/3/library/functions.html#complex"
+            }
+            Self::ComplexStrParsing { .. } => {
+                "Input should be a valid complex string following the rules at https://docs.python.org/3/library/functions.html#complex"
+            }
         }
     }
 
@@ -636,12 +659,24 @@ impl ErrorType {
         };
         match self {
             Self::NoSuchAttribute { attribute, .. } => render!(tmpl, attribute),
-            Self::JsonInvalid { error, .. } => render!(tmpl, error),
+            Self::JsonInvalid { error, .. }
+            | Self::GetAttributeError { error, .. }
+            | Self::IterationError { error, .. }
+            | Self::DatetimeObjectInvalid { error, .. }
+            | Self::UrlParsing { error, .. }
+            | Self::UuidParsing { error, .. } => render!(tmpl, error),
+            Self::MappingType { error, .. }
+            | Self::DateParsing { error, .. }
+            | Self::DateFromDatetimeParsing { error, .. }
+            | Self::TimeParsing { error, .. }
+            | Self::DatetimeParsing { error, .. }
+            | Self::DatetimeFromDateParsing { error, .. }
+            | Self::TimeDeltaParsing { error, .. }
+            | Self::UrlSyntaxViolation { error, .. } => render!(tmpl, error),
             Self::NeedsPythonObject { method_name, .. } => render!(tmpl, method_name),
-            Self::GetAttributeError { error, .. } => render!(tmpl, error),
-            Self::ModelType { class_name, .. } => render!(tmpl, class_name),
-            Self::DataclassType { class_name, .. } => render!(tmpl, class_name),
-            Self::DataclassExactType { class_name, .. } => render!(tmpl, class_name),
+            Self::ModelType { class_name, .. }
+            | Self::DataclassType { class_name, .. }
+            | Self::DataclassExactType { class_name, .. } => render!(tmpl, class_name),
             Self::GreaterThan { gt, .. } => to_string_render!(tmpl, gt),
             Self::GreaterThanEqual { ge, .. } => to_string_render!(tmpl, ge),
             Self::LessThan { lt, .. } => to_string_render!(tmpl, lt),
@@ -666,26 +701,18 @@ impl ErrorType {
                 let actual_length = actual_length.map_or(Cow::Borrowed("more"), |v| Cow::Owned(v.to_string()));
                 to_string_render!(tmpl, field_type, max_length, actual_length, expected_plural,)
             }
-            Self::IterationError { error, .. } => render!(tmpl, error),
-            Self::StringTooShort { min_length, .. } => {
+            Self::StringTooShort { min_length, .. } | Self::BytesTooShort { min_length, .. } => {
                 let expected_plural = plural_s(*min_length);
                 to_string_render!(tmpl, min_length, expected_plural)
             }
-            Self::StringTooLong { max_length, .. } => {
+            Self::StringTooLong { max_length, .. }
+            | Self::BytesTooLong { max_length, .. }
+            | Self::UrlTooLong { max_length, .. } => {
                 let expected_plural = plural_s(*max_length);
                 to_string_render!(tmpl, max_length, expected_plural)
             }
             Self::StringPatternMismatch { pattern, .. } => render!(tmpl, pattern),
             Self::Enum { expected, .. } => to_string_render!(tmpl, expected),
-            Self::MappingType { error, .. } => render!(tmpl, error),
-            Self::BytesTooShort { min_length, .. } => {
-                let expected_plural = plural_s(*min_length);
-                to_string_render!(tmpl, min_length, expected_plural)
-            }
-            Self::BytesTooLong { max_length, .. } => {
-                let expected_plural = plural_s(*max_length);
-                to_string_render!(tmpl, max_length, expected_plural)
-            }
             Self::BytesInvalidEncoding {
                 encoding,
                 encoding_error,
@@ -709,18 +736,10 @@ impl ErrorType {
                 ..
             } => PydanticCustomError::format_message(message_template, context.as_ref().map(|c| c.bind(py))),
             Self::LiteralError { expected, .. } => render!(tmpl, expected),
-            Self::DateParsing { error, .. } => render!(tmpl, error),
-            Self::DateFromDatetimeParsing { error, .. } => render!(tmpl, error),
-            Self::TimeParsing { error, .. } => render!(tmpl, error),
-            Self::DatetimeParsing { error, .. } => render!(tmpl, error),
-            Self::DatetimeFromDateParsing { error, .. } => render!(tmpl, error),
-            Self::DatetimeObjectInvalid { error, .. } => render!(tmpl, error),
             Self::TimezoneOffset {
                 tz_expected, tz_actual, ..
             } => to_string_render!(tmpl, tz_expected, tz_actual),
-            Self::TimeDeltaParsing { error, .. } => render!(tmpl, error),
-            Self::IsInstanceOf { class, .. } => render!(tmpl, class),
-            Self::IsSubclassOf { class, .. } => render!(tmpl, class),
+            Self::IsInstanceOf { class, .. } | Self::IsSubclassOf { class, .. } => render!(tmpl, class),
             Self::UnionTagInvalid {
                 discriminator,
                 tag,
@@ -728,14 +747,7 @@ impl ErrorType {
                 ..
             } => render!(tmpl, discriminator, tag, expected_tags),
             Self::UnionTagNotFound { discriminator, .. } => render!(tmpl, discriminator),
-            Self::UrlParsing { error, .. } => render!(tmpl, error),
-            Self::UrlSyntaxViolation { error, .. } => render!(tmpl, error),
-            Self::UrlTooLong { max_length, .. } => {
-                let expected_plural = plural_s(*max_length);
-                to_string_render!(tmpl, max_length, expected_plural)
-            }
             Self::UrlScheme { expected_schemes, .. } => render!(tmpl, expected_schemes),
-            Self::UuidParsing { error, .. } => render!(tmpl, error),
             Self::UuidVersion { expected_version, .. } => to_string_render!(tmpl, expected_version),
             Self::DecimalMaxDigits { max_digits, .. } => {
                 let expected_plural = plural_s(*max_digits);
@@ -809,13 +821,16 @@ impl From<Int> for Number {
     }
 }
 
-impl FromPyObject<'_> for Number {
-    fn extract_bound(obj: &Bound<'_, PyAny>) -> PyResult<Self> {
-        if let Some(int) = extract_i64(obj) {
+impl FromPyObject<'_, '_> for Number {
+    type Error = PyErr;
+
+    #[allow(clippy::same_functions_in_if_condition)] // https://github.com/rust-lang/rust-clippy/issues/10928
+    fn extract(obj: Borrowed<'_, '_, PyAny>) -> PyResult<Self> {
+        if let Ok(int) = obj.extract() {
             Ok(Number::Int(int))
-        } else if let Ok(float) = obj.extract::<f64>() {
+        } else if let Ok(float) = obj.extract() {
             Ok(Number::Float(float))
-        } else if let Ok(string) = obj.extract::<String>() {
+        } else if let Ok(string) = obj.extract() {
             Ok(Number::String(string))
         } else {
             py_err!(PyTypeError; "Expected int or float or String, got {}", obj.get_type())
