@@ -1,4 +1,5 @@
 use std::borrow::Cow;
+use std::cell::RefCell;
 use std::fmt;
 use std::fmt::{Display, Write};
 use std::str::from_utf8;
@@ -18,8 +19,8 @@ use crate::build_tools::py_schema_error_type;
 use crate::errors::LocItem;
 use crate::get_pydantic_version;
 use crate::input::InputType;
-use crate::serializers::{Extra, SerMode, SerializationState};
-use crate::tools::{safe_repr, write_truncated_to_limited_bytes, SchemaDict};
+use crate::serializers::{Extra, SerMode, SerializationConfig, SerializationState, WarningsMode};
+use crate::tools::{SchemaDict, safe_repr, write_truncated_to_limited_bytes};
 
 use super::line_error::ValLineError;
 use super::location::Location;
@@ -27,7 +28,7 @@ use super::types::ErrorType;
 use super::value_exception::PydanticCustomError;
 use super::{InputValue, ValError};
 
-#[pyclass(extends=PyValueError, module="pydantic_core._pydantic_core", subclass)]
+#[pyclass(extends=PyValueError, module="pydantic_core._pydantic_core", subclass, skip_from_py_object, frozen)]
 #[derive(Clone)]
 #[cfg_attr(debug_assertions, derive(Debug))]
 pub struct ValidationError {
@@ -75,12 +76,13 @@ impl ValidationError {
                 let validation_error = Self::new(line_errors, title, input_type, hide_input);
                 match Bound::new(py, validation_error) {
                     Ok(err) => {
-                        if validation_error_cause {
-                            // Will return an import error if the backport was needed and not installed:
-                            if let Some(cause_problem) = ValidationError::maybe_add_cause(err.borrow(), py) {
-                                return cause_problem;
-                            }
+                        // Will return an import error if the backport was needed and not installed:
+                        if validation_error_cause
+                            && let Some(cause_problem) = ValidationError::maybe_add_cause(err.borrow(), py)
+                        {
+                            return cause_problem;
                         }
+
                         PyErr::from_value(err.into_any())
                     }
                     Err(err) => err,
@@ -110,7 +112,9 @@ impl ValidationError {
     }
 
     pub fn use_default_error() -> PyErr {
-        py_schema_error_type!("Uncaught `PydanticUseDefault` exception: the error was raised in a field validator and no default value is available for that field.")
+        py_schema_error_type!(
+            "Uncaught `PydanticUseDefault` exception: the error was raised in a field validator and no default value is available for that field."
+        )
     }
 
     fn maybe_add_cause(self_: PyRef<'_, Self>, py: Python) -> Option<PyErr> {
@@ -179,7 +183,11 @@ impl ValidationError {
                         Ok(group_cls) => group_cls.call1((title, user_py_errs)).ok(),
                         Err(_) => None,
                     },
-                    Err(_) => return Some(PyImportError::new_err("validation_error_cause flag requires the exceptiongroup module backport to be installed when used on Python <3.11.")),
+                    Err(_) => {
+                        return Some(PyImportError::new_err(
+                            "validation_error_cause flag requires the exceptiongroup module backport to be installed when used on Python <3.11.",
+                        ));
+                    }
                 }
             };
 
@@ -236,11 +244,7 @@ fn get_formated_url(py: Python) -> &'static str {
 }
 
 fn get_url_prefix(py: Python<'_>, include_url: bool) -> Option<&str> {
-    if include_url {
-        Some(get_formated_url(py))
-    } else {
-        None
-    }
+    if include_url { Some(get_formated_url(py)) } else { None }
 }
 
 // used to convert a validation error back to ValError for wrap functions
@@ -337,15 +341,30 @@ impl ValidationError {
         include_context: bool,
         include_input: bool,
     ) -> PyResult<Bound<'py, PyString>> {
-        let state = SerializationState::new("iso8601", "iso8601", "utf8", "constants")?;
-        let extra = state.extra(py, &SerMode::Json, None, false, false, true, None, false, None);
-        let serializer = ValidationErrorSerializer {
+        let config = SerializationConfig::default();
+        let extra = Extra::new(
+            py,
+            SerMode::Json,
+            None,
+            false,
+            false,
+            false,
+            false,
+            false,
+            true,
+            None,
+            false,
+            None,
+            None,
+        );
+        let mut state = SerializationState::new(config, WarningsMode::None, None, None, extra)?;
+        let mut serializer = ValidationErrorSerializer {
             py,
             line_errors: &self.line_errors,
             url_prefix: get_url_prefix(py, include_url),
             include_context,
             include_input,
-            extra: &extra,
+            state: &mut state,
             input_type: &self.input_type,
         };
 
@@ -406,7 +425,7 @@ pub fn pretty_py_line_errors<'a>(
 }
 
 /// `PyLineError` are the public version of `ValLineError`, as help and used in `ValidationError`s
-#[pyclass]
+#[pyclass(from_py_object)]
 #[derive(Clone)]
 #[cfg_attr(debug_assertions, derive(Debug))]
 pub struct PyLineError {
@@ -430,18 +449,18 @@ impl TryFrom<&Bound<'_, PyAny>> for PyLineError {
     type Error = PyErr;
 
     fn try_from(value: &Bound<'_, PyAny>) -> PyResult<Self> {
-        let dict = value.downcast::<PyDict>()?;
+        let dict = value.cast::<PyDict>()?;
         let py = value.py();
 
         let type_raw = dict
             .get_item(intern!(py, "type"))?
             .ok_or_else(|| PyKeyError::new_err("type"))?;
 
-        let error_type = if let Ok(type_str) = type_raw.downcast::<PyString>() {
+        let error_type = if let Ok(type_str) = type_raw.cast::<PyString>() {
             let context: Option<Bound<'_, PyDict>> = dict.get_as(intern!(py, "ctx"))?;
             ErrorType::new(py, type_str.to_str()?, context)?
-        } else if let Ok(custom_error) = type_raw.extract::<PydanticCustomError>() {
-            ErrorType::new_custom_error(py, custom_error)
+        } else if let Ok(custom_error) = type_raw.cast::<PydanticCustomError>() {
+            ErrorType::new_custom_error(py, custom_error.get().clone())
         } else {
             return Err(PyTypeError::new_err(
                 "`type` should be a `str` or `PydanticCustomError`",
@@ -491,20 +510,14 @@ impl PyLineError {
         if include_input {
             dict.set_item("input", &self.input_value)?;
         }
-        if include_context {
-            if let Some(context) = self.error_type.py_dict(py)? {
-                dict.set_item("ctx", context)?;
-            }
+        if include_context && let Some(context) = self.error_type.py_dict(py)? {
+            dict.set_item("ctx", context)?;
         }
-        if let Some(url_prefix) = url_prefix {
-            match self.error_type {
-                ErrorType::CustomError { .. } => {
-                    // Don't add URLs for custom errors
-                }
-                _ => {
-                    dict.set_item("url", self.get_error_url(url_prefix))?;
-                }
-            }
+        if let Some(url_prefix) = url_prefix
+            // Don't add URLs for custom errors
+            && !matches!(self.error_type, ErrorType::CustomError { .. })
+        {
+            dict.set_item("url", self.get_error_url(url_prefix))?;
         }
         Ok(dict)
     }
@@ -536,22 +549,16 @@ impl PyLineError {
                 write!(output, ", input_type={type_}")?;
             }
         }
-        if let Some(url_prefix) = url_prefix {
-            match self.error_type {
-                ErrorType::CustomError { .. } => {
-                    // Don't display URLs for custom errors
-                    output.push(']');
-                }
-                _ => {
-                    write!(
-                        output,
-                        "]\n    For further information visit {}",
-                        self.get_error_url(url_prefix)
-                    )?;
-                }
-            }
-        } else {
-            output.push(']');
+        output.push(']');
+        if let Some(url_prefix) = url_prefix
+            // Don't display URLs for custom errors
+            && !matches!(self.error_type, ErrorType::CustomError { .. })
+        {
+            write!(
+                output,
+                "\n    For further information visit {}",
+                self.get_error_url(url_prefix)
+            )?;
         }
         Ok(output)
     }
@@ -568,18 +575,18 @@ where
     S::Error::custom(error.to_string())
 }
 
-struct ValidationErrorSerializer<'py> {
+struct ValidationErrorSerializer<'slf, 'py> {
     py: Python<'py>,
     line_errors: &'py [PyLineError],
     url_prefix: Option<&'py str>,
     include_context: bool,
     include_input: bool,
-    extra: &'py Extra<'py>,
+    state: &'slf mut SerializationState<'py>,
     input_type: &'py InputType,
 }
 
-impl Serialize for ValidationErrorSerializer<'_> {
-    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
+impl ValidationErrorSerializer<'_, '_> {
+    fn serialize<S>(&mut self, serializer: S) -> Result<S::Ok, S::Error>
     where
         S: Serializer,
     {
@@ -591,7 +598,7 @@ impl Serialize for ValidationErrorSerializer<'_> {
                 url_prefix: self.url_prefix,
                 include_context: self.include_context,
                 include_input: self.include_input,
-                extra: self.extra,
+                state: RefCell::new(self.state),
                 input_type: self.input_type,
             };
             seq.serialize_element(&line_s)?;
@@ -600,17 +607,17 @@ impl Serialize for ValidationErrorSerializer<'_> {
     }
 }
 
-struct PyLineErrorSerializer<'py> {
+struct PyLineErrorSerializer<'slf, 'py> {
     py: Python<'py>,
     line_error: &'py PyLineError,
     url_prefix: Option<&'py str>,
     include_context: bool,
     include_input: bool,
-    extra: &'py Extra<'py>,
+    state: RefCell<&'slf mut SerializationState<'py>>,
     input_type: &'py InputType,
 }
 
-impl Serialize for PyLineErrorSerializer<'_> {
+impl Serialize for PyLineErrorSerializer<'_, '_> {
     fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
     where
         S: Serializer,
@@ -621,6 +628,7 @@ impl Serialize for PyLineErrorSerializer<'_> {
             .filter(|b| *b)
             .count();
         let mut map = serializer.serialize_map(Some(size))?;
+        let mut state = self.state.borrow_mut();
 
         map.serialize_entry("type", &self.line_error.error_type.type_string())?;
 
@@ -634,16 +642,13 @@ impl Serialize for PyLineErrorSerializer<'_> {
         map.serialize_entry("msg", &msg)?;
 
         if self.include_input {
-            map.serialize_entry(
-                "input",
-                &self.extra.serialize_infer(self.line_error.input_value.bind(py)),
-            )?;
+            map.serialize_entry("input", &state.serialize_infer(self.line_error.input_value.bind(py)))?;
         }
 
-        if self.include_context {
-            if let Some(context) = self.line_error.error_type.py_dict(py).map_err(py_err_json::<S>)? {
-                map.serialize_entry("ctx", &self.extra.serialize_infer(context.bind(py)))?;
-            }
+        if self.include_context
+            && let Some(context) = self.line_error.error_type.py_dict(py).map_err(py_err_json::<S>)?
+        {
+            map.serialize_entry("ctx", &state.serialize_infer(context.bind(py)))?;
         }
         if let Some(url_prefix) = self.url_prefix {
             map.serialize_entry("url", &self.line_error.get_error_url(url_prefix))?;
