@@ -3,12 +3,11 @@ use pyo3::prelude::*;
 use pyo3::types::{PyDict, PyString};
 
 use crate::input::InputType;
-use crate::tools::extract_i64;
 
-use super::line_error::AsErrorValue;
+use super::line_error::ToErrorValue;
 use super::{ErrorType, ValError};
 
-#[pyclass(extends=PyException, module="pydantic_core._pydantic_core")]
+#[pyclass(extends=PyException, module="pydantic_core._pydantic_core", skip_from_py_object)]
 #[derive(Debug, Clone)]
 pub struct PydanticOmit {}
 
@@ -34,7 +33,7 @@ impl PydanticOmit {
     }
 }
 
-#[pyclass(extends=PyException, module="pydantic_core._pydantic_core")]
+#[pyclass(extends=PyException, module="pydantic_core._pydantic_core", skip_from_py_object)]
 #[derive(Debug, Clone)]
 pub struct PydanticUseDefault {}
 
@@ -54,7 +53,7 @@ impl PydanticUseDefault {
     }
 }
 
-#[pyclass(extends=PyValueError, module="pydantic_core._pydantic_core")]
+#[pyclass(extends=PyValueError, module="pydantic_core._pydantic_core", subclass, frozen, skip_from_py_object)]
 #[derive(Debug, Clone, Default)]
 pub struct PydanticCustomError {
     error_type: String,
@@ -65,22 +64,23 @@ pub struct PydanticCustomError {
 #[pymethods]
 impl PydanticCustomError {
     #[new]
-    pub fn py_new(py: Python, error_type: String, message_template: String, context: Option<&PyDict>) -> Self {
+    #[pyo3(signature = (error_type, message_template, context = None, /))]
+    pub fn py_new(error_type: String, message_template: String, context: Option<Bound<'_, PyDict>>) -> Self {
         Self {
             error_type,
             message_template,
-            context: context.map(|c| c.into_py(py)),
+            context: context.map(Bound::unbind),
         }
     }
 
     #[getter(r#type)]
-    pub fn error_type(&self) -> String {
-        self.error_type.clone()
+    pub fn error_type(&self) -> &str {
+        &self.error_type
     }
 
     #[getter]
-    pub fn message_template(&self) -> String {
-        self.message_template.clone()
+    pub fn message_template(&self) -> &str {
+        &self.message_template
     }
 
     #[getter]
@@ -89,7 +89,7 @@ impl PydanticCustomError {
     }
 
     pub fn message(&self, py: Python) -> PyResult<String> {
-        Self::format_message(&self.message_template, self.context.as_ref().map(|c| c.as_ref(py)))
+        Self::format_message(&self.message_template, self.context.as_ref().map(|c| c.bind(py)))
     }
 
     fn __str__(&self, py: Python) -> PyResult<String> {
@@ -98,15 +98,15 @@ impl PydanticCustomError {
 
     fn __repr__(&self, py: Python) -> PyResult<String> {
         let msg = self.message(py)?;
-        match { self.context.as_ref() } {
-            Some(ctx) => Ok(format!("{msg} [type={}, context={}]", self.error_type, ctx.as_ref(py))),
+        match self.context.as_ref() {
+            Some(ctx) => Ok(format!("{msg} [type={}, context={}]", self.error_type, ctx.bind(py))),
             None => Ok(format!("{msg} [type={}, context=None]", self.error_type)),
         }
     }
 }
 
 impl PydanticCustomError {
-    pub fn into_val_error(self, input: &impl AsErrorValue) -> ValError {
+    pub fn into_val_error(self, input: impl ToErrorValue) -> ValError {
         let error_type = ErrorType::CustomError {
             error_type: self.error_type,
             message_template: self.message_template,
@@ -115,14 +115,14 @@ impl PydanticCustomError {
         ValError::new(error_type, input)
     }
 
-    pub fn format_message(message_template: &str, context: Option<&PyDict>) -> PyResult<String> {
+    pub fn format_message(message_template: &str, context: Option<&Bound<'_, PyDict>>) -> PyResult<String> {
         let mut message = message_template.to_string();
         if let Some(ctx) = context {
-            for (key, value) in ctx {
-                let key: &PyString = key.downcast()?;
-                if let Ok(py_str) = value.downcast::<PyString>() {
+            for (key, value) in ctx.iter() {
+                let key = key.cast::<PyString>()?;
+                if let Ok(py_str) = value.cast::<PyString>() {
                     message = message.replace(&format!("{{{}}}", key.to_str()?), py_str.to_str()?);
-                } else if let Some(value_int) = extract_i64(value) {
+                } else if let Ok(value_int) = value.extract::<i64>() {
                     message = message.replace(&format!("{{{}}}", key.to_str()?), &value_int.to_string());
                 } else {
                     // fallback for anything else just in case
@@ -134,7 +134,7 @@ impl PydanticCustomError {
     }
 }
 
-#[pyclass(extends=PyValueError, module="pydantic_core._pydantic_core")]
+#[pyclass(extends=PyValueError, module="pydantic_core._pydantic_core", skip_from_py_object, frozen)]
 #[derive(Debug, Clone)]
 pub struct PydanticKnownError {
     error_type: ErrorType,
@@ -143,7 +143,8 @@ pub struct PydanticKnownError {
 #[pymethods]
 impl PydanticKnownError {
     #[new]
-    pub fn py_new(py: Python, error_type: &str, context: Option<&PyDict>) -> PyResult<Self> {
+    #[pyo3(signature = (error_type, context=None, /))]
+    pub fn py_new(py: Python, error_type: &str, context: Option<Bound<'_, PyDict>>) -> PyResult<Self> {
         let error_type = ErrorType::new(py, error_type, context)?;
         Ok(Self { error_type })
     }
@@ -173,19 +174,15 @@ impl PydanticKnownError {
 
     fn __repr__(&self, py: Python) -> PyResult<String> {
         let msg = self.message(py)?;
-        match { self.context(py)?.as_ref() } {
-            Some(ctx) => Ok(format!(
-                "{msg} [type={}, context={}]",
-                self.error_type(),
-                ctx.as_ref(py)
-            )),
+        match self.context(py)?.as_ref() {
+            Some(ctx) => Ok(format!("{msg} [type={}, context={}]", self.error_type(), ctx.bind(py))),
             None => Ok(format!("{msg} [type={}, context=None]", self.error_type())),
         }
     }
 }
 
 impl PydanticKnownError {
-    pub fn into_val_error(self, input: &impl AsErrorValue) -> ValError {
+    pub fn into_val_error(self, input: impl ToErrorValue) -> ValError {
         ValError::new(self.error_type, input)
     }
 }
