@@ -1,12 +1,13 @@
 use std::convert::Infallible;
 use std::fmt;
 
+use jiter::JsonValue;
 use pyo3::exceptions::PyValueError;
 use pyo3::types::{PyDict, PyList, PyString};
-use pyo3::{intern, prelude::*, IntoPyObjectExt};
+use pyo3::{IntoPyObjectExt, intern, prelude::*};
 
 use crate::errors::{ErrorTypeDefaults, InputValue, LocItem, ValError, ValResult};
-use crate::lookup_key::{LookupKey, LookupPath};
+use crate::lookup_key::LookupPath;
 use crate::tools::py_err;
 use crate::validators::{TemporalUnitMode, ValBytesMode};
 
@@ -44,7 +45,7 @@ impl TryFrom<&str> for InputType {
             "python" => Ok(Self::Python),
             "json" => Ok(Self::Json),
             "string" => Ok(Self::String),
-            s => py_err!(PyValueError; "Invalid error mode: {}", s),
+            s => py_err!(PyValueError; "Invalid error mode: {s}"),
         }
     }
 }
@@ -73,6 +74,10 @@ pub trait Input<'py>: fmt::Debug {
         None
     }
 
+    fn as_json(&self) -> Option<&JsonValue<'_>> {
+        None
+    }
+
     fn as_kwargs(&self, py: Python<'py>) -> Option<Bound<'py, PyDict>>;
 
     type Arguments<'a>: Arguments<'py>
@@ -85,7 +90,7 @@ pub trait Input<'py>: fmt::Debug {
 
     fn validate_dataclass_args<'a>(&'a self, dataclass_name: &str) -> ValResult<Self::Arguments<'a>>;
 
-    fn validate_str(&self, strict: bool, coerce_numbers_to_str: bool) -> ValMatch<EitherString<'_>>;
+    fn validate_str(&self, strict: bool, coerce_numbers_to_str: bool) -> ValMatch<EitherString<'_, 'py>>;
 
     fn validate_bytes<'a>(&'a self, strict: bool, mode: ValBytesMode) -> ValMatch<EitherBytes<'a, 'py>>;
 
@@ -103,7 +108,7 @@ pub trait Input<'py>: fmt::Debug {
 
     /// Extract a String from the input, only allowing exact
     /// matches for a String (no subclasses)
-    fn exact_str(&self) -> ValResult<EitherString<'_>> {
+    fn exact_str(&self) -> ValResult<EitherString<'_, 'py>> {
         self.validate_str(true, false).and_then(|val_match| {
             val_match
                 .require_exact()
@@ -120,11 +125,7 @@ pub trait Input<'py>: fmt::Debug {
         Self: 'a;
 
     fn validate_dict(&self, strict: bool) -> ValResult<Self::Dict<'_>> {
-        if strict {
-            self.strict_dict()
-        } else {
-            self.lax_dict()
-        }
+        if strict { self.strict_dict() } else { self.lax_dict() }
     }
     fn strict_dict(&self) -> ValResult<Self::Dict<'_>>;
     #[cfg_attr(has_coverage_attribute, coverage(off))]
@@ -223,7 +224,7 @@ pub trait KeywordArgs<'py> {
     where
         Self: 'a;
     fn len(&self) -> usize;
-    fn get_item<'k>(&self, key: &'k LookupKey) -> ValResult<Option<(&'k LookupPath, Self::Item<'_>)>>;
+    fn get_item(&self, key: &LookupPath) -> ValResult<Option<Self::Item<'_>>>;
     fn iter(&self) -> impl Iterator<Item = ValResult<(Self::Key<'_>, Self::Item<'_>)>>;
 }
 
@@ -243,7 +244,7 @@ pub trait ValidatedDict<'py> {
     type Item<'a>: BorrowInput<'py>
     where
         Self: 'a;
-    fn get_item<'k>(&self, key: &'k LookupKey) -> ValResult<Option<(&'k LookupPath, Self::Item<'_>)>>;
+    fn get_item(&self, key: &LookupPath) -> ValResult<Option<Self::Item<'_>>>;
     // FIXME this is a bit of a leaky abstraction
     fn is_py_get_attr(&self) -> bool {
         false
@@ -285,7 +286,7 @@ pub enum Never {}
 impl<'py> ValidatedDict<'py> for Never {
     type Key<'a> = Bound<'py, PyAny>;
     type Item<'a> = Bound<'py, PyAny>;
-    fn get_item<'k>(&self, _key: &'k LookupKey) -> ValResult<Option<(&'k LookupPath, Self::Item<'_>)>> {
+    fn get_item<'k>(&self, _key: &LookupPath) -> ValResult<Option<Self::Item<'_>>> {
         unreachable!()
     }
     fn iterate<'a, R>(
@@ -371,7 +372,7 @@ impl<'py> KeywordArgs<'py> for Never {
     fn len(&self) -> usize {
         unreachable!()
     }
-    fn get_item<'k>(&self, _key: &'k LookupKey) -> ValResult<Option<(&'k LookupPath, Self::Item<'_>)>> {
+    fn get_item<'k>(&self, _key: &LookupPath) -> ValResult<Option<Self::Item<'_>>> {
         unreachable!()
     }
     fn iter(&self) -> impl Iterator<Item = ValResult<(Self::Key<'_>, Self::Item<'_>)>> {
