@@ -14,27 +14,16 @@ use pyo3::PyTypeCheck;
 use pyo3::PyTypeInfo;
 use speedate::MicrosecondsPrecisionOverflowBehavior;
 
+use crate::ArgsKwargs;
 use crate::errors::{ErrorType, ErrorTypeDefaults, InputValue, LocItem, ValError, ValResult};
-use crate::tools::{extract_i64, safe_repr};
-use crate::validators::complex::string_to_complex;
-use crate::validators::decimal::{create_decimal, get_decimal_type};
+use crate::lookup_key::LookupPath;
+use crate::tools::safe_repr;
 use crate::validators::Exactness;
 use crate::validators::TemporalUnitMode;
 use crate::validators::ValBytesMode;
-use crate::ArgsKwargs;
+use crate::validators::complex::{get_complex_type, string_to_complex};
+use crate::validators::decimal::{create_decimal, get_decimal_type};
 
-use super::datetime::{
-    bytes_as_date, bytes_as_datetime, bytes_as_time, bytes_as_timedelta, date_as_datetime, float_as_datetime,
-    float_as_duration, float_as_time, int_as_datetime, int_as_duration, int_as_time, EitherDate, EitherDateTime,
-    EitherTime,
-};
-use super::input_abstract::ValMatch;
-use super::return_enums::EitherComplex;
-use super::return_enums::{iterate_attributes, iterate_mapping_items, ValidationMatch};
-use super::shared::{
-    decimal_as_int, float_as_int, fraction_as_int, get_enum_meta_object, int_as_bool, str_as_bool, str_as_float,
-    str_as_int,
-};
 use super::Arguments;
 use super::ConsumeIterator;
 use super::KeywordArgs;
@@ -43,9 +32,21 @@ use super::ValidatedDict;
 use super::ValidatedList;
 use super::ValidatedSet;
 use super::ValidatedTuple;
+use super::datetime::{
+    EitherDate, EitherDateTime, EitherTime, bytes_as_date, bytes_as_datetime, bytes_as_time, bytes_as_timedelta,
+    date_as_datetime, float_as_datetime, float_as_duration, float_as_time, int_as_datetime, int_as_duration,
+    int_as_time,
+};
+use super::input_abstract::ValMatch;
+use super::return_enums::EitherComplex;
+use super::return_enums::{ValidationMatch, iterate_attributes, iterate_mapping_items};
+use super::shared::{
+    decimal_as_int, float_as_int, fraction_as_int, get_enum_meta_object, int_as_bool, str_as_bool, str_as_float,
+    str_as_int,
+};
 use super::{
-    py_string_str, BorrowInput, EitherBytes, EitherFloat, EitherInt, EitherString, EitherTimedelta, GenericIterator,
-    Input,
+    BorrowInput, EitherBytes, EitherFloat, EitherInt, EitherString, EitherTimedelta, GenericIterator, Input,
+    py_string_str,
 };
 
 static FRACTION_TYPE: PyOnceLock<Py<PyType>> = PyOnceLock::new();
@@ -63,7 +64,7 @@ pub fn get_fraction_type(py: Python<'_>) -> &Bound<'_, PyType> {
 }
 
 pub(crate) fn downcast_python_input<'py, T: PyTypeCheck>(input: &(impl Input<'py> + ?Sized)) -> Option<&Bound<'py, T>> {
-    input.as_python().and_then(|any| any.downcast::<T>().ok())
+    input.as_python().and_then(|any| any.cast::<T>().ok())
 }
 
 pub(crate) fn input_as_python_instance<'a, 'py>(
@@ -75,9 +76,9 @@ pub(crate) fn input_as_python_instance<'a, 'py>(
 
 impl From<&Bound<'_, PyAny>> for LocItem {
     fn from(py_any: &Bound<'_, PyAny>) -> Self {
-        if let Ok(py_str) = py_any.downcast::<PyString>() {
+        if let Ok(py_str) = py_any.cast::<PyString>() {
             py_str.to_string_lossy().as_ref().into()
-        } else if let Some(key_int) = extract_i64(py_any) {
+        } else if let Ok(key_int) = py_any.extract::<i64>() {
             key_int.into()
         } else {
             safe_repr(py_any).to_string().into()
@@ -110,7 +111,7 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
     }
 
     fn as_kwargs(&self, _py: Python<'py>) -> Option<Bound<'py, PyDict>> {
-        self.downcast::<PyDict>().ok().map(Bound::to_owned)
+        self.cast::<PyDict>().ok().map(Bound::to_owned)
     }
 
     type Arguments<'a>
@@ -119,15 +120,13 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
         Self: 'a;
 
     fn validate_args(&self) -> ValResult<PyArgs<'py>> {
-        if let Ok(dict) = self.downcast::<PyDict>() {
+        if let Ok(dict) = self.cast::<PyDict>() {
             Ok(PyArgs::new(None, Some(dict.clone())))
-        } else if let Ok(args_kwargs) = self.extract::<ArgsKwargs>() {
-            let args = args_kwargs.args.into_bound(self.py());
-            let kwargs = args_kwargs.kwargs.map(|d| d.into_bound(self.py()));
-            Ok(PyArgs::new(Some(args), kwargs))
-        } else if let Ok(tuple) = self.downcast::<PyTuple>() {
+        } else if let Ok(args_kwargs) = self.cast::<ArgsKwargs>() {
+            Ok(PyArgs::from_bound_args(args_kwargs))
+        } else if let Ok(tuple) = self.cast::<PyTuple>() {
             Ok(PyArgs::new(Some(tuple.clone()), None))
-        } else if let Ok(list) = self.downcast::<PyList>() {
+        } else if let Ok(list) = self.cast::<PyList>() {
             Ok(PyArgs::new(Some(list.to_tuple()), None))
         } else {
             Err(ValError::new(ErrorTypeDefaults::ArgumentsType, self))
@@ -135,22 +134,18 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
     }
 
     fn validate_args_v3(&self) -> ValResult<PyArgs<'py>> {
-        if let Ok(args_kwargs) = self.extract::<ArgsKwargs>() {
-            let args = args_kwargs.args.into_bound(self.py());
-            let kwargs = args_kwargs.kwargs.map(|d| d.into_bound(self.py()));
-            Ok(PyArgs::new(Some(args), kwargs))
+        if let Ok(args_kwargs) = self.cast::<ArgsKwargs>() {
+            Ok(PyArgs::from_bound_args(args_kwargs))
         } else {
             Err(ValError::new(ErrorTypeDefaults::ArgumentsType, self))
         }
     }
 
     fn validate_dataclass_args<'a>(&'a self, class_name: &str) -> ValResult<PyArgs<'py>> {
-        if let Ok(dict) = self.downcast::<PyDict>() {
+        if let Ok(dict) = self.cast::<PyDict>() {
             Ok(PyArgs::new(None, Some(dict.clone())))
-        } else if let Ok(args_kwargs) = self.extract::<ArgsKwargs>() {
-            let args = args_kwargs.args.into_bound(self.py());
-            let kwargs = args_kwargs.kwargs.map(|d| d.into_bound(self.py()));
-            Ok(PyArgs::new(Some(args), kwargs))
+        } else if let Ok(args_kwargs) = self.cast::<ArgsKwargs>() {
+            Ok(PyArgs::from_bound_args(args_kwargs))
         } else {
             let class_name = class_name.to_string();
             Err(ValError::new(
@@ -168,9 +163,9 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
         strict: bool,
         coerce_numbers_to_str: bool,
     ) -> ValResult<ValidationMatch<EitherString<'_, 'py>>> {
-        if let Ok(py_str) = self.downcast_exact::<PyString>() {
+        if let Ok(py_str) = self.cast_exact::<PyString>() {
             return Ok(ValidationMatch::exact(py_str.clone().into()));
-        } else if let Ok(py_str) = self.downcast::<PyString>() {
+        } else if let Ok(py_str) = self.cast::<PyString>() {
             // force to a rust string to make sure behavior is consistent whether or not we go via a
             // rust string in StrConstrainedValidator - e.g. to_lower
             return Ok(ValidationMatch::strict(py_string_str(py_str)?.into()));
@@ -178,12 +173,12 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
 
         'lax: {
             if !strict {
-                return if let Ok(bytes) = self.downcast::<PyBytes>() {
+                return if let Ok(bytes) = self.cast::<PyBytes>() {
                     match from_utf8(bytes.as_bytes()) {
                         Ok(str) => Ok(str.into()),
                         Err(_) => Err(ValError::new(ErrorTypeDefaults::StringUnicode, self)),
                     }
-                } else if let Ok(py_byte_array) = self.downcast::<PyByteArray>() {
+                } else if let Ok(py_byte_array) = self.cast::<PyByteArray>() {
                     match bytearray_to_str(py_byte_array) {
                         Ok(py_str) => Ok(py_str.into()),
                         Err(_) => Err(ValError::new(ErrorTypeDefaults::StringUnicode, self)),
@@ -215,21 +210,21 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
         strict: bool,
         mode: ValBytesMode,
     ) -> ValResult<ValidationMatch<EitherBytes<'a, 'py>>> {
-        if let Ok(py_bytes) = self.downcast_exact::<PyBytes>() {
+        if let Ok(py_bytes) = self.cast_exact::<PyBytes>() {
             return Ok(ValidationMatch::exact(py_bytes.into()));
-        } else if let Ok(py_bytes) = self.downcast::<PyBytes>() {
+        } else if let Ok(py_bytes) = self.cast::<PyBytes>() {
             return Ok(ValidationMatch::strict(py_bytes.into()));
         }
 
         'lax: {
             if !strict {
-                return if let Ok(py_str) = self.downcast::<PyString>() {
+                return if let Ok(py_str) = self.cast::<PyString>() {
                     let str = py_string_str(py_str)?;
                     match mode.deserialize_string(str) {
                         Ok(b) => Ok(b),
                         Err(e) => Err(ValError::new(e, self)),
                     }
-                } else if let Ok(py_byte_array) = self.downcast::<PyByteArray>() {
+                } else if let Ok(py_byte_array) = self.cast::<PyByteArray>() {
                     Ok(py_byte_array.to_vec().into())
                 } else {
                     break 'lax;
@@ -242,22 +237,22 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
     }
 
     fn validate_bool(&self, strict: bool) -> ValResult<ValidationMatch<bool>> {
-        if let Ok(bool) = self.downcast::<PyBool>() {
+        if let Ok(bool) = self.cast::<PyBool>() {
             return Ok(ValidationMatch::exact(bool.is_true()));
         }
 
         if !strict {
             if let Some(s) = maybe_as_string(self, ErrorTypeDefaults::BoolParsing)? {
                 return str_as_bool(self, s).map(ValidationMatch::lax);
-            } else if let Some(int) = extract_i64(self) {
+            } else if let Ok(int) = self.extract() {
                 return int_as_bool(self, int).map(ValidationMatch::lax);
-            } else if let Ok(float) = self.extract::<f64>() {
-                if let Ok(int) = float_as_int(self, float) {
-                    return int
-                        .as_bool()
-                        .ok_or_else(|| ValError::new(ErrorTypeDefaults::BoolParsing, self))
-                        .map(ValidationMatch::lax);
-                };
+            } else if let Ok(float) = self.extract::<f64>()
+                && let Ok(int) = float_as_int(self, float)
+            {
+                return int
+                    .as_bool()
+                    .ok_or_else(|| ValError::new(ErrorTypeDefaults::BoolParsing, self))
+                    .map(ValidationMatch::lax);
             }
         }
 
@@ -315,7 +310,7 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
     }
 
     fn exact_str(&self) -> ValResult<EitherString<'_, 'py>> {
-        if let Ok(py_str) = self.downcast_exact() {
+        if let Ok(py_str) = self.cast_exact() {
             Ok(EitherString::Py(py_str.clone()))
         } else {
             Err(ValError::new(ErrorTypeDefaults::IntType, self))
@@ -323,15 +318,13 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
     }
 
     fn validate_float(&self, strict: bool) -> ValResult<ValidationMatch<EitherFloat<'_>>> {
-        if let Ok(float) = self.downcast_exact::<PyFloat>() {
+        if let Ok(float) = self.cast_exact::<PyFloat>() {
             return Ok(ValidationMatch::exact(EitherFloat::Py(float.clone())));
         }
 
-        if !strict {
-            if let Some(s) = maybe_as_string(self, ErrorTypeDefaults::FloatParsing)? {
-                // checking for bytes and string is fast, so do this before isinstance(float)
-                return str_as_float(self, s).map(ValidationMatch::lax);
-            }
+        if !strict && let Some(s) = maybe_as_string(self, ErrorTypeDefaults::FloatParsing)? {
+            // checking for bytes and string is fast, so do this before isinstance(float)
+            return str_as_float(self, s).map(ValidationMatch::lax);
         }
 
         if let Ok(float) = self.extract::<f64>() {
@@ -367,6 +360,14 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
             if self.is_instance_of::<PyFloat>() {
                 return create_decimal(self.str()?.as_any(), self).map(ValidationMatch::lax);
             }
+
+            // Handle three-tuple constructor: (sign, digits_tuple, exponent)
+            if let Ok(tuple) = self.cast_exact::<PyTuple>()
+                && tuple.len() == 3
+                && let Ok(decimal) = create_decimal(self, self)
+            {
+                return Ok(ValidationMatch::lax(decimal));
+            }
         }
 
         if self.is_instance(decimal_type)? {
@@ -395,19 +396,19 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
         Self: 'a;
 
     fn strict_dict<'a>(&'a self) -> ValResult<GenericPyMapping<'a, 'py>> {
-        if let Ok(dict) = self.downcast_exact::<PyDict>() {
+        if let Ok(dict) = self.cast_exact::<PyDict>() {
             Ok(GenericPyMapping::Dict(dict))
         } else if self.is_instance_of::<PyDict>() {
-            Ok(GenericPyMapping::Mapping(self.downcast::<PyMapping>()?))
+            Ok(GenericPyMapping::Mapping(self.cast::<PyMapping>()?))
         } else {
             Err(ValError::new(ErrorTypeDefaults::DictType, self))
         }
     }
 
     fn lax_dict<'a>(&'a self) -> ValResult<GenericPyMapping<'a, 'py>> {
-        if let Ok(dict) = self.downcast_exact::<PyDict>() {
+        if let Ok(dict) = self.cast_exact::<PyDict>() {
             Ok(GenericPyMapping::Dict(dict))
-        } else if let Ok(mapping) = self.downcast::<PyMapping>() {
+        } else if let Ok(mapping) = self.cast::<PyMapping>() {
             Ok(GenericPyMapping::Mapping(mapping))
         } else {
             Err(ValError::new(ErrorTypeDefaults::DictType, self))
@@ -421,12 +422,10 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
     ) -> ValResult<GenericPyMapping<'a, 'py>> {
         if from_attributes {
             // if from_attributes, first try a dict, then mapping then from_attributes
-            if let Ok(dict) = self.downcast::<PyDict>() {
+            if let Ok(dict) = self.cast::<PyDict>() {
                 return Ok(GenericPyMapping::Dict(dict));
-            } else if !strict {
-                if let Ok(mapping) = self.downcast::<PyMapping>() {
-                    return Ok(GenericPyMapping::Mapping(mapping));
-                }
+            } else if !strict && let Ok(mapping) = self.cast::<PyMapping>() {
+                return Ok(GenericPyMapping::Mapping(mapping));
             }
 
             if from_attributes_applicable(self) {
@@ -454,12 +453,10 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
         Self: 'a;
 
     fn validate_list<'a>(&'a self, strict: bool) -> ValMatch<PySequenceIterable<'a, 'py>> {
-        if let Ok(list) = self.downcast::<PyList>() {
+        if let Ok(list) = self.cast::<PyList>() {
             return Ok(ValidationMatch::exact(PySequenceIterable::List(list)));
-        } else if !strict {
-            if let Ok(other) = extract_sequence_iterable(self) {
-                return Ok(ValidationMatch::lax(other));
-            }
+        } else if !strict && let Ok(other) = extract_sequence_iterable(self) {
+            return Ok(ValidationMatch::lax(other));
         }
 
         Err(ValError::new(ErrorTypeDefaults::ListType, self))
@@ -471,12 +468,10 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
         Self: 'a;
 
     fn validate_tuple<'a>(&'a self, strict: bool) -> ValMatch<PySequenceIterable<'a, 'py>> {
-        if let Ok(tup) = self.downcast::<PyTuple>() {
+        if let Ok(tup) = self.cast::<PyTuple>() {
             return Ok(ValidationMatch::exact(PySequenceIterable::Tuple(tup)));
-        } else if !strict {
-            if let Ok(other) = extract_sequence_iterable(self) {
-                return Ok(ValidationMatch::lax(other));
-            }
+        } else if !strict && let Ok(other) = extract_sequence_iterable(self) {
+            return Ok(ValidationMatch::lax(other));
         }
 
         Err(ValError::new(ErrorTypeDefaults::TupleType, self))
@@ -488,24 +483,20 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
         Self: 'a;
 
     fn validate_set<'a>(&'a self, strict: bool) -> ValMatch<PySequenceIterable<'a, 'py>> {
-        if let Ok(set) = self.downcast::<PySet>() {
+        if let Ok(set) = self.cast::<PySet>() {
             return Ok(ValidationMatch::exact(PySequenceIterable::Set(set)));
-        } else if !strict {
-            if let Ok(other) = extract_sequence_iterable(self) {
-                return Ok(ValidationMatch::lax(other));
-            }
+        } else if !strict && let Ok(other) = extract_sequence_iterable(self) {
+            return Ok(ValidationMatch::lax(other));
         }
 
         Err(ValError::new(ErrorTypeDefaults::SetType, self))
     }
 
     fn validate_frozenset<'a>(&'a self, strict: bool) -> ValMatch<PySequenceIterable<'a, 'py>> {
-        if let Ok(frozenset) = self.downcast::<PyFrozenSet>() {
+        if let Ok(frozenset) = self.cast::<PyFrozenSet>() {
             return Ok(ValidationMatch::exact(PySequenceIterable::FrozenSet(frozenset)));
-        } else if !strict {
-            if let Ok(other) = extract_sequence_iterable(self) {
-                return Ok(ValidationMatch::lax(other));
-            }
+        } else if !strict && let Ok(other) = extract_sequence_iterable(self) {
+            return Ok(ValidationMatch::lax(other));
         }
 
         Err(ValError::new(ErrorTypeDefaults::FrozenSetType, self))
@@ -520,21 +511,21 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
     }
 
     fn validate_date(&self, strict: bool, mode: TemporalUnitMode) -> ValResult<ValidationMatch<EitherDate<'py>>> {
-        if let Ok(date) = self.downcast_exact::<PyDate>() {
+        if let Ok(date) = self.cast_exact::<PyDate>() {
             Ok(ValidationMatch::exact(date.clone().into()))
         } else if self.is_instance_of::<PyDateTime>() {
             // have to check if it's a datetime first, otherwise the line below converts to a date
             // even if we later try coercion from a datetime, we don't want to return a datetime now
             Err(ValError::new(ErrorTypeDefaults::DateType, self))
-        } else if let Ok(date) = self.downcast::<PyDate>() {
+        } else if let Ok(date) = self.cast::<PyDate>() {
             Ok(ValidationMatch::strict(date.clone().into()))
         } else if let Some(bytes) = {
             if strict {
                 None
-            } else if let Ok(py_str) = self.downcast::<PyString>() {
+            } else if let Ok(py_str) = self.cast::<PyString>() {
                 let str = py_string_str(py_str)?;
                 Some(str.as_bytes())
-            } else if let Ok(py_bytes) = self.downcast::<PyBytes>() {
+            } else if let Ok(py_bytes) = self.cast::<PyBytes>() {
                 Some(py_bytes.as_bytes())
             } else {
                 None
@@ -551,22 +542,22 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
         strict: bool,
         microseconds_overflow_behavior: MicrosecondsPrecisionOverflowBehavior,
     ) -> ValResult<ValidationMatch<EitherTime<'py>>> {
-        if let Ok(time) = self.downcast_exact::<PyTime>() {
+        if let Ok(time) = self.cast_exact::<PyTime>() {
             return Ok(ValidationMatch::exact(time.clone().into()));
-        } else if let Ok(time) = self.downcast::<PyTime>() {
+        } else if let Ok(time) = self.cast::<PyTime>() {
             return Ok(ValidationMatch::strict(time.clone().into()));
         }
 
         'lax: {
             if !strict {
-                return if let Ok(py_str) = self.downcast::<PyString>() {
+                return if let Ok(py_str) = self.cast::<PyString>() {
                     let str = py_string_str(py_str)?;
                     bytes_as_time(self, str.as_bytes(), microseconds_overflow_behavior)
-                } else if let Ok(py_bytes) = self.downcast::<PyBytes>() {
+                } else if let Ok(py_bytes) = self.cast::<PyBytes>() {
                     bytes_as_time(self, py_bytes.as_bytes(), microseconds_overflow_behavior)
                 } else if self.is_exact_instance_of::<PyBool>() {
                     Err(ValError::new(ErrorTypeDefaults::TimeType, self))
-                } else if let Some(int) = extract_i64(self) {
+                } else if let Ok(int) = self.extract() {
                     int_as_time(self, int, 0)
                 } else if let Ok(float) = self.extract::<f64>() {
                     float_as_time(self, float)
@@ -586,26 +577,26 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
         microseconds_overflow_behavior: MicrosecondsPrecisionOverflowBehavior,
         mode: TemporalUnitMode,
     ) -> ValResult<ValidationMatch<EitherDateTime<'py>>> {
-        if let Ok(dt) = self.downcast_exact::<PyDateTime>() {
+        if let Ok(dt) = self.cast_exact::<PyDateTime>() {
             return Ok(ValidationMatch::exact(dt.clone().into()));
-        } else if let Ok(dt) = self.downcast::<PyDateTime>() {
+        } else if let Ok(dt) = self.cast::<PyDateTime>() {
             return Ok(ValidationMatch::strict(dt.clone().into()));
         }
 
         'lax: {
             if !strict {
-                return if let Ok(py_str) = self.downcast::<PyString>() {
+                return if let Ok(py_str) = self.cast::<PyString>() {
                     let str = py_string_str(py_str)?;
                     bytes_as_datetime(self, str.as_bytes(), microseconds_overflow_behavior, mode)
-                } else if let Ok(py_bytes) = self.downcast::<PyBytes>() {
+                } else if let Ok(py_bytes) = self.cast::<PyBytes>() {
                     bytes_as_datetime(self, py_bytes.as_bytes(), microseconds_overflow_behavior, mode)
                 } else if self.is_exact_instance_of::<PyBool>() {
                     Err(ValError::new(ErrorTypeDefaults::DatetimeType, self))
-                } else if let Some(int) = extract_i64(self) {
+                } else if let Ok(int) = self.extract() {
                     int_as_datetime(self, int, 0, mode)
                 } else if let Ok(float) = self.extract::<f64>() {
                     float_as_datetime(self, float, mode)
-                } else if let Ok(date) = self.downcast::<PyDate>() {
+                } else if let Ok(date) = self.cast::<PyDate>() {
                     Ok(date_as_datetime(date)?)
                 } else {
                     break 'lax;
@@ -633,12 +624,12 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
 
         'lax: {
             if !strict {
-                return if let Ok(py_str) = self.downcast::<PyString>() {
+                return if let Ok(py_str) = self.cast::<PyString>() {
                     let str = py_string_str(py_str)?;
                     bytes_as_timedelta(self, str.as_bytes(), microseconds_overflow_behavior)
-                } else if let Ok(py_bytes) = self.downcast::<PyBytes>() {
+                } else if let Ok(py_bytes) = self.cast::<PyBytes>() {
                     bytes_as_timedelta(self, py_bytes.as_bytes(), microseconds_overflow_behavior)
-                } else if let Some(int) = extract_i64(self) {
+                } else if let Ok(int) = self.extract() {
                     Ok(int_as_duration(self, int)?.into())
                 } else if let Ok(float) = self.extract::<f64>() {
                     Ok(float_as_duration(self, float)?.into())
@@ -653,7 +644,7 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
     }
 
     fn validate_complex<'a>(&'a self, strict: bool, py: Python<'py>) -> ValResult<ValidationMatch<EitherComplex<'py>>> {
-        if let Ok(complex) = self.downcast::<PyComplex>() {
+        if let Ok(complex) = self.cast::<PyComplex>() {
             return Ok(ValidationMatch::strict(EitherComplex::Py(complex.to_owned())));
         }
         if strict {
@@ -669,25 +660,24 @@ impl<'py> Input<'py> for Bound<'py, PyAny> {
             ));
         }
 
-        if let Ok(s) = self.downcast::<PyString>() {
+        if let Ok(s) = self.cast::<PyString>()
+            && let Ok(c) = string_to_complex(s, self)
+        {
             // If input is not a valid complex string, instead of telling users to correct
             // the string, it makes more sense to tell them to provide any acceptable value
             // since they might have just given values of some incorrect types instead
             // of actually trying some complex strings.
-            if let Ok(c) = string_to_complex(s, self) {
-                return Ok(ValidationMatch::lax(EitherComplex::Py(c)));
-            }
-        } else if self.is_exact_instance_of::<PyFloat>() {
-            return Ok(ValidationMatch::lax(EitherComplex::Complex([
-                self.extract::<f64>().unwrap(),
-                0.0,
-            ])));
-        } else if self.is_exact_instance_of::<PyInt>() {
-            return Ok(ValidationMatch::lax(EitherComplex::Complex([
-                self.extract::<i64>().unwrap() as f64,
-                0.0,
-            ])));
+            return Ok(ValidationMatch::lax(EitherComplex::Py(c)));
         }
+
+        // Delegate to the constructor directly
+        // (see https://docs.python.org/3/library/functions.html#complex):
+        if let Ok(complex_obj) = get_complex_type(py).call1((self,))
+            && let Ok(complex) = complex_obj.cast_into()
+        {
+            return Ok(ValidationMatch::lax(EitherComplex::Py(complex)));
+        }
+
         Err(ValError::new(ErrorTypeDefaults::ComplexType, self))
     }
 }
@@ -713,7 +703,7 @@ fn from_attributes_applicable(obj: &Bound<'_, PyAny>) -> bool {
         .get_type()
         .getattr(intern!(obj.py(), "__module__"))
         .ok()
-        .and_then(|module_name| module_name.downcast_into::<PyString>().ok())
+        .and_then(|module_name| module_name.cast_into::<PyString>().ok())
     else {
         return false;
     };
@@ -726,9 +716,9 @@ fn from_attributes_applicable(obj: &Bound<'_, PyAny>) -> bool {
 
 /// Utility for extracting a string from a PyAny, if possible.
 fn maybe_as_string<'a>(v: &'a Bound<'_, PyAny>, unicode_error: ErrorType) -> ValResult<Option<&'a str>> {
-    if let Ok(py_string) = v.downcast::<PyString>() {
+    if let Ok(py_string) = v.cast::<PyString>() {
         py_string_str(py_string).map(Some)
-    } else if let Ok(bytes) = v.downcast::<PyBytes>() {
+    } else if let Ok(bytes) = v.cast::<PyBytes>() {
         match from_utf8(bytes.as_bytes()) {
             Ok(s) => Ok(Some(s)),
             Err(_) => Err(ValError::new(unicode_error, v)),
@@ -746,7 +736,7 @@ fn bytearray_to_str<'py>(bytearray: &Bound<'py, PyByteArray>) -> PyResult<Bound<
     let py = bytearray.py();
     let py_string = bytearray
         .call_method1(intern!(py, "decode"), (intern!(py, "utf-8"),))?
-        .downcast_into()?;
+        .cast_into()?;
     Ok(py_string)
 }
 
@@ -780,6 +770,15 @@ impl<'py> PyArgs<'py> {
             kwargs: kwargs.map(PyKwargs),
         }
     }
+
+    fn from_bound_args(py_args: &Bound<'py, ArgsKwargs>) -> Self {
+        let py = py_args.py();
+        let args_kwargs = py_args.get();
+        Self::new(
+            Some(args_kwargs.args.bind(py).clone()),
+            args_kwargs.kwargs.as_ref().map(|d| d.bind(py).clone()),
+        )
+    }
 }
 
 impl<'py> Arguments<'py> for PyArgs<'py> {
@@ -829,10 +828,7 @@ impl<'py> KeywordArgs<'py> for PyKwargs<'py> {
         self.0.len()
     }
 
-    fn get_item<'k>(
-        &self,
-        key: &'k crate::lookup_key::LookupKey,
-    ) -> ValResult<Option<(&'k crate::lookup_key::LookupPath, Self::Item<'_>)>> {
+    fn get_item<'k>(&self, key: &LookupPath) -> ValResult<Option<Self::Item<'_>>> {
         key.py_get_dict_item(&self.0).map_err(Into::into)
     }
 
@@ -859,10 +855,7 @@ impl<'py> ValidatedDict<'py> for GenericPyMapping<'_, 'py> {
     where
         Self: 'a;
 
-    fn get_item<'k>(
-        &self,
-        key: &'k crate::lookup_key::LookupKey,
-    ) -> ValResult<Option<(&'k crate::lookup_key::LookupPath, Self::Item<'_>)>> {
+    fn get_item(&self, key: &LookupPath) -> ValResult<Option<Self::Item<'_>>> {
         match self {
             Self::Dict(dict) => key.py_get_dict_item(dict).map_err(Into::into),
             Self::Mapping(mapping) => key.py_get_mapping_item(mapping).map_err(Into::into),
@@ -916,13 +909,13 @@ pub enum PySequenceIterable<'a, 'py> {
 /// or frozenset
 fn extract_sequence_iterable<'a, 'py>(obj: &'a Bound<'py, PyAny>) -> ValResult<PySequenceIterable<'a, 'py>> {
     // Handle concrete non-overlapping types first, then abstract types
-    if let Ok(iterable) = obj.downcast::<PyList>() {
+    if let Ok(iterable) = obj.cast::<PyList>() {
         Ok(PySequenceIterable::List(iterable))
-    } else if let Ok(iterable) = obj.downcast::<PyTuple>() {
+    } else if let Ok(iterable) = obj.cast::<PyTuple>() {
         Ok(PySequenceIterable::Tuple(iterable))
-    } else if let Ok(iterable) = obj.downcast::<PySet>() {
+    } else if let Ok(iterable) = obj.cast::<PySet>() {
         Ok(PySequenceIterable::Set(iterable))
-    } else if let Ok(iterable) = obj.downcast::<PyFrozenSet>() {
+    } else if let Ok(iterable) = obj.cast::<PyFrozenSet>() {
         Ok(PySequenceIterable::FrozenSet(iterable))
     } else {
         // Try to get this as a generable iterable thing, but exclude string and mapping types
@@ -930,11 +923,10 @@ fn extract_sequence_iterable<'a, 'py>(obj: &'a Bound<'py, PyAny>) -> ValResult<P
             || obj.is_instance_of::<PyBytes>()
             || obj.is_instance_of::<PyByteArray>()
             || obj.is_instance_of::<PyDict>()
-            || obj.downcast::<PyMapping>().is_ok())
+            || obj.cast::<PyMapping>().is_ok())
+            && let Ok(iter) = obj.try_iter()
         {
-            if let Ok(iter) = obj.try_iter() {
-                return Ok(PySequenceIterable::Iterator(iter));
-            }
+            return Ok(PySequenceIterable::Iterator(iter));
         }
 
         Err(ValError::new(ErrorTypeDefaults::IterableType, obj))
