@@ -12,8 +12,8 @@ mod shared;
 
 pub use datetime::TzInfo;
 pub(crate) use datetime::{
-    duration_as_pytimedelta, pydate_as_date, pydatetime_as_datetime, pytime_as_time, EitherDate, EitherDateTime,
-    EitherTimedelta,
+    EitherDate, EitherDateTime, EitherTimedelta, duration_as_pytimedelta, pydate_as_date, pydatetime_as_datetime,
+    pytime_as_time,
 };
 pub(crate) use input_abstract::{
     Arguments, BorrowInput, ConsumeIterator, Input, InputType, KeywordArgs, PositionalArgs, ValidatedDict,
@@ -22,15 +22,11 @@ pub(crate) use input_abstract::{
 pub(crate) use input_python::{downcast_python_input, input_as_python_instance};
 pub(crate) use input_string::StringMapping;
 pub(crate) use return_enums::{
-    no_validator_iter_to_vec, py_string_str, validate_iter_to_set, validate_iter_to_vec, EitherBytes, EitherFloat,
-    EitherInt, EitherString, GenericIterator, Int, MaxLengthCheck, ValidationMatch,
+    EitherBytes, EitherFloat, EitherInt, EitherString, GenericIterator, Int, MaxLengthCheck, ValidationMatch,
+    no_validator_iter_to_vec, py_string_str, validate_iter_to_set, validate_iter_to_vec,
 };
 
 // Defined here as it's not exported by pyo3
 pub fn py_error_on_minusone(py: Python<'_>, result: c_int) -> PyResult<()> {
-    if result != -1 {
-        Ok(())
-    } else {
-        Err(PyErr::fetch(py))
-    }
+    if result != -1 { Ok(()) } else { Err(PyErr::fetch(py)) }
 }
