@@ -15,16 +15,16 @@ use pyo3::types::PyFunction;
 use pyo3::types::{PyBytes, PyComplex, PyFloat, PyFrozenSet, PyIterator, PyMapping, PySet, PyString};
 
 use pyo3::IntoPyObjectExt;
-use serde::{ser::Error, Serialize, Serializer};
+use serde::{Serialize, Serializer, ser::Error};
 
 use crate::errors::{
-    py_err_string, ErrorType, ErrorTypeDefaults, InputValue, ToErrorValue, ValError, ValLineError, ValResult,
+    ErrorType, ErrorTypeDefaults, InputValue, ToErrorValue, ValError, ValLineError, ValResult, py_err_string,
 };
 use crate::py_gc::PyGcTraverse;
-use crate::tools::{extract_i64, extract_int, new_py_string, py_err};
+use crate::tools::new_py_string;
 use crate::validators::{CombinedValidator, Exactness, ValidationState, Validator};
 
-use super::{py_error_on_minusone, BorrowInput, Input};
+use super::{BorrowInput, Input, py_error_on_minusone};
 
 pub struct ValidationMatch<T>(T, Exactness);
 
@@ -126,8 +126,8 @@ pub(crate) fn validate_iter_to_vec<'py>(
     validator: &CombinedValidator,
     state: &mut ValidationState<'_, 'py>,
     fail_fast: bool,
-) -> ValResult<Vec<PyObject>> {
-    let mut output: Vec<PyObject> = Vec::with_capacity(capacity);
+) -> ValResult<Vec<Py<PyAny>>> {
+    let mut output: Vec<Py<PyAny>> = Vec::with_capacity(capacity);
     let mut errors: Vec<ValLineError> = Vec::new();
     let allow_partial = state.allow_partial;
 
@@ -164,13 +164,13 @@ pub(crate) fn validate_iter_to_vec<'py>(
 }
 
 pub trait BuildSet {
-    fn build_add(&self, item: PyObject) -> PyResult<()>;
+    fn build_add(&self, item: Py<PyAny>) -> PyResult<()>;
 
     fn build_len(&self) -> usize;
 }
 
 impl BuildSet for Bound<'_, PySet> {
-    fn build_add(&self, item: PyObject) -> PyResult<()> {
+    fn build_add(&self, item: Py<PyAny>) -> PyResult<()> {
         self.add(item)
     }
 
@@ -180,7 +180,7 @@ impl BuildSet for Bound<'_, PySet> {
 }
 
 impl BuildSet for Bound<'_, PyFrozenSet> {
-    fn build_add(&self, item: PyObject) -> PyResult<()> {
+    fn build_add(&self, item: Py<PyAny>) -> PyResult<()> {
         py_error_on_minusone(self.py(), unsafe {
             // Safety: self.as_ptr() the _only_ pointer to the `frozenset`, and it's allowed
             // to mutate this via the C API when nothing else can refer to it.
@@ -236,21 +236,21 @@ pub(crate) fn validate_iter_to_set<'py>(
         let item = item_result.map_err(|e| any_next_error!(py, e, input, index))?;
         match validate_add(py, set, item, state, validator) {
             Ok(()) => {
-                if let Some(max_length) = max_length {
-                    if set.build_len() > max_length {
-                        return Err(ValError::new(
-                            ErrorType::TooLong {
-                                field_type: field_type.to_string(),
-                                max_length,
-                                // The logic here is that it doesn't matter how many elements the
-                                // input actually had; all we know is it had more than the allowed
-                                // number of deduplicated elements.
-                                actual_length: None,
-                                context: None,
-                            },
-                            input,
-                        ));
-                    }
+                if let Some(max_length) = max_length
+                    && set.build_len() > max_length
+                {
+                    return Err(ValError::new(
+                        ErrorType::TooLong {
+                            field_type: field_type.to_string(),
+                            max_length,
+                            // The logic here is that it doesn't matter how many elements the
+                            // input actually had; all we know is it had more than the allowed
+                            // number of deduplicated elements.
+                            actual_length: None,
+                            context: None,
+                        },
+                        input,
+                    ));
                 }
             }
             Err(ValError::LineErrors(line_errors)) => {
@@ -278,7 +278,7 @@ pub(crate) fn no_validator_iter_to_vec<'py>(
     input: &(impl Input<'py> + ?Sized),
     iter: impl Iterator<Item = PyResult<impl BorrowInput<'py>>>,
     mut max_length_check: MaxLengthCheck<'_, impl Input<'py> + ?Sized>,
-) -> ValResult<Vec<PyObject>> {
+) -> ValResult<Vec<Py<PyAny>>> {
     iter.enumerate()
         .map(|(index, result)| {
             let v = result.map_err(|e| any_next_error!(py, e, input, index))?;
@@ -327,31 +327,28 @@ const MAPPING_TUPLE_ERROR: &str = "Mapping items must be tuples of (key, value)
 pub(crate) fn iterate_attributes<'a, 'py>(
     object: &'a Bound<'py, PyAny>,
 ) -> PyResult<impl Iterator<Item = ValResult<(Bound<'py, PyAny>, Bound<'py, PyAny>)>> + 'a> {
-    let mut attributes_iterator = object.dir()?.into_iter();
-
-    Ok(std::iter::from_fn(move || {
-        // loop until we find an attribute who's name does not start with underscore,
-        // or we get to the end of the list of attributes
-        let name = attributes_iterator.next()?;
-        // from benchmarks this is 14x faster than using the python `startswith` method
-        let name_cow = match name.downcast::<PyString>() {
-            Ok(name) => name.to_string_lossy(),
+    Ok(object.dir()?.into_iter().filter_map(|attr| {
+        let name = match attr.cast_into::<PyString>() {
+            Ok(name) => name,
             Err(e) => return Some(Err(e.into())),
         };
-        if !name_cow.as_ref().starts_with('_') {
-            // getattr is most likely to fail due to an exception in a @property, skip
-            if let Ok(attr) = object.getattr(name_cow.as_ref()) {
+
+        // skip private attributes
+        // from benchmarks this is 14x faster than using the python `startswith` method
+        if !name.to_string_lossy().starts_with('_')
+                // getattr is most likely to fail due to an exception in a @property, skip
+                && let Ok(attr) = object.getattr(&name)
                 // we don't want bound methods to be included, is there a better way to check?
                 // ref https://stackoverflow.com/a/18955425/949890
-                let is_bound = matches!(attr.hasattr(intern!(attr.py(), "__self__")), Ok(true));
+                && matches!(attr.hasattr(intern!(attr.py(), "__self__")), Ok(true))
                 // the PyFunction::is_type_of(attr) catches `staticmethod`, but also any other function,
                 // I think that's better than including static methods in the yielded attributes,
                 // if someone really wants fields, they can use an explicit field, or a function to modify input
-                if !is_bound && !attr.is_instance_of::<PyFunction>() {
-                    return Some(Ok((name, attr)));
-                }
-            }
+                && !attr.is_instance_of::<PyFunction>()
+        {
+            return Some(Ok((name.into_any(), attr)));
         }
+
         None
     }))
 }
@@ -400,7 +397,7 @@ impl From<&Bound<'_, PyAny>> for GenericIterator<'_> {
 
 #[derive(Debug, Clone)]
 pub struct GenericPyIterator {
-    obj: PyObject,
+    obj: Py<PyAny>,
     iter: Py<PyIterator>,
     index: usize,
 }
@@ -468,20 +465,20 @@ impl<'data> GenericJsonIterator<'data> {
 }
 
 #[cfg_attr(debug_assertions, derive(Debug))]
-pub enum EitherString<'a> {
+pub enum EitherString<'a, 'py> {
     Cow(Cow<'a, str>),
-    Py(Bound<'a, PyString>),
+    Py(Bound<'py, PyString>),
 }
 
-impl<'a> EitherString<'a> {
+impl<'py> EitherString<'_, 'py> {
     pub fn as_cow(&self) -> ValResult<Cow<'_, str>> {
         match self {
-            Self::Cow(data) => Ok(data.clone()),
+            Self::Cow(data) => Ok(Cow::Borrowed(data)),
             Self::Py(py_str) => Ok(Cow::Borrowed(py_string_str(py_str)?)),
         }
     }
 
-    pub fn as_py_string(&'a self, py: Python<'a>, cache_str: StringCacheMode) -> Bound<'a, PyString> {
+    pub fn as_py_string(&self, py: Python<'py>, cache_str: StringCacheMode) -> Bound<'py, PyString> {
         match self {
             Self::Cow(cow) => new_py_string(py, cow.as_ref(), cache_str),
             Self::Py(py_string) => py_string.clone(),
@@ -489,20 +486,20 @@ impl<'a> EitherString<'a> {
     }
 }
 
-impl<'a> From<&'a str> for EitherString<'a> {
+impl<'a> From<&'a str> for EitherString<'a, '_> {
     fn from(data: &'a str) -> Self {
         Self::Cow(Cow::Borrowed(data))
     }
 }
 
-impl From<String> for EitherString<'_> {
+impl From<String> for EitherString<'_, '_> {
     fn from(data: String) -> Self {
         Self::Cow(Cow::Owned(data))
     }
 }
 
-impl<'a> From<Bound<'a, PyString>> for EitherString<'a> {
-    fn from(date: Bound<'a, PyString>) -> Self {
+impl<'py> From<Bound<'py, PyString>> for EitherString<'_, 'py> {
+    fn from(date: Bound<'py, PyString>) -> Self {
         Self::Py(date)
     }
 }
@@ -568,8 +565,7 @@ pub enum EitherInt<'py> {
 
 impl<'py> EitherInt<'py> {
     pub fn upcast(py_any: &Bound<'py, PyAny>) -> ValResult<Self> {
-        // Safety: we know that py_any is a python int
-        if let Some(int_64) = extract_i64(py_any) {
+        if let Ok(int_64) = py_any.extract() {
             Ok(Self::I64(int_64))
         } else {
             let big_int: BigInt = py_any.extract()?;
@@ -706,11 +702,17 @@ impl Rem for &Int {
     }
 }
 
-impl FromPyObject<'_> for Int {
-    fn extract_bound(obj: &Bound<'_, PyAny>) -> PyResult<Self> {
-        match extract_int(obj) {
-            Some(i) => Ok(i),
-            None => py_err!(PyTypeError; "Expected int, got {}", obj.get_type()),
+impl FromPyObject<'_, '_> for Int {
+    type Error = PyErr;
+
+    #[allow(clippy::same_functions_in_if_condition)] // https://github.com/rust-lang/rust-clippy/issues/10928
+    fn extract(obj: Borrowed<'_, '_, PyAny>) -> PyResult<Self> {
+        if let Ok(i) = obj.extract() {
+            Ok(Int::I64(i))
+        } else if let Ok(b) = obj.extract() {
+            Ok(Int::Big(b))
+        } else {
+            Err(PyTypeError::new_err(format!("Expected int, got {}", obj.get_type())))
         }
     }
 }
