@@ -1,28 +1,33 @@
 use std::borrow::Cow;
+use std::cell::RefCell;
 use std::fmt::Debug;
 use std::io::{self, Write};
 use std::sync::Arc;
 
 use pyo3::exceptions::PyTypeError;
-use pyo3::prelude::*;
-use pyo3::sync::PyOnceLock;
+use pyo3::pybacked::PyBackedStr;
 use pyo3::types::{PyDict, PyString};
-use pyo3::{intern, PyTraverseError, PyVisit};
+use pyo3::{IntoPyObjectExt, prelude::*};
+use pyo3::{PyTraverseError, PyVisit, intern};
 
 use enum_dispatch::enum_dispatch;
-use serde::Serialize;
+use serde::ser::SerializeMap as _;
+use serde::{Serialize, Serializer};
 use serde_json::ser::{Formatter, PrettyFormatter};
 
 use crate::build_tools::py_schema_err;
 use crate::build_tools::py_schema_error_type;
 use crate::definitions::DefinitionsBuilder;
 use crate::py_gc::PyGcTraverse;
+use crate::serializers::SerMode;
+use crate::serializers::errors::WrappedSerError;
+use crate::serializers::polymorphism_trampoline::PolymorphismTrampoline;
 use crate::serializers::ser::PythonSerializer;
 use crate::serializers::type_serializers::any::AnySerializer;
-use crate::tools::{py_err, SchemaDict};
+use crate::tools::{SchemaDict, py_err};
 
 use super::errors::se_err_py_err;
-use super::extra::Extra;
+use super::extra::SerializationState;
 use super::infer::{infer_json_key, infer_serialize, infer_to_python};
 use super::ob_type::{IsType, ObType};
 
@@ -61,16 +66,16 @@ macro_rules! combined_serializer {
                     $(
                         <$b_serializer>::EXPECTED_TYPE => match <$b_serializer>::build(schema, config, definitions) {
                             Ok(serializer) => Ok(serializer),
-                            Err(err) => py_schema_err!("Error building `{}` serializer:\n  {}", lookup_type, err),
+                            Err(err) => py_schema_err!("Error building `{lookup_type}` serializer:\n  {err}"),
                         },
                     )*
                     $(
                         <$builder>::EXPECTED_TYPE => match <$builder>::build(schema, config, definitions) {
                             Ok(serializer) => Ok(serializer),
-                            Err(err) => py_schema_err!("Error building `{}` serializer:\n  {}", lookup_type, err),
+                            Err(err) => py_schema_err!("Error building `{lookup_type}` serializer:\n  {err}"),
                         },
                     )*
-                    _ => py_schema_err!("Unknown serialization schema type: `{}`", lookup_type),
+                    _ => py_schema_err!("Unknown serialization schema type: `{lookup_type}`"),
                 }
             }
         }
@@ -89,6 +94,9 @@ combined_serializer! {
         Fields: super::fields::GeneralFieldsSerializer;
         // prebuilt serializers are manually constructed, and thus manually added to the `CombinedSerializer` enum
         Prebuilt: super::prebuilt::PrebuiltSerializer;
+        // polymorphism trampoline is manually constructed to wrap models and dataclasses with
+        // polymorphic serialization
+        PolymorphismTrampoline: super::polymorphism_trampoline::PolymorphismTrampoline;
     }
     // `find_only` is for type_serializers which are built directly via the `type` key and `find_serializer`
     // but aren't actually used for serialization, e.g. their `build` method must return another serializer
@@ -108,7 +116,6 @@ combined_serializer! {
         super::type_serializers::function::FunctionPlainSerializerBuilder;
         super::type_serializers::function::FunctionWrapSerializerBuilder;
         super::type_serializers::model::ModelFieldsBuilder;
-        super::type_serializers::typed_dict::TypedDictBuilder;
     }
     // `both` means the struct is added to both the `CombinedSerializer` enum and the match statement in
     // `find_serializer` so they can be used via a `type` str.
@@ -149,20 +156,11 @@ combined_serializer! {
         Recursive: super::type_serializers::definitions::DefinitionRefSerializer;
         Tuple: super::type_serializers::tuple::TupleSerializer;
         Complex: super::type_serializers::complex::ComplexSerializer;
+        TypedDict: super::type_serializers::typed_dict::TypedDictSerializer;
     }
 }
 
 impl CombinedSerializer {
-    // Used when creating the base serializer instance, to avoid reusing the instance
-    // when unpickling:
-    pub fn build_base(
-        schema: &Bound<'_, PyDict>,
-        config: Option<&Bound<'_, PyDict>>,
-        definitions: &mut DefinitionsBuilder<Arc<CombinedSerializer>>,
-    ) -> PyResult<Arc<CombinedSerializer>> {
-        Self::_build(schema, config, definitions, false)
-    }
-
     fn _build(
         schema: &Bound<'_, PyDict>,
         config: Option<&Bound<'_, PyDict>>,
@@ -184,7 +182,7 @@ impl CombinedSerializer {
                         config,
                         definitions,
                     )
-                    .map_err(|err| py_schema_error_type!("Error building `function-plain` serializer:\n  {}", err));
+                    .map_err(|err| py_schema_error_type!("Error building `function-plain` serializer:\n  {err}"));
                 }
                 Some("function-wrap") => {
                     // `function-wrap` is also a special case, not included in `find_serializer` since it mean
@@ -195,27 +193,29 @@ impl CombinedSerializer {
                         config,
                         definitions,
                     )
-                    .map_err(|err| py_schema_error_type!("Error building `function-wrap` serializer:\n  {}", err));
+                    .map_err(|err| py_schema_error_type!("Error building `function-wrap` serializer:\n  {err}"));
                 }
-                // applies to lists tuples and dicts, does not override the main schema `type`
-                Some("include-exclude-sequence" | "include-exclude-dict") => (),
-                // applies specifically to bytes, does not override the main schema `type`
-                Some("base64") => (),
+                Some(
+                    // applies to lists tuples and dicts, does not override the main schema `type`
+                    "include-exclude-sequence" | "include-exclude-dict"
+                    // applies specifically to bytes, does not override the main schema `type`
+                    | "base64"
+                )
+                // if `schema.serialization.type` is None, fall back to `schema.type`
+                | None => (),
                 Some(ser_type) => {
                     // otherwise if `schema.serialization.type` is defined, use that with `find_serializer`
                     // instead of `schema.type`. In this case it's an error if a serializer isn't found.
                     return Self::find_serializer(ser_type, &ser_schema, config, definitions);
                 }
-                // if `schema.serialization.type` is None, fall back to `schema.type`
-                None => (),
-            };
+            }
         }
 
         let type_: Bound<'_, PyString> = schema.get_as_req(type_key)?;
         let type_ = type_.to_str()?;
 
         if use_prebuilt {
-            // if we have a SchemaValidator on the type already, use it
+            // if we have a SchemaSerializer on the type already, use it
             if let Ok(Some(prebuilt_serializer)) =
                 super::prebuilt::PrebuiltSerializer::try_get_from_schema(type_, schema)
             {
@@ -223,75 +223,108 @@ impl CombinedSerializer {
             }
         }
 
-        Self::find_serializer(type_, schema, config, definitions)
+        let serializer = Self::find_serializer(type_, schema, config, definitions)?;
+        Self::maybe_wrap_in_polymorphism_trampoline(serializer, schema)
+    }
+
+    fn maybe_wrap_in_polymorphism_trampoline(
+        serializer: Arc<CombinedSerializer>,
+        schema: &Bound<'_, PyDict>,
+    ) -> PyResult<Arc<CombinedSerializer>> {
+        let py = schema.py();
+        let type_: Bound<'_, PyString> = schema.get_as_req(intern!(py, "type"))?;
+        let type_ = type_.to_str()?;
+
+        // Note: it could make sense to generalize this behavior for any type that may have subclasses,
+        // but apart from models and dataclasses, that would be for arbitrary types where custom serialization
+        // has to be defined already.
+        if type_ == "model" || type_ == "dataclass" {
+            // Get polymorphic serialization from config
+            let config = schema.get_as::<Bound<'_, PyDict>>(intern!(py, "config"))?;
+            let polymorphic_serialization: bool = config
+                .and_then(|cfg| cfg.get_as(intern!(py, "polymorphic_serialization")).transpose())
+                .unwrap_or(Ok(false))?;
+
+            // Unconditionally wrap in PolymorphismTrampoline, because runtime flag might still enable it
+            Ok(Arc::new(
+                PolymorphismTrampoline::new(
+                    schema.get_as_req(intern!(py, "cls"))?,
+                    serializer,
+                    polymorphic_serialization,
+                )
+                .into(),
+            ))
+        } else {
+            Ok(serializer)
+        }
     }
 
     /// Main recursive way to call serializers, supports possible recursive type inference by
     /// switching to type inference mode eagerly.
-    pub fn to_python(
+    pub fn to_python<'py>(
         &self,
-        value: &Bound<'_, PyAny>,
-        include: Option<&Bound<'_, PyAny>>,
-        exclude: Option<&Bound<'_, PyAny>>,
-        extra: &Extra,
+        value: &Bound<'py, PyAny>,
+        state: &mut SerializationState<'py>,
     ) -> PyResult<Py<PyAny>> {
-        if extra.serialize_as_any {
-            infer_to_python(value, include, exclude, extra)
+        if state.extra.serialize_as_any {
+            infer_to_python(value, state)
         } else {
-            self.to_python_no_infer(value, include, exclude, extra)
+            self.to_python_no_infer(value, state)
         }
     }
 
     /// Variant of the above which does not fall back to inference mode immediately
     #[inline]
-    pub fn to_python_no_infer(
+    pub fn to_python_no_infer<'py>(
         &self,
-        value: &Bound<'_, PyAny>,
-        include: Option<&Bound<'_, PyAny>>,
-        exclude: Option<&Bound<'_, PyAny>>,
-        extra: &Extra,
+        value: &Bound<'py, PyAny>,
+        state: &mut SerializationState<'py>,
     ) -> PyResult<Py<PyAny>> {
-        TypeSerializer::to_python(self, value, include, exclude, extra)
+        TypeSerializer::to_python(self, value, state)
     }
 
-    pub fn json_key<'a>(&self, key: &'a Bound<'_, PyAny>, extra: &Extra) -> PyResult<Cow<'a, str>> {
-        if extra.serialize_as_any {
-            infer_json_key(key, extra)
+    pub fn json_key<'a, 'py>(
+        &self,
+        key: &'a Bound<'py, PyAny>,
+        state: &mut SerializationState<'py>,
+    ) -> PyResult<Cow<'a, str>> {
+        if state.extra.serialize_as_any {
+            infer_json_key(key, state)
         } else {
-            self.json_key_no_infer(key, extra)
+            self.json_key_no_infer(key, state)
         }
     }
 
     #[inline]
-    pub fn json_key_no_infer<'a>(&self, key: &'a Bound<'_, PyAny>, extra: &Extra) -> PyResult<Cow<'a, str>> {
-        TypeSerializer::json_key(self, key, extra)
+    pub fn json_key_no_infer<'a, 'py>(
+        &self,
+        key: &'a Bound<'py, PyAny>,
+        state: &mut SerializationState<'py>,
+    ) -> PyResult<Cow<'a, str>> {
+        TypeSerializer::json_key(self, key, state)
     }
 
-    pub fn serde_serialize<S: serde::ser::Serializer>(
+    pub fn serde_serialize<'py, S: serde::ser::Serializer>(
         &self,
-        value: &Bound<'_, PyAny>,
+        value: &Bound<'py, PyAny>,
         serializer: S,
-        include: Option<&Bound<'_, PyAny>>,
-        exclude: Option<&Bound<'_, PyAny>>,
-        extra: &Extra,
+        state: &mut SerializationState<'py>,
     ) -> Result<S::Ok, S::Error> {
-        if extra.serialize_as_any {
-            infer_serialize(value, serializer, include, exclude, extra)
+        if state.extra.serialize_as_any {
+            infer_serialize(value, serializer, state)
         } else {
-            self.serde_serialize_no_infer(value, serializer, include, exclude, extra)
+            self.serde_serialize_no_infer(value, serializer, state)
         }
     }
 
     #[inline]
-    pub fn serde_serialize_no_infer<S: serde::ser::Serializer>(
+    pub fn serde_serialize_no_infer<'py, S: serde::ser::Serializer>(
         &self,
-        value: &Bound<'_, PyAny>,
+        value: &Bound<'py, PyAny>,
         serializer: S,
-        include: Option<&Bound<'_, PyAny>>,
-        exclude: Option<&Bound<'_, PyAny>>,
-        extra: &Extra,
+        state: &mut SerializationState<'py>,
     ) -> Result<S::Ok, S::Error> {
-        TypeSerializer::serde_serialize(self, value, serializer, include, exclude, extra)
+        TypeSerializer::serde_serialize(self, value, serializer, state)
     }
 }
 
@@ -304,7 +337,11 @@ impl BuildSerializer for CombinedSerializer {
         config: Option<&Bound<'_, PyDict>>,
         definitions: &mut DefinitionsBuilder<Arc<CombinedSerializer>>,
     ) -> PyResult<Arc<CombinedSerializer>> {
-        Self::_build(schema, config, definitions, true)
+        // Read use_prebuilt from the definitions builder - this ensures all nested
+        // serializers respect the same setting as the top-level build
+        let use_prebuilt = definitions.use_prebuilt();
+        let serializer = Self::_build(schema, config, definitions, use_prebuilt)?;
+        Self::maybe_wrap_in_polymorphism_trampoline(serializer, schema)
     }
 }
 
@@ -352,44 +389,42 @@ impl PyGcTraverse for CombinedSerializer {
             CombinedSerializer::Tuple(inner) => inner.py_gc_traverse(visit),
             CombinedSerializer::Uuid(inner) => inner.py_gc_traverse(visit),
             CombinedSerializer::Complex(inner) => inner.py_gc_traverse(visit),
+            CombinedSerializer::TypedDict(inner) => inner.py_gc_traverse(visit),
+            CombinedSerializer::PolymorphismTrampoline(inner) => inner.py_gc_traverse(visit),
         }
     }
 }
 
 #[enum_dispatch(CombinedSerializer)]
 pub(crate) trait TypeSerializer: Send + Sync + Debug {
-    fn to_python(
-        &self,
-        value: &Bound<'_, PyAny>,
-        include: Option<&Bound<'_, PyAny>>,
-        exclude: Option<&Bound<'_, PyAny>>,
-        extra: &Extra,
-    ) -> PyResult<Py<PyAny>>;
+    fn to_python<'py>(&self, value: &Bound<'py, PyAny>, state: &mut SerializationState<'py>) -> PyResult<Py<PyAny>>;
 
-    fn json_key<'a>(&self, key: &'a Bound<'_, PyAny>, extra: &Extra) -> PyResult<Cow<'a, str>>;
-
-    fn invalid_as_json_key<'a>(
+    fn json_key<'a, 'py>(
         &self,
-        key: &'a Bound<'_, PyAny>,
-        extra: &Extra,
+        key: &'a Bound<'py, PyAny>,
+        state: &mut SerializationState<'py>,
+    ) -> PyResult<Cow<'a, str>>;
+
+    fn invalid_as_json_key<'a, 'py>(
+        &self,
+        key: &'a Bound<'py, PyAny>,
+        state: &mut SerializationState<'py>,
         expected_type: &'static str,
     ) -> PyResult<Cow<'a, str>> {
-        match extra.ob_type_lookup.is_type(key, ObType::None) {
-            IsType::Exact | IsType::Subclass => py_err!(PyTypeError; "`{}` not valid as object key", expected_type),
+        match state.extra.ob_type_lookup.is_type(key, ObType::None) {
+            IsType::Exact | IsType::Subclass => py_err!(PyTypeError; "`{expected_type}` not valid as object key"),
             IsType::False => {
-                extra.warnings.on_fallback_py(self.get_name(), key, extra)?;
-                infer_json_key(key, extra)
+                state.warn_fallback_py(self.get_name(), key)?;
+                infer_json_key(key, state)
             }
         }
     }
 
-    fn serde_serialize<S: serde::ser::Serializer>(
+    fn serde_serialize<'py, S: serde::ser::Serializer>(
         &self,
-        value: &Bound<'_, PyAny>,
+        value: &Bound<'py, PyAny>,
         serializer: S,
-        include: Option<&Bound<'_, PyAny>>,
-        exclude: Option<&Bound<'_, PyAny>>,
-        extra: &Extra,
+        state: &mut SerializationState<'py>,
     ) -> Result<S::Ok, S::Error>;
 
     fn get_name(&self) -> &str;
@@ -404,58 +439,50 @@ pub(crate) trait TypeSerializer: Send + Sync + Debug {
     }
 }
 
-pub(crate) struct PydanticSerializer<'py> {
-    value: &'py Bound<'py, PyAny>,
-    serializer: &'py CombinedSerializer,
-    include: Option<&'py Bound<'py, PyAny>>,
-    exclude: Option<&'py Bound<'py, PyAny>>,
-    extra: &'py Extra<'py>,
+pub(crate) struct PydanticSerializer<'slf, 'py> {
+    value: &'slf Bound<'py, PyAny>,
+    serializer: &'slf CombinedSerializer,
+    /// RefCell to allow mutable access to the state during serialization, we expect it
+    /// to only ever be borrowed mutably once at a time.
+    state: RefCell<&'slf mut SerializationState<'py>>,
 }
 
-impl<'py> PydanticSerializer<'py> {
+impl<'slf, 'py> PydanticSerializer<'slf, 'py> {
     pub(crate) fn new(
-        value: &'py Bound<'py, PyAny>,
-        serializer: &'py CombinedSerializer,
-        include: Option<&'py Bound<'py, PyAny>>,
-        exclude: Option<&'py Bound<'py, PyAny>>,
-        extra: &'py Extra<'py>,
+        value: &'slf Bound<'py, PyAny>,
+        serializer: &'slf CombinedSerializer,
+        state: &'slf mut SerializationState<'py>,
     ) -> Self {
         Self {
             value,
-            serializer: if extra.serialize_as_any {
+            serializer: if state.extra.serialize_as_any {
                 AnySerializer::get()
             } else {
                 serializer
             },
-            include,
-            exclude,
-            extra,
+            state: RefCell::new(state),
         }
     }
 
     /// Same as above but will not fall back to type inference when `serialize_as_any` is set
     pub(crate) fn new_no_infer(
-        value: &'py Bound<'py, PyAny>,
-        serializer: &'py CombinedSerializer,
-        include: Option<&'py Bound<'py, PyAny>>,
-        exclude: Option<&'py Bound<'py, PyAny>>,
-        extra: &'py Extra<'py>,
+        value: &'slf Bound<'py, PyAny>,
+        serializer: &'slf CombinedSerializer,
+        state: &'slf mut SerializationState<'py>,
     ) -> Self {
         Self {
             value,
             serializer,
-            include,
-            exclude,
-            extra,
+            state: RefCell::new(state),
         }
     }
 }
 
-impl Serialize for PydanticSerializer<'_> {
+impl Serialize for PydanticSerializer<'_, '_> {
     fn serialize<S: serde::ser::Serializer>(&self, serializer: S) -> Result<S::Ok, S::Error> {
         // inference is handled in the constructor
         self.serializer
-            .serde_serialize_no_infer(self.value, serializer, self.include, self.exclude, self.extra)
+            .serde_serialize_no_infer(self.value, serializer, &mut self.state.borrow_mut())
     }
 }
 
@@ -541,17 +568,15 @@ impl Formatter for EscapeNonAsciiPrettyFormatter<'_> {
 }
 
 #[allow(clippy::too_many_arguments)]
-pub(crate) fn to_json_bytes(
-    value: &Bound<'_, PyAny>,
+pub(crate) fn to_json_bytes<'py>(
+    value: &Bound<'py, PyAny>,
     serializer: &CombinedSerializer,
-    include: Option<&Bound<'_, PyAny>>,
-    exclude: Option<&Bound<'_, PyAny>>,
-    extra: &Extra,
+    state: &mut SerializationState<'py>,
     indent: Option<usize>,
     ensure_ascii: bool,
     expected_json_size: usize,
 ) -> PyResult<Vec<u8>> {
-    let serializer = PydanticSerializer::new(value, serializer, include, exclude, extra);
+    let serializer = PydanticSerializer::new(value, serializer, state);
 
     let writer: Vec<u8> = Vec::with_capacity(expected_json_size);
 
@@ -585,38 +610,222 @@ pub(crate) fn to_json_bytes(
     Ok(bytes)
 }
 
-#[allow(clippy::type_complexity)]
-pub(super) fn any_dataclass_iter<'a, 'py>(
-    dataclass: &'a Bound<'py, PyAny>,
-) -> PyResult<(
-    impl Iterator<Item = PyResult<(Bound<'py, PyAny>, Bound<'py, PyAny>)>> + 'a,
-    Bound<'py, PyDict>,
-)>
-where
-    'py: 'a,
-{
-    let py = dataclass.py();
-    let fields = dataclass
-        .getattr(intern!(py, "__dataclass_fields__"))?
-        .downcast_into::<PyDict>()?;
-    let field_type_marker = get_field_marker(py)?;
+/// Common interface for doing serialization
+pub(crate) trait DoSerialize {
+    type Ok;
+    type Error: From<PyErr>;
 
-    let next = move |(field_name, field): (Bound<'py, PyAny>, Bound<'py, PyAny>)| -> PyResult<Option<(Bound<'py, PyAny>, Bound<'py, PyAny>)>> {
-        let field_type = field.getattr(intern!(py, "_field_type"))?;
-        if field_type.is(field_type_marker) {
-            let value = dataclass.getattr(field_name.downcast::<PyString>()?)?;
-            Ok(Some((field_name, value)))
+    fn serialize_no_infer<'py>(
+        self,
+        serializer: &CombinedSerializer,
+        value: &Bound<'py, PyAny>,
+        state: &mut SerializationState<'py>,
+    ) -> Result<Self::Ok, Self::Error>;
+
+    fn serialize_fallback<'py>(
+        self,
+        name: &str,
+        value: &Bound<'py, PyAny>,
+        state: &mut SerializationState<'py>,
+    ) -> Result<Self::Ok, Self::Error>;
+
+    fn serialize_str(self, value: &Bound<'_, PyString>) -> Result<Self::Ok, Self::Error>;
+
+    type Map: SerializeMap<Ok = Self::Ok, Error = Self::Error>;
+
+    fn serialize_map(self) -> Result<Self::Map, Self::Error>;
+}
+
+pub(crate) trait SerializeMap {
+    type Ok;
+    type Error: From<PyErr>;
+
+    /// Serialize a key-value entry into the map
+    fn serialize_entry<'py>(
+        &mut self,
+        key: &Bound<'py, PyAny>,
+        key_serializer: &CombinedSerializer,
+        value: &Bound<'py, PyAny>,
+        value_serializer: &CombinedSerializer,
+        state: &mut SerializationState<'py>,
+    ) -> Result<(), Self::Error>;
+
+    /// Serialize a string key-value entry into the map (avoids needing to "serialize" the key)
+    fn serialize_entry_string_key<'py>(
+        &mut self,
+        key: &PyBackedStr,
+        value: &Bound<'py, PyAny>,
+        value_serializer: &CombinedSerializer,
+        state: &mut SerializationState<'py>,
+    ) -> Result<(), Self::Error>;
+
+    fn end(self) -> Result<Self::Ok, Self::Error>;
+}
+
+/// Helper to create a `SerializeToPython` instance
+pub fn serialize_to_python(py: Python<'_>) -> SerializeToPython<'_> {
+    SerializeToPython(py)
+}
+
+/// Helper to create a `SerializeToJson` instance
+pub fn serialize_to_json<S>(serializer: S) -> SerializeToJson<S> {
+    SerializeToJson { serializer }
+}
+
+pub struct SerializeToPython<'py>(Python<'py>);
+
+impl<'s> DoSerialize for SerializeToPython<'s> {
+    type Ok = Py<PyAny>;
+    type Error = PyErr;
+
+    fn serialize_no_infer<'py>(
+        self,
+        serializer: &CombinedSerializer,
+        value: &Bound<'py, PyAny>,
+        state: &mut SerializationState<'py>,
+    ) -> PyResult<Py<PyAny>> {
+        serializer.to_python_no_infer(value, state)
+    }
+
+    fn serialize_fallback<'py>(
+        self,
+        name: &str,
+        value: &Bound<'py, PyAny>,
+        state: &mut SerializationState<'py>,
+    ) -> PyResult<Py<PyAny>> {
+        state.warn_fallback_py(name, value)?;
+        infer_to_python(value, state)
+    }
+
+    fn serialize_str(self, value: &Bound<'_, PyString>) -> Result<Py<PyAny>, PyErr> {
+        value.into_py_any(value.py())
+    }
+
+    type Map = Bound<'s, PyDict>;
+
+    fn serialize_map(self) -> Result<Bound<'s, PyDict>, PyErr> {
+        Ok(PyDict::new(self.0))
+    }
+}
+
+impl SerializeMap for Bound<'_, PyDict> {
+    type Ok = Py<PyAny>;
+    type Error = PyErr;
+
+    fn serialize_entry<'py>(
+        &mut self,
+        key: &Bound<'py, PyAny>,
+        key_serializer: &CombinedSerializer,
+        value: &Bound<'py, PyAny>,
+        value_serializer: &CombinedSerializer,
+        state: &mut SerializationState<'py>,
+    ) -> Result<(), Self::Error> {
+        if matches!(state.extra.mode, SerMode::Json) {
+            let key = key_serializer.json_key_no_infer(key, state)?;
+            let py_value = value_serializer.to_python_no_infer(value, state)?;
+            self.set_item(key, py_value)?;
         } else {
-            Ok(None)
+            let py_key = key_serializer.to_python_no_infer(key, state)?;
+            let py_value = value_serializer.to_python_no_infer(value, state)?;
+            self.set_item(py_key, py_value)?;
         }
-    };
+        Ok(())
+    }
 
-    Ok((fields.iter().filter_map(move |field| next(field).transpose()), fields))
+    fn serialize_entry_string_key<'py>(
+        &mut self,
+        key: &PyBackedStr,
+        value: &Bound<'py, PyAny>,
+        value_serializer: &CombinedSerializer,
+        state: &mut SerializationState<'py>,
+    ) -> Result<(), Self::Error> {
+        let py_value = value_serializer.to_python_no_infer(value, state)?;
+        self.set_item(key, py_value)?;
+        Ok(())
+    }
+
+    fn end(self) -> Result<Self::Ok, Self::Error> {
+        Ok(self.into_any().unbind())
+    }
 }
 
-static DC_FIELD_MARKER: PyOnceLock<Py<PyAny>> = PyOnceLock::new();
+pub(crate) struct SerializeToJson<S> {
+    serializer: S,
+}
 
-/// needed to match the logic from dataclasses.fields `tuple(f for f in fields.values() if f._field_type is _FIELD)`
-fn get_field_marker(py: Python<'_>) -> PyResult<&Bound<'_, PyAny>> {
-    DC_FIELD_MARKER.import(py, "dataclasses", "_FIELD")
+impl<S: Serializer> DoSerialize for SerializeToJson<S> {
+    type Ok = S::Ok;
+    type Error = WrappedSerError<S::Error>;
+
+    fn serialize_no_infer<'py>(
+        self,
+        serializer: &CombinedSerializer,
+        value: &Bound<'py, PyAny>,
+        state: &mut SerializationState<'py>,
+    ) -> Result<S::Ok, WrappedSerError<S::Error>> {
+        serializer
+            .serde_serialize_no_infer(value, self.serializer, state)
+            .map_err(WrappedSerError)
+    }
+
+    fn serialize_fallback<'py>(
+        self,
+        name: &str,
+        value: &Bound<'py, PyAny>,
+        state: &mut SerializationState<'py>,
+    ) -> Result<S::Ok, WrappedSerError<S::Error>> {
+        state.warn_fallback_ser::<S>(name, value).map_err(WrappedSerError)?;
+        infer_serialize(value, self.serializer, state).map_err(WrappedSerError)
+    }
+
+    fn serialize_str(self, value: &Bound<'_, PyString>) -> Result<S::Ok, WrappedSerError<S::Error>> {
+        let s = value.to_str()?;
+        self.serializer.serialize_str(s).map_err(WrappedSerError)
+    }
+
+    type Map = SerdeMapSerializer<S>;
+
+    fn serialize_map(self) -> Result<SerdeMapSerializer<S>, Self::Error> {
+        Ok(SerdeMapSerializer::<S> {
+            // choice not to pass size_hint here is because JSON maps don't need a size hint
+            map: self.serializer.serialize_map(None).map_err(WrappedSerError)?,
+        })
+    }
+}
+
+pub(crate) struct SerdeMapSerializer<S: Serializer> {
+    map: S::SerializeMap,
+}
+
+impl<S: Serializer> SerializeMap for SerdeMapSerializer<S> {
+    type Ok = S::Ok;
+    type Error = WrappedSerError<S::Error>;
+
+    fn serialize_entry<'py>(
+        &mut self,
+        key: &Bound<'py, PyAny>,
+        key_serializer: &CombinedSerializer,
+        value: &Bound<'py, PyAny>,
+        value_serializer: &CombinedSerializer,
+        state: &mut SerializationState<'py>,
+    ) -> Result<(), Self::Error> {
+        let key = key_serializer.json_key_no_infer(key, state)?;
+        let value_ser = PydanticSerializer::new_no_infer(value, value_serializer, state);
+        self.map.serialize_entry(&key, &value_ser).map_err(WrappedSerError)
+    }
+
+    fn serialize_entry_string_key<'py>(
+        &mut self,
+        key: &PyBackedStr,
+        value: &Bound<'py, PyAny>,
+        value_serializer: &CombinedSerializer,
+        state: &mut SerializationState<'py>,
+    ) -> Result<(), Self::Error> {
+        let value_ser = PydanticSerializer::new_no_infer(value, value_serializer, state);
+        self.map.serialize_entry(&**key, &value_ser).map_err(WrappedSerError)
+    }
+
+    fn end(self) -> Result<Self::Ok, Self::Error> {
+        self.map.end().map_err(WrappedSerError)
+    }
 }
