@@ -1,19 +1,24 @@
 use pyo3::intern;
 use pyo3::prelude::*;
+use pyo3::pybacked::PyBackedStr;
 use pyo3::types::{PyDict, PyList, PyString, PyType};
 use std::borrow::Cow;
+use std::sync::Arc;
 
 use ahash::AHashMap;
-use serde::ser::SerializeMap;
 
-use crate::build_tools::{py_schema_error_type, ExtraBehavior};
+use crate::build_tools::{ExtraBehavior, py_schema_error_type};
 use crate::definitions::DefinitionsBuilder;
+use crate::serializers::SerializationState;
+use crate::serializers::errors::unwrap_ser_error;
+use crate::serializers::shared::DoSerialize;
+use crate::serializers::shared::serialize_to_json;
+use crate::serializers::shared::serialize_to_python;
 use crate::tools::SchemaDict;
 
 use super::{
-    infer_json_key, infer_json_key_known, infer_serialize, infer_to_python, py_err_se_err, BuildSerializer,
-    CombinedSerializer, ComputedFields, Extra, FieldsMode, GeneralFieldsSerializer, ObType, SerCheck, SerField,
-    TypeSerializer,
+    BuildSerializer, CombinedSerializer, ComputedFields, FieldsMode, GeneralFieldsSerializer, ObType, SerCheck,
+    SerField, TypeSerializer, infer_json_key, infer_json_key_known,
 };
 
 pub struct DataclassArgsBuilder;
@@ -24,12 +29,12 @@ impl BuildSerializer for DataclassArgsBuilder {
     fn build(
         schema: &Bound<'_, PyDict>,
         config: Option<&Bound<'_, PyDict>>,
-        definitions: &mut DefinitionsBuilder<CombinedSerializer>,
-    ) -> PyResult<CombinedSerializer> {
+        definitions: &mut DefinitionsBuilder<Arc<CombinedSerializer>>,
+    ) -> PyResult<Arc<CombinedSerializer>> {
         let py = schema.py();
 
         let fields_list: Bound<'_, PyList> = schema.get_as_req(intern!(py, "fields"))?;
-        let mut fields: AHashMap<String, SerField> = AHashMap::with_capacity(fields_list.len());
+        let mut fields = AHashMap::with_capacity(fields_list.len());
 
         let fields_mode = match ExtraBehavior::from_schema_or_config(py, schema, config, ExtraBehavior::Ignore)? {
             ExtraBehavior::Allow => FieldsMode::TypedDictAllow,
@@ -39,29 +44,27 @@ impl BuildSerializer for DataclassArgsBuilder {
         let serialize_by_alias = config.get_as(intern!(py, "serialize_by_alias"))?;
 
         for (index, item) in fields_list.iter().enumerate() {
-            let field_info = item.downcast::<PyDict>()?;
-            let name: String = field_info.get_as_req(intern!(py, "name"))?;
+            let field_info = item.cast::<PyDict>()?;
+            let key: PyBackedStr = field_info.get_as_req(intern!(py, "name"))?;
 
-            let key_py: Py<PyString> = PyString::new(py, &name).into();
             if !field_info.get_as(intern!(py, "init_only"))?.unwrap_or(false) {
                 if field_info.get_as(intern!(py, "serialization_exclude"))? == Some(true) {
                     fields.insert(
-                        name,
-                        SerField::new(py, key_py, None, None, true, serialize_by_alias, None),
+                        key.clone_ref(py),
+                        SerField::new(key, None, None, true, serialize_by_alias, None),
                     );
                 } else {
                     let schema = field_info.get_as_req(intern!(py, "schema"))?;
                     let serializer = CombinedSerializer::build(&schema, config, definitions)
-                        .map_err(|e| py_schema_error_type!("Field `{}`:\n  {}", index, e))?;
+                        .map_err(|e| py_schema_error_type!("Field `{index}`:\n  {e}"))?;
 
                     let alias = field_info.get_as(intern!(py, "serialization_alias"))?;
                     let serialization_exclude_if: Option<Py<PyAny>> =
                         field_info.get_as(intern!(py, "serialization_exclude_if"))?;
                     fields.insert(
-                        name,
+                        key.clone_ref(py),
                         SerField::new(
-                            py,
-                            key_py,
+                            key,
                             alias,
                             Some(serializer),
                             true,
@@ -74,14 +77,14 @@ impl BuildSerializer for DataclassArgsBuilder {
         }
         let computed_fields = ComputedFields::new(schema, config, definitions)?;
 
-        Ok(GeneralFieldsSerializer::new(fields, fields_mode, None, computed_fields).into())
+        Ok(CombinedSerializer::Fields(GeneralFieldsSerializer::new(fields, fields_mode, None, computed_fields)).into())
     }
 }
 
 #[derive(Debug)]
 pub struct DataclassSerializer {
     class: Py<PyType>,
-    serializer: Box<CombinedSerializer>,
+    serializer: Arc<CombinedSerializer>,
     fields: Vec<Py<PyString>>,
     name: String,
 }
@@ -92,8 +95,8 @@ impl BuildSerializer for DataclassSerializer {
     fn build(
         schema: &Bound<'_, PyDict>,
         _config: Option<&Bound<'_, PyDict>>,
-        definitions: &mut DefinitionsBuilder<CombinedSerializer>,
-    ) -> PyResult<CombinedSerializer> {
+        definitions: &mut DefinitionsBuilder<Arc<CombinedSerializer>>,
+    ) -> PyResult<Arc<CombinedSerializer>> {
         let py = schema.py();
 
         // models ignore the parent config and always use the config from this model
@@ -101,27 +104,27 @@ impl BuildSerializer for DataclassSerializer {
 
         let class: Bound<'_, PyType> = schema.get_as_req(intern!(py, "cls"))?;
         let sub_schema = schema.get_as_req(intern!(py, "schema"))?;
-        let serializer = Box::new(CombinedSerializer::build(&sub_schema, config.as_ref(), definitions)?);
+        let serializer = CombinedSerializer::build(&sub_schema, config.as_ref(), definitions)?;
 
         let fields = schema
             .get_as_req::<Bound<'_, PyList>>(intern!(py, "fields"))?
             .iter()
-            .map(|s| Ok(s.downcast_into::<PyString>()?.unbind()))
+            .map(|s| Ok(s.cast_into::<PyString>()?.unbind()))
             .collect::<PyResult<Vec<_>>>()?;
 
-        Ok(Self {
+        Ok(CombinedSerializer::Dataclass(Self {
             class: class.clone().unbind(),
             serializer,
             fields,
             name: class.getattr(intern!(py, "__name__"))?.extract()?,
-        }
+        })
         .into())
     }
 }
 
 impl DataclassSerializer {
-    fn allow_value(&self, value: &Bound<'_, PyAny>, extra: &Extra) -> PyResult<bool> {
-        match extra.check {
+    fn allow_value(&self, value: &Bound<'_, PyAny>, state: &SerializationState<'_>) -> PyResult<bool> {
+        match state.check {
             SerCheck::Strict => Ok(value.get_type().is(self.class.bind(value.py()))),
             SerCheck::Lax => value.is_instance(self.class.bind(value.py())),
             SerCheck::None => value.hasattr(intern!(value.py(), "__dataclass_fields__")),
@@ -138,84 +141,62 @@ impl DataclassSerializer {
         }
         Ok(dict)
     }
+
+    fn serialize<'py, S: DoSerialize>(
+        &self,
+        value: &Bound<'py, PyAny>,
+        state: &mut SerializationState<'py>,
+        do_serialize: S,
+    ) -> Result<S::Ok, S::Error> {
+        if self.allow_value(value, state)? {
+            let state = &mut state.scoped_set(|s| &mut s.model, Some(value.clone()));
+            if let CombinedSerializer::Fields(ref fields_serializer) = *self.serializer {
+                // Fast path uses known dataclass fields to avoid needing to create a dict of field values
+                fields_serializer.serialize_iterators(
+                    value,
+                    known_dataclass_iter(&self.fields, value),
+                    std::iter::empty(),
+                    state,
+                    do_serialize,
+                )
+            } else {
+                let inner_value = self.get_inner_value(value)?;
+                do_serialize.serialize_no_infer(&self.serializer, &inner_value, state)
+            }
+        } else {
+            do_serialize.serialize_fallback(self.get_name(), value, state)
+        }
+    }
 }
 
 impl_py_gc_traverse!(DataclassSerializer { class, serializer });
 
 impl TypeSerializer for DataclassSerializer {
-    fn to_python(
-        &self,
-        value: &Bound<'_, PyAny>,
-        include: Option<&Bound<'_, PyAny>>,
-        exclude: Option<&Bound<'_, PyAny>>,
-        extra: &Extra,
-    ) -> PyResult<PyObject> {
-        let model = Some(value);
-        let dc_extra = Extra { model, ..*extra };
-        if self.allow_value(value, &dc_extra)? {
-            let py = value.py();
-            if let CombinedSerializer::Fields(ref fields_serializer) = *self.serializer {
-                let output_dict: Bound<PyDict> = fields_serializer.main_to_python(
-                    py,
-                    known_dataclass_iter(&self.fields, value),
-                    include,
-                    exclude,
-                    dc_extra,
-                )?;
+    fn to_python<'py>(&self, value: &Bound<'py, PyAny>, state: &mut SerializationState<'py>) -> PyResult<Py<PyAny>> {
+        self.serialize(value, state, serialize_to_python(value.py()))
+    }
 
-                fields_serializer.add_computed_fields_python(model, &output_dict, include, exclude, extra)?;
-                Ok(output_dict.into())
-            } else {
-                let inner_value = self.get_inner_value(value)?;
-                self.serializer.to_python(&inner_value, include, exclude, &dc_extra)
-            }
+    fn json_key<'a, 'py>(
+        &self,
+        key: &'a Bound<'py, PyAny>,
+        state: &mut SerializationState<'py>,
+    ) -> PyResult<Cow<'a, str>> {
+        if self.allow_value(key, state)? {
+            infer_json_key_known(ObType::Dataclass, key, state)
         } else {
-            extra.warnings.on_fallback_py(self.get_name(), value, &dc_extra)?;
-            infer_to_python(value, include, exclude, &dc_extra)
+            state.warn_fallback_py(&self.name, key)?;
+            infer_json_key(key, state)
         }
     }
 
-    fn json_key<'a>(&self, key: &'a Bound<'_, PyAny>, extra: &Extra) -> PyResult<Cow<'a, str>> {
-        if self.allow_value(key, extra)? {
-            infer_json_key_known(ObType::Dataclass, key, extra)
-        } else {
-            extra.warnings.on_fallback_py(&self.name, key, extra)?;
-            infer_json_key(key, extra)
-        }
-    }
-
-    fn serde_serialize<S: serde::ser::Serializer>(
+    fn serde_serialize<'py, S: serde::ser::Serializer>(
         &self,
-        value: &Bound<'_, PyAny>,
+        value: &Bound<'py, PyAny>,
         serializer: S,
-        include: Option<&Bound<'_, PyAny>>,
-        exclude: Option<&Bound<'_, PyAny>>,
-        extra: &Extra,
+        state: &mut SerializationState<'py>,
     ) -> Result<S::Ok, S::Error> {
-        let model = Some(value);
-        let dc_extra = Extra { model, ..*extra };
-        if self.allow_value(value, &dc_extra).map_err(py_err_se_err)? {
-            if let CombinedSerializer::Fields(ref fields_serializer) = *self.serializer {
-                let expected_len = self.fields.len() + fields_serializer.computed_field_count();
-                let mut map = fields_serializer.main_serde_serialize(
-                    known_dataclass_iter(&self.fields, value),
-                    expected_len,
-                    serializer,
-                    include,
-                    exclude,
-                    dc_extra,
-                )?;
-                fields_serializer.add_computed_fields_json::<S>(model, &mut map, include, exclude, extra)?;
-                map.end()
-            } else {
-                let inner_value = self.get_inner_value(value).map_err(py_err_se_err)?;
-                self.serializer
-                    .serde_serialize(&inner_value, serializer, include, exclude, &dc_extra)
-            }
-        } else {
-            extra.warnings.on_fallback_ser::<S>(self.get_name(), value, &dc_extra)?;
-            infer_serialize(value, serializer, include, exclude, &dc_extra)
-        }
+        self.serialize(value, state, serialize_to_json(serializer))
+            .map_err(unwrap_ser_error)
     }
 
     fn get_name(&self) -> &str {
