@@ -9,11 +9,15 @@ use serde::ser::Error;
 
 use crate::build_tools::py_schema_err;
 use crate::definitions::DefinitionsBuilder;
+use crate::serializers::SerializationState;
+use crate::serializers::errors::unwrap_ser_error;
+use crate::serializers::shared::DoSerialize;
+use crate::serializers::shared::serialize_to_json;
+use crate::serializers::shared::serialize_to_python;
 use crate::tools::SchemaDict;
 
 use super::simple::none_json_key;
-use super::string::serialize_py_str;
-use super::{py_err_se_err, BuildSerializer, CombinedSerializer, Extra, PydanticSerializationError, TypeSerializer};
+use super::{BuildSerializer, CombinedSerializer, Extra, PydanticSerializationError, TypeSerializer, py_err_se_err};
 
 #[derive(Debug, Clone, Eq, PartialEq)]
 pub(super) enum WhenUsed {
@@ -108,27 +112,21 @@ impl FormatSerializer {
 impl_py_gc_traverse!(FormatSerializer { format_func });
 
 impl TypeSerializer for FormatSerializer {
-    fn to_python(
-        &self,
-        value: &Bound<'_, PyAny>,
-        _include: Option<&Bound<'_, PyAny>>,
-        _exclude: Option<&Bound<'_, PyAny>>,
-        extra: &Extra,
-    ) -> PyResult<Py<PyAny>> {
-        if self.when_used.should_use(value, extra) {
+    fn to_python(&self, value: &Bound<'_, PyAny>, state: &mut SerializationState<'_>) -> PyResult<Py<PyAny>> {
+        if self.when_used.should_use(value, &state.extra) {
             self.call(value).map_err(PydanticSerializationError::new_err)
         } else {
             Ok(value.clone().unbind())
         }
     }
 
-    fn json_key<'a>(&self, key: &'a Bound<'_, PyAny>, _extra: &Extra) -> PyResult<Cow<'a, str>> {
+    fn json_key<'a>(&self, key: &'a Bound<'_, PyAny>, _state: &mut SerializationState<'_>) -> PyResult<Cow<'a, str>> {
         if self.when_used.should_use_json(key) {
             let py_str = self
                 .call(key)
                 .map_err(PydanticSerializationError::new_err)?
                 .into_bound(key.py())
-                .downcast_into::<PyString>()?;
+                .cast_into::<PyString>()?;
             Ok(Cow::Owned(py_str.to_str()?.to_owned()))
         } else {
             none_json_key()
@@ -139,15 +137,15 @@ impl TypeSerializer for FormatSerializer {
         &self,
         value: &Bound<'_, PyAny>,
         serializer: S,
-        _include: Option<&Bound<'_, PyAny>>,
-        _exclude: Option<&Bound<'_, PyAny>>,
-        _extra: &Extra,
+        _state: &mut SerializationState<'_>,
     ) -> Result<S::Ok, S::Error> {
         if self.when_used.should_use_json(value) {
             match self.call(value) {
                 Ok(v) => {
-                    let py_str = v.bind(value.py()).downcast().map_err(py_err_se_err)?;
-                    serialize_py_str(py_str, serializer)
+                    let py_str = v.bind(value.py()).cast().map_err(py_err_se_err)?;
+                    serialize_to_json(serializer)
+                        .serialize_str(py_str)
+                        .map_err(unwrap_ser_error)
                 }
                 Err(e) => Err(S::Error::custom(e)),
             }
@@ -184,21 +182,19 @@ impl BuildSerializer for ToStringSerializer {
 impl_py_gc_traverse!(ToStringSerializer {});
 
 impl TypeSerializer for ToStringSerializer {
-    fn to_python(
-        &self,
-        value: &Bound<'_, PyAny>,
-        _include: Option<&Bound<'_, PyAny>>,
-        _exclude: Option<&Bound<'_, PyAny>>,
-        extra: &Extra,
-    ) -> PyResult<Py<PyAny>> {
-        if self.when_used.should_use(value, extra) {
-            value.str().map(Into::into)
+    fn to_python(&self, value: &Bound<'_, PyAny>, state: &mut SerializationState<'_>) -> PyResult<Py<PyAny>> {
+        if self.when_used.should_use(value, &state.extra) {
+            serialize_via_str(value, serialize_to_python(value.py()))
         } else {
             Ok(value.clone().unbind())
         }
     }
 
-    fn json_key<'a>(&self, key: &'a Bound<'_, PyAny>, _extra: &Extra) -> PyResult<Cow<'a, str>> {
+    fn json_key<'a, 'py>(
+        &self,
+        key: &'a Bound<'py, PyAny>,
+        _state: &mut SerializationState<'py>,
+    ) -> PyResult<Cow<'a, str>> {
         if self.when_used.should_use_json(key) {
             Ok(Cow::Owned(key.str()?.to_string_lossy().into_owned()))
         } else {
@@ -210,13 +206,10 @@ impl TypeSerializer for ToStringSerializer {
         &self,
         value: &Bound<'_, PyAny>,
         serializer: S,
-        _include: Option<&Bound<'_, PyAny>>,
-        _exclude: Option<&Bound<'_, PyAny>>,
-        _extra: &Extra,
+        _state: &mut SerializationState<'_>,
     ) -> Result<S::Ok, S::Error> {
         if self.when_used.should_use_json(value) {
-            let s = value.str().map_err(py_err_se_err)?;
-            serialize_py_str(&s, serializer)
+            serialize_via_str(value, serialize_to_json(serializer)).map_err(|e| e.0)
         } else {
             serializer.serialize_none()
         }
@@ -226,3 +219,9 @@ impl TypeSerializer for ToStringSerializer {
         Self::EXPECTED_TYPE
     }
 }
+
+/// Serialize a value by calling `str()` on it
+pub fn serialize_via_str<S: DoSerialize>(value: &Bound<'_, PyAny>, do_serialize: S) -> Result<S::Ok, S::Error> {
+    let s = value.str()?;
+    do_serialize.serialize_str(&s)
+}
