@@ -7,15 +7,17 @@ use std::sync::Arc;
 
 use serde::ser::SerializeSeq;
 
+use crate::PydanticSerializationUnexpectedValue;
 use crate::definitions::DefinitionsBuilder;
+use crate::serializers::SerializationState;
+use crate::serializers::extra::IncludeExclude;
 use crate::serializers::extra::SerCheck;
 use crate::serializers::type_serializers::any::AnySerializer;
 use crate::tools::SchemaDict;
-use crate::PydanticSerializationUnexpectedValue;
 
 use super::{
-    infer_json_key, infer_serialize, infer_to_python, py_err_se_err, BuildSerializer, CombinedSerializer, Extra,
-    PydanticSerializer, SchemaFilter, SerMode, TypeSerializer,
+    BuildSerializer, CombinedSerializer, PydanticSerializer, SchemaFilter, SerMode, TypeSerializer, infer_json_key,
+    infer_serialize, infer_to_python, py_err_se_err,
 };
 
 #[derive(Debug)]
@@ -38,7 +40,7 @@ impl BuildSerializer for TupleSerializer {
         let items: Bound<'_, PyList> = schema.get_as_req(intern!(py, "items_schema"))?;
         let serializers: Vec<Arc<CombinedSerializer>> = items
             .iter()
-            .map(|item| CombinedSerializer::build(item.downcast()?, config, definitions))
+            .map(|item| CombinedSerializer::build(item.cast()?, config, definitions))
             .collect::<PyResult<_>>()?;
 
         let mut serializer_names = serializers.iter().map(|v| v.get_name()).collect::<Vec<_>>();
@@ -61,91 +63,82 @@ impl BuildSerializer for TupleSerializer {
 impl_py_gc_traverse!(TupleSerializer { serializers });
 
 impl TypeSerializer for TupleSerializer {
-    fn to_python(
-        &self,
-        value: &Bound<'_, PyAny>,
-        include: Option<&Bound<'_, PyAny>>,
-        exclude: Option<&Bound<'_, PyAny>>,
-        extra: &Extra,
-    ) -> PyResult<Py<PyAny>> {
-        match value.downcast::<PyTuple>() {
+    fn to_python<'py>(&self, value: &Bound<'py, PyAny>, state: &mut SerializationState<'py>) -> PyResult<Py<PyAny>> {
+        match value.cast::<PyTuple>() {
             Ok(py_tuple) => {
                 let py = value.py();
 
                 let n_items = py_tuple.len();
                 let mut items = Vec::with_capacity(n_items);
 
-                self.for_each_tuple_item_and_serializer(py_tuple, include, exclude, extra, |entry| {
+                self.for_each_tuple_item_and_serializer(py_tuple, state, |entry| {
                     entry
                         .serializer
-                        .to_python(&entry.item, entry.include.as_ref(), entry.exclude.as_ref(), extra)
+                        .to_python(&entry.item, entry.state)
                         .map(|item| items.push(item))
                 })??;
 
-                match extra.mode {
+                match state.extra.mode {
                     SerMode::Json => Ok(PyList::new(py, items)?.into()),
                     _ => Ok(PyTuple::new(py, items)?.into()),
                 }
             }
             Err(_) => {
-                extra.warnings.on_fallback_py(&self.name, value, extra)?;
-                infer_to_python(value, include, exclude, extra)
+                state.warn_fallback_py(&self.name, value)?;
+                infer_to_python(value, state)
             }
         }
     }
 
-    fn json_key<'a>(&self, key: &'a Bound<'_, PyAny>, extra: &Extra) -> PyResult<Cow<'a, str>> {
-        match key.downcast::<PyTuple>() {
+    fn json_key<'a, 'py>(
+        &self,
+        key: &'a Bound<'py, PyAny>,
+        state: &mut SerializationState<'py>,
+    ) -> PyResult<Cow<'a, str>> {
+        match key.cast::<PyTuple>() {
             Ok(py_tuple) => {
                 let mut key_builder = KeyBuilder::new();
 
-                self.for_each_tuple_item_and_serializer(py_tuple, None, None, extra, |entry| {
+                let state = &mut state.scoped_include_exclude(IncludeExclude::empty());
+                self.for_each_tuple_item_and_serializer(py_tuple, state, |entry| {
                     entry
                         .serializer
-                        .json_key(&entry.item, extra)
+                        .json_key(&entry.item, entry.state)
                         .map(|key| key_builder.push(&key))
                 })??;
 
                 Ok(Cow::Owned(key_builder.finish()))
             }
             Err(_) => {
-                extra.warnings.on_fallback_py(&self.name, key, extra)?;
-                infer_json_key(key, extra)
+                state.warn_fallback_py(&self.name, key)?;
+                infer_json_key(key, state)
             }
         }
     }
 
-    fn serde_serialize<S: serde::ser::Serializer>(
+    fn serde_serialize<'py, S: serde::ser::Serializer>(
         &self,
-        value: &Bound<'_, PyAny>,
+        value: &Bound<'py, PyAny>,
         serializer: S,
-        include: Option<&Bound<'_, PyAny>>,
-        exclude: Option<&Bound<'_, PyAny>>,
-        extra: &Extra,
+        state: &mut SerializationState<'py>,
     ) -> Result<S::Ok, S::Error> {
-        match value.downcast::<PyTuple>() {
+        match value.cast::<PyTuple>() {
             Ok(py_tuple) => {
-                let py_tuple = py_tuple.downcast::<PyTuple>().map_err(py_err_se_err)?;
+                let py_tuple = py_tuple.cast::<PyTuple>().map_err(py_err_se_err)?;
 
                 let n_items = py_tuple.len();
                 let mut seq = serializer.serialize_seq(Some(n_items))?;
 
-                self.for_each_tuple_item_and_serializer(py_tuple, include, exclude, extra, |entry| {
-                    seq.serialize_element(&PydanticSerializer::new(
-                        &entry.item,
-                        entry.serializer,
-                        entry.include.as_ref(),
-                        entry.exclude.as_ref(),
-                        extra,
-                    ))
+                self.for_each_tuple_item_and_serializer(py_tuple, state, |entry| {
+                    seq.serialize_element(&PydanticSerializer::new(&entry.item, entry.serializer, entry.state))
                 })
                 .map_err(py_err_se_err)??;
 
                 seq.end()
             }
             Err(_) => {
-                extra.warnings.on_fallback_ser::<S>(&self.name, value, extra)?;
-                infer_serialize(value, serializer, include, exclude, extra)
+                state.warn_fallback_ser::<S>(&self.name, value)?;
+                infer_serialize(value, serializer, state)
             }
         }
     }
@@ -161,9 +154,8 @@ impl TypeSerializer for TupleSerializer {
 
 struct TupleSerializerEntry<'a, 'py> {
     item: Bound<'py, PyAny>,
-    include: Option<Bound<'py, PyAny>>,
-    exclude: Option<Bound<'py, PyAny>>,
     serializer: &'a CombinedSerializer,
+    state: &'a mut SerializationState<'py>,
 }
 
 impl TupleSerializer {
@@ -173,13 +165,11 @@ impl TupleSerializer {
     ///
     /// The error type E is the type of the error returned by the closure, which is why there are two
     /// levels of `Result`.
-    fn for_each_tuple_item_and_serializer<E>(
+    fn for_each_tuple_item_and_serializer<'py, E>(
         &self,
-        tuple: &Bound<'_, PyTuple>,
-        include: Option<&Bound<'_, PyAny>>,
-        exclude: Option<&Bound<'_, PyAny>>,
-        extra: &Extra,
-        mut f: impl for<'a, 'py> FnMut(TupleSerializerEntry<'a, 'py>) -> Result<(), E>,
+        tuple: &Bound<'py, PyTuple>,
+        state: &mut SerializationState<'py>,
+        mut f: impl for<'a> FnMut(TupleSerializerEntry<'a, 'py>) -> Result<(), E>,
     ) -> PyResult<Result<(), E>> {
         let n_items = tuple.len();
         let mut py_tuple_iter = tuple.iter();
@@ -190,13 +180,12 @@ impl TupleSerializer {
                     let Some(element) = py_tuple_iter.next() else {
                         break;
                     };
-                    let op_next = self.filter.index_filter(index, include, exclude, Some(n_items))?;
-                    if let Some((next_include, next_exclude)) = op_next {
+                    if let Some(next_include_exclude) = self.filter.index_filter(index, state, Some(n_items))? {
+                        let state = &mut state.scoped_include_exclude(next_include_exclude);
                         if let Err(e) = f(TupleSerializerEntry {
                             item: element,
-                            include: next_include,
-                            exclude: next_exclude,
                             serializer,
+                            state,
                         }) {
                             return Ok(Err(e));
                         };
@@ -210,10 +199,10 @@ impl TupleSerializer {
             let n_variadic_items = (n_items + 1).saturating_sub(self.serializers.len());
             let serializers_iter = self.serializers[..variadic_item_index]
                 .iter()
-                .chain(iter::repeat(&self.serializers[variadic_item_index]).take(n_variadic_items))
+                .chain(iter::repeat_n(&self.serializers[variadic_item_index], n_variadic_items))
                 .chain(self.serializers[variadic_item_index + 1..].iter());
             use_serializers!(serializers_iter);
-        } else if extra.check == SerCheck::Strict && n_items != self.serializers.len() {
+        } else if state.check == SerCheck::Strict && n_items != self.serializers.len() {
             return Err(PydanticSerializationUnexpectedValue::new_from_msg(Some(format!(
                 "Expected {} items, but got {}",
                 self.serializers.len(),
@@ -225,22 +214,20 @@ impl TupleSerializer {
             let mut warned = false;
             for (i, element) in py_tuple_iter.enumerate() {
                 if !warned {
-                    extra
+                    state
                         .warnings
                         .register_warning(PydanticSerializationUnexpectedValue::new_from_msg(Some(
                             "Unexpected extra items present in tuple".to_string(),
                         )));
                     warned = true;
                 }
-                let op_next = self
-                    .filter
-                    .index_filter(i + self.serializers.len(), include, exclude, Some(n_items))?;
-                if let Some((next_include, next_exclude)) = op_next {
+                let index = i + self.serializers.len();
+                if let Some(next_include_exclude) = self.filter.index_filter(index, state, Some(n_items))? {
+                    let state = &mut state.scoped_include_exclude(next_include_exclude);
                     if let Err(e) = f(TupleSerializerEntry {
                         item: element,
-                        include: next_include,
-                        exclude: next_exclude,
                         serializer: &CombinedSerializer::Any(AnySerializer),
+                        state,
                     }) {
                         return Ok(Err(e));
                     }
