@@ -1,28 +1,34 @@
 use std::borrow::Cow;
 use std::collections::hash_map::DefaultHasher;
-use std::fmt;
-use std::fmt::Formatter;
+use std::fmt::{self, Display};
+use std::fmt::{Formatter, Write};
 use std::hash::{Hash, Hasher};
 use std::sync::OnceLock;
 
 use idna::punycode::decode_to_string;
 use jiter::{PartialMode, StringCacheMode};
-use percent_encoding::{percent_encode, NON_ALPHANUMERIC};
+use percent_encoding::{AsciiSet, CONTROLS, percent_encode};
 use pyo3::exceptions::PyValueError;
 use pyo3::pyclass::CompareOp;
 use pyo3::sync::OnceLockExt;
 use pyo3::types::{PyDict, PyType};
-use pyo3::{intern, prelude::*, IntoPyObjectExt};
+use pyo3::{IntoPyObjectExt, intern, prelude::*};
 use url::Url;
 
+use crate::ValidationError;
 use crate::input::InputType;
 use crate::recursion_guard::RecursionState;
 use crate::tools::SchemaDict;
 use crate::validators::url::{MultiHostUrlValidator, UrlValidator};
 use crate::validators::{Extra, ValidationState, Validator};
-use crate::ValidationError;
 
-#[pyclass(name = "Url", module = "pydantic_core._pydantic_core", subclass, frozen)]
+#[pyclass(
+    name = "Url",
+    module = "pydantic_core._pydantic_core",
+    subclass,
+    frozen,
+    skip_from_py_object
+)]
 #[derive(Clone)]
 #[cfg_attr(debug_assertions, derive(Debug))]
 pub struct PyUrl {
@@ -94,6 +100,7 @@ impl PyUrl {
                     ),
                     &mut RecursionState::default(),
                     PartialMode::Off,
+                    None,
                 ),
             )
             .map_err(|e| {
@@ -103,7 +110,7 @@ impl PyUrl {
                 };
                 ValidationError::from_val_error(py, name, InputType::Python, e, None, false, false)
             })?
-            .downcast_bound::<Self>(py)?
+            .cast_bound::<Self>(py)?
             .get()
             .clone(); // FIXME: avoid the clone, would need to make `validate` be aware of what URL subclass to create
         Ok(url_obj)
@@ -229,6 +236,27 @@ impl PyUrl {
         path: Option<&str>,
         query: Option<&str>,
         fragment: Option<&str>,
+        // encode_credentials: bool, // TODO: re-enable this
+    ) -> PyResult<Bound<'py, PyAny>> {
+        Self::build_inner(
+            cls, scheme, host, username, password, port, path, query, fragment, false,
+        )
+    }
+}
+
+impl PyUrl {
+    #[allow(clippy::too_many_arguments)]
+    fn build_inner<'py>(
+        cls: &Bound<'py, PyType>,
+        scheme: &str,
+        host: &str,
+        username: Option<&str>,
+        password: Option<&str>,
+        port: Option<u16>,
+        path: Option<&str>,
+        query: Option<&str>,
+        fragment: Option<&str>,
+        encode_credentials: bool,
     ) -> PyResult<Bound<'py, PyAny>> {
         let url_host = UrlHostParts {
             username: username.map(Into::into),
@@ -236,7 +264,10 @@ impl PyUrl {
             host: Some(host.into()),
             port,
         };
-        let mut url = format!("{scheme}://{url_host}");
+        let mut url = format!("{scheme}://");
+        url_host
+            .to_writer(&mut url, encode_credentials)
+            .expect("writing to string should not fail");
         if let Some(path) = path {
             url.push('/');
             url.push_str(path);
@@ -253,7 +284,13 @@ impl PyUrl {
     }
 }
 
-#[pyclass(name = "MultiHostUrl", module = "pydantic_core._pydantic_core", subclass, frozen)]
+#[pyclass(
+    name = "MultiHostUrl",
+    module = "pydantic_core._pydantic_core",
+    subclass,
+    frozen,
+    skip_from_py_object
+)]
 #[derive(Clone, Hash)]
 #[cfg_attr(debug_assertions, derive(Debug))]
 pub struct PyMultiHostUrl {
@@ -299,6 +336,7 @@ impl PyMultiHostUrl {
                     ),
                     &mut RecursionState::default(),
                     PartialMode::Off,
+                    None,
                 ),
             )
             .map_err(|e| {
@@ -308,7 +346,7 @@ impl PyMultiHostUrl {
                 };
                 ValidationError::from_val_error(py, name, InputType::Python, e, None, false, false)
             })?
-            .downcast_bound::<Self>(py)?
+            .cast_bound::<Self>(py)?
             .get()
             .clone(); // FIXME: avoid the clone, would need to make `validate` be aware of what URL subclass to create
         Ok(url_obj)
@@ -360,53 +398,49 @@ impl PyMultiHostUrl {
             let host_offset = scheme.len() + 3;
 
             let mut full_url = self.ref_url.unicode_string(py).into_owned();
-            full_url.insert(host_offset, ',');
+            let mut extra_hosts = String::new();
 
             // special urls will have had a trailing slash added, non-special urls will not
             // hence we need to remove the last char if the scheme is special
             #[allow(clippy::bool_to_int_with_if)]
             let sub = if scheme_is_special(scheme) { 1 } else { 0 };
 
-            let hosts = extra_urls
-                .iter()
-                .map(|url| {
-                    let str = unicode_url(url.as_str(), url);
-                    str[host_offset..str.len() - sub].to_string()
-                })
-                .collect::<Vec<String>>()
-                .join(",");
-            full_url.insert_str(host_offset, &hosts);
+            for url in extra_urls {
+                let str = unicode_url(url.as_str(), url);
+                extra_hosts.push_str(&str[host_offset..str.len() - sub]);
+                extra_hosts.push(',');
+            }
+
+            full_url.insert_str(host_offset, &extra_hosts);
             Cow::Owned(full_url)
         } else {
             self.ref_url.unicode_string(py)
         }
     }
 
-    pub fn __str__(&self, py: Python<'_>) -> String {
+    pub fn __str__(&self, py: Python<'_>) -> Cow<'_, str> {
         if let Some(extra_urls) = &self.extra_urls {
             let scheme = self.ref_url.lib_url.scheme();
             let host_offset = scheme.len() + 3;
 
             let mut full_url = self.ref_url.serialized(py).to_string();
-            full_url.insert(host_offset, ',');
+            let mut extra_hosts = String::new();
 
             // special urls will have had a trailing slash added, non-special urls will not
             // hence we need to remove the last char if the scheme is special
             #[allow(clippy::bool_to_int_with_if)]
             let sub = if scheme_is_special(scheme) { 1 } else { 0 };
 
-            let hosts = extra_urls
-                .iter()
-                .map(|url| {
-                    let str = url.as_str();
-                    &str[host_offset..str.len() - sub]
-                })
-                .collect::<Vec<&str>>()
-                .join(",");
-            full_url.insert_str(host_offset, &hosts);
-            full_url
+            for url in extra_urls {
+                let str = url.as_str();
+                extra_hosts.push_str(&str[host_offset..str.len() - sub]);
+                extra_hosts.push(',');
+            }
+
+            full_url.insert_str(host_offset, &extra_hosts);
+            Cow::Owned(full_url)
         } else {
-            self.ref_url.__str__(py).to_string()
+            Cow::Borrowed(self.ref_url.__str__(py))
         }
     }
 
@@ -439,14 +473,14 @@ impl PyMultiHostUrl {
         self.clone().into_py_any(py)
     }
 
-    fn __getnewargs__(&self, py: Python<'_>) -> (String,) {
+    fn __getnewargs__(&self, py: Python<'_>) -> (Cow<'_, str>,) {
         (self.__str__(py),)
     }
 
     #[classmethod]
     #[pyo3(signature=(*, scheme, hosts=None, path=None, query=None, fragment=None, host=None, username=None, password=None, port=None))]
     #[allow(clippy::too_many_arguments)]
-    pub fn build<'py>(
+    fn build<'py>(
         cls: &Bound<'py, PyType>,
         scheme: &str,
         hosts: Option<Vec<UrlHostParts>>,
@@ -458,39 +492,67 @@ impl PyMultiHostUrl {
         username: Option<&str>,
         password: Option<&str>,
         port: Option<u16>,
+        // encode_credentials: bool, // TODO: re-enable this
     ) -> PyResult<Bound<'py, PyAny>> {
-        let mut url =
-            if hosts.is_some() && (host.is_some() || username.is_some() || password.is_some() || port.is_some()) {
-                return Err(PyValueError::new_err(
-                    "expected one of `hosts` or singular values to be set.",
-                ));
-            } else if let Some(hosts) = hosts {
-                // check all of host / user / password / port empty
-                // build multi-host url
-                let mut multi_url = format!("{scheme}://");
-                for (index, single_host) in hosts.iter().enumerate() {
-                    if single_host.is_empty() {
-                        return Err(PyValueError::new_err(
-                            "expected one of 'host', 'username', 'password' or 'port' to be set",
-                        ));
-                    }
-                    multi_url.push_str(&single_host.to_string());
-                    if index != hosts.len() - 1 {
-                        multi_url.push(',');
-                    }
+        Self::build_inner(
+            cls, scheme, hosts, path, query, fragment, host, username, password, port,
+            false, // TODO: re-enable this
+        )
+    }
+}
+
+impl PyMultiHostUrl {
+    #[allow(clippy::too_many_arguments)]
+    fn build_inner<'py>(
+        cls: &Bound<'py, PyType>,
+        scheme: &str,
+        hosts: Option<Vec<UrlHostParts>>,
+        path: Option<&str>,
+        query: Option<&str>,
+        fragment: Option<&str>,
+        // convenience parameters to build with a single host
+        host: Option<&str>,
+        username: Option<&str>,
+        password: Option<&str>,
+        port: Option<u16>,
+        encode_credentials: bool,
+    ) -> PyResult<Bound<'py, PyAny>> {
+        let mut url = format!("{scheme}://");
+
+        if hosts.is_some() && (host.is_some() || username.is_some() || password.is_some() || port.is_some()) {
+            return Err(PyValueError::new_err(
+                "expected one of `hosts` or singular values to be set.",
+            ));
+        } else if let Some(hosts) = hosts {
+            // check all of host / user / password / port empty
+            // build multi-host url
+            let len = hosts.len();
+            for (index, single_host) in hosts.into_iter().enumerate() {
+                if single_host.is_empty() {
+                    return Err(PyValueError::new_err(
+                        "expected one of 'host', 'username', 'password' or 'port' to be set",
+                    ));
                 }
-                multi_url
-            } else if host.is_some() {
-                let url_host = UrlHostParts {
-                    username: username.map(Into::into),
-                    password: password.map(Into::into),
-                    host: host.map(Into::into),
-                    port,
-                };
-                format!("{scheme}://{url_host}")
-            } else {
-                return Err(PyValueError::new_err("expected either `host` or `hosts` to be set"));
+                single_host
+                    .to_writer(&mut url, encode_credentials)
+                    .expect("writing to string should not fail");
+                if index != len - 1 {
+                    url.push(',');
+                }
+            }
+        } else if host.is_some() {
+            let url_host = UrlHostParts {
+                username: username.map(Into::into),
+                password: password.map(Into::into),
+                host: host.map(Into::into),
+                port,
             };
+            url_host
+                .to_writer(&mut url, encode_credentials)
+                .expect("writing to string should not fail");
+        } else {
+            return Err(PyValueError::new_err("expected either `host` or `hosts` to be set"));
+        }
 
         if let Some(path) = path {
             url.push('/');
@@ -508,23 +570,57 @@ impl PyMultiHostUrl {
     }
 }
 
-pub struct UrlHostParts {
+struct UrlHostParts {
     username: Option<String>,
     password: Option<String>,
     host: Option<String>,
     port: Option<u16>,
 }
 
+struct MaybeEncoded<'a>(&'a str, bool);
+
+impl fmt::Display for MaybeEncoded<'_> {
+    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
+        if self.1 {
+            write!(f, "{}", encode_userinfo_component(self.0))
+        } else {
+            write!(f, "{}", self.0)
+        }
+    }
+}
+
 impl UrlHostParts {
     fn is_empty(&self) -> bool {
         self.host.is_none() && self.password.is_none() && self.host.is_none() && self.port.is_none()
     }
+
+    fn to_writer(&self, mut w: impl Write, encode_credentials: bool) -> fmt::Result {
+        match (&self.username, &self.password) {
+            (Some(username), None) => write!(w, "{}@", MaybeEncoded(username, encode_credentials))?,
+            (None, Some(password)) => write!(w, ":{}@", MaybeEncoded(password, encode_credentials))?,
+            (Some(username), Some(password)) => write!(
+                w,
+                "{}:{}@",
+                MaybeEncoded(username, encode_credentials),
+                MaybeEncoded(password, encode_credentials)
+            )?,
+            (None, None) => {}
+        }
+        if let Some(host) = &self.host {
+            write!(w, "{host}")?;
+        }
+        if let Some(port) = self.port {
+            write!(w, ":{port}")?;
+        }
+        Ok(())
+    }
 }
 
-impl FromPyObject<'_> for UrlHostParts {
-    fn extract_bound(ob: &Bound<'_, PyAny>) -> PyResult<Self> {
+impl FromPyObject<'_, '_> for UrlHostParts {
+    type Error = PyErr;
+    fn extract(ob: Borrowed<'_, '_, PyAny>) -> PyResult<Self> {
         let py = ob.py();
-        let dict = ob.downcast::<PyDict>()?;
+        let dict = ob.cast::<PyDict>()?;
         Ok(UrlHostParts {
             username: dict.get_as(intern!(py, "username"))?,
             password: dict.get_as(intern!(py, "password"))?,
@@ -534,29 +630,6 @@ impl FromPyObject<'_> for UrlHostParts {
     }
 }
 
-impl fmt::Display for UrlHostParts {
-    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
-        match (&self.username, &self.password) {
-            (Some(username), None) => write!(f, "{}@", encode_userinfo_component(username))?,
-            (None, Some(password)) => write!(f, ":{}@", encode_userinfo_component(password))?,
-            (Some(username), Some(password)) => write!(
-                f,
-                "{}:{}@",
-                encode_userinfo_component(username),
-                encode_userinfo_component(password)
-            )?,
-            (None, None) => {}
-        }
-        if let Some(host) = &self.host {
-            write!(f, "{host}")?;
-        }
-        if let Some(port) = self.port {
-            write!(f, ":{port}")?;
-        }
-        Ok(())
-    }
-}
-
 fn host_to_dict<'a>(py: Python<'a>, lib_url: &Url) -> PyResult<Bound<'a, PyDict>> {
     let dict = PyDict::new(py);
     dict.set_item("username", Some(lib_url.username()).filter(|s| !s.is_empty()))?;
@@ -602,14 +675,38 @@ fn is_punnycode_domain(lib_url: &Url, domain: &str) -> bool {
     scheme_is_special(lib_url.scheme()) && domain.split('.').any(|part| part.starts_with(PUNYCODE_PREFIX))
 }
 
-fn encode_userinfo_component(value: &str) -> Cow<'_, str> {
-    let encoded = percent_encode(value.as_bytes(), NON_ALPHANUMERIC).to_string();
-    if encoded == value {
-        Cow::Borrowed(value)
-    } else {
-        Cow::Owned(encoded)
-    }
+/// See <https://url.spec.whatwg.org/#userinfo-percent-encode-set>
+const USERINFO_ENCODE_SET: &AsciiSet = &CONTROLS
+    // query percent-encodes is controls plus the below
+    .add(b' ')
+    .add(b'"')
+    .add(b'#')
+    .add(b'<')
+    .add(b'>')
+    // path percent-encodes is query percent-encodes plus the below
+    .add(b'?')
+    .add(b'^')
+    .add(b'`')
+    .add(b'{')
+    .add(b'}')
+    // userinfo percent-encodes is path percent-encodes plus the below
+    .add(b'/')
+    .add(b':')
+    .add(b';')
+    .add(b'=')
+    .add(b'@')
+    .add(b'[')
+    .add(b'\\')
+    .add(b']')
+    .add(b'|')
+    // https://datatracker.ietf.org/doc/html/rfc3986.html#section-2.4
+    // we must also percent-encode '%'
+    .add(b'%');
+
+fn encode_userinfo_component(value: &str) -> impl Display + '_ {
+    percent_encode(value.as_bytes(), USERINFO_ENCODE_SET)
 }
+
 // based on https://github.com/servo/rust-url/blob/1c1e406874b3d2aa6f36c5d2f3a5c2ea74af9efb/url/src/parser.rs#L161-L167
 pub fn scheme_is_special(scheme: &str) -> bool {
     matches!(scheme, "http" | "https" | "ws" | "wss" | "ftp" | "file")
