@@ -11,7 +11,7 @@ use strum::EnumMessage;
 use crate::build_tools::{is_strict, py_schema_error_type};
 use crate::build_tools::{py_schema_err, schema_or_config_same};
 use crate::errors::ToErrorValue;
-use crate::errors::{py_err_string, ErrorType, ErrorTypeDefaults, ValError, ValResult};
+use crate::errors::{ErrorType, ErrorTypeDefaults, ValError, ValResult, py_err_string};
 use crate::input::{EitherDateTime, Input};
 
 use super::Exactness;
@@ -93,16 +93,16 @@ impl Validator for DateTimeValidator {
             };
             macro_rules! check_constraint {
                 ($constraint:ident, $error:ident) => {
-                    if let Some(constraint) = &constraints.$constraint {
-                        if !speedate_dt.$constraint(constraint) {
-                            return Err(ValError::new(
-                                ErrorType::$error {
-                                    $constraint: constraint.to_string().into(),
-                                    context: None,
-                                },
-                                input,
-                            ));
-                        }
+                    if let Some(constraint) = &constraints.$constraint
+                        && !speedate_dt.$constraint(constraint)
+                    {
+                        return Err(ValError::new(
+                            ErrorType::$error {
+                                $constraint: constraint.to_string().into(),
+                                context: None,
+                            },
+                            input,
+                        ));
                     }
                 };
             }
@@ -247,7 +247,7 @@ impl NowOp {
         match s {
             "past" => Ok(NowOp::Past),
             "future" => Ok(NowOp::Future),
-            _ => py_schema_err!("Invalid now_op {:?}", s),
+            _ => py_schema_err!("Invalid now_op {s:?}"),
         }
     }
 }
@@ -296,17 +296,16 @@ impl TZConstraint {
         match s {
             "naive" => Ok(TZConstraint::Naive),
             "aware" => Ok(TZConstraint::Aware(None)),
-            _ => py_schema_err!("Invalid tz_constraint {:?}", s),
+            _ => py_schema_err!("Invalid tz_constraint {s:?}"),
         }
     }
 
     pub(super) fn from_py(schema: &Bound<'_, PyDict>) -> PyResult<Option<Self>> {
         let py = schema.py();
-        let tz_constraint = match schema.get_item(intern!(py, "tz_constraint"))? {
-            Some(c) => c,
-            None => return Ok(None),
+        let Some(tz_constraint) = schema.get_item(intern!(py, "tz_constraint"))? else {
+            return Ok(None);
         };
-        if let Ok(s) = tz_constraint.downcast::<PyString>() {
+        if let Ok(s) = tz_constraint.cast::<PyString>() {
             let s = s.to_str()?;
             Ok(Some(Self::from_str(s)?))
         } else {
