@@ -1,5 +1,6 @@
 // Validator for Enums, so named because "enum" is a reserved keyword in Rust.
 use std::marker::PhantomData;
+use std::sync::Arc;
 
 use pyo3::exceptions::PyTypeError;
 use pyo3::intern;
@@ -9,10 +10,11 @@ use pyo3::types::{PyDict, PyFloat, PyInt, PyList, PyString, PyType};
 use crate::build_tools::{is_strict, py_schema_err};
 use crate::errors::{ErrorType, ValError, ValResult};
 use crate::input::{Input, InputType};
-use crate::tools::{safe_repr, SchemaDict};
+use crate::tools::{SchemaDict, safe_repr};
+use crate::validators::literal::expected_repr;
 
 use super::is_instance::class_repr;
-use super::literal::{expected_repr_name, LiteralLookup};
+use super::literal::LiteralLookup;
 use super::{BuildValidator, CombinedValidator, DefinitionsBuilder, Exactness, ValidationState, Validator};
 
 #[derive(Debug, Clone)]
@@ -24,8 +26,8 @@ impl BuildValidator for BuildEnumValidator {
     fn build(
         schema: &Bound<'_, PyDict>,
         config: Option<&Bound<'_, PyDict>>,
-        _definitions: &mut DefinitionsBuilder<CombinedValidator>,
-    ) -> PyResult<CombinedValidator> {
+        _definitions: &mut DefinitionsBuilder<Arc<CombinedValidator>>,
+    ) -> PyResult<Arc<CombinedValidator>> {
         let members: Bound<PyList> = schema.get_as_req(intern!(schema.py(), "members"))?;
         if members.is_empty() {
             return py_schema_err!("`members` should have length > 0");
@@ -33,15 +35,13 @@ impl BuildValidator for BuildEnumValidator {
 
         let py = schema.py();
         let value_str = intern!(py, "value");
-        let expected: Vec<(Bound<'_, PyAny>, PyObject)> = members
+        let expected: Vec<(Bound<'_, PyAny>, Py<PyAny>)> = members
             .iter()
             .map(|v| Ok((v.getattr(value_str)?, v.into())))
             .collect::<PyResult<_>>()?;
 
-        let repr_args: Vec<String> = expected
-            .iter()
-            .map(|(k, _)| k.repr()?.extract())
-            .collect::<PyResult<_>>()?;
+        let repr_args: Vec<_> = expected.iter().map(|(k, _)| k.repr()).collect::<PyResult<_>>()?;
+        let expected_repr = expected_repr(&repr_args)?;
 
         let class: Bound<PyType> = schema.get_as_req(intern!(py, "cls"))?;
         let class_repr = class_repr(schema, &class)?;
@@ -55,7 +55,7 @@ impl BuildValidator for BuildEnumValidator {
                     class: class.clone().into(),
                     lookup,
                     missing: schema.get_as(intern!(py, "missing"))?,
-                    expected_repr: expected_repr_name(repr_args, "").0,
+                    expected_repr,
                     strict: is_strict(schema, config)?,
                     class_repr: class_repr.clone(),
                     name: format!("{}[{class_repr}]", $name_prefix),
@@ -65,11 +65,11 @@ impl BuildValidator for BuildEnumValidator {
 
         let sub_type: Option<String> = schema.get_as(intern!(py, "sub_type"))?;
         match sub_type.as_deref() {
-            Some("int") => Ok(CombinedValidator::IntEnum(build!(IntEnumValidator, "int-enum"))),
-            Some("str") => Ok(CombinedValidator::StrEnum(build!(StrEnumValidator, "str-enum"))),
-            Some("float") => Ok(CombinedValidator::FloatEnum(build!(FloatEnumValidator, "float-enum"))),
+            Some("int") => Ok(CombinedValidator::IntEnum(build!(IntEnumValidator, "int-enum")).into()),
+            Some("str") => Ok(CombinedValidator::StrEnum(build!(StrEnumValidator, "str-enum")).into()),
+            Some("float") => Ok(CombinedValidator::FloatEnum(build!(FloatEnumValidator, "float-enum")).into()),
             Some(_) => py_schema_err!("`sub_type` must be one of: 'int', 'str', 'float' or None"),
-            None => Ok(CombinedValidator::PlainEnum(build!(PlainEnumValidator, "enum"))),
+            None => Ok(CombinedValidator::PlainEnum(build!(PlainEnumValidator, "enum")).into()),
         }
     }
 }
@@ -78,17 +78,17 @@ pub trait EnumValidateValue: std::fmt::Debug + Clone + Send + Sync {
     fn validate_value<'py, I: Input<'py> + ?Sized>(
         py: Python<'py>,
         input: &I,
-        lookup: &LiteralLookup<PyObject>,
+        lookup: &LiteralLookup<Py<PyAny>>,
         strict: bool,
-    ) -> ValResult<Option<PyObject>>;
+    ) -> ValResult<Option<Py<PyAny>>>;
 }
 
 #[derive(Debug, Clone)]
 pub struct EnumValidator<T: EnumValidateValue> {
     phantom: PhantomData<T>,
     class: Py<PyType>,
-    lookup: LiteralLookup<PyObject>,
-    missing: Option<PyObject>,
+    lookup: LiteralLookup<Py<PyAny>>,
+    missing: Option<Py<PyAny>>,
     expected_repr: String,
     strict: bool,
     class_repr: String,
@@ -101,7 +101,7 @@ impl<T: EnumValidateValue> Validator for EnumValidator<T> {
         py: Python<'py>,
         input: &(impl Input<'py> + ?Sized),
         state: &mut ValidationState<'_, 'py>,
-    ) -> ValResult<PyObject> {
+    ) -> ValResult<Py<PyAny>> {
         let class = self.class.bind(py);
         if let Some(exact_py_input) = input.as_python().filter(|any| any.is_exact_instance(class)) {
             return Ok(exact_py_input.clone().unbind());
@@ -174,23 +174,21 @@ impl EnumValidateValue for PlainEnumValidator {
     fn validate_value<'py, I: Input<'py> + ?Sized>(
         py: Python<'py>,
         input: &I,
-        lookup: &LiteralLookup<PyObject>,
+        lookup: &LiteralLookup<Py<PyAny>>,
         strict: bool,
-    ) -> ValResult<Option<PyObject>> {
+    ) -> ValResult<Option<Py<PyAny>>> {
         match lookup.validate(py, input)? {
             Some((_, v)) => Ok(Some(v.clone_ref(py))),
             None => {
-                if !strict {
-                    if let Some(py_input) = input.as_python() {
-                        // necessary for compatibility with 2.6, where str and int subclasses are allowed
-                        if py_input.is_instance_of::<PyString>() {
-                            return Ok(lookup.validate_str(input, false)?.map(|v| v.clone_ref(py)));
-                        } else if py_input.is_instance_of::<PyInt>() {
-                            return Ok(lookup.validate_int(py, input, false)?.map(|v| v.clone_ref(py)));
-                        // necessary for compatibility with 2.6, where float values are allowed for int enums in lax mode
-                        } else if py_input.is_instance_of::<PyFloat>() {
-                            return Ok(lookup.validate_int(py, input, false)?.map(|v| v.clone_ref(py)));
-                        }
+                if !strict && let Some(py_input) = input.as_python() {
+                    // necessary for compatibility with 2.6, where str and int subclasses are allowed
+                    if py_input.is_instance_of::<PyString>() {
+                        return Ok(lookup.validate_str(input, false)?.map(|v| v.clone_ref(py)));
+                    } else if py_input.is_instance_of::<PyInt>() {
+                        return Ok(lookup.validate_int(py, input, false)?.map(|v| v.clone_ref(py)));
+                    // necessary for compatibility with 2.6, where float values are allowed for int enums in lax mode
+                    } else if py_input.is_instance_of::<PyFloat>() {
+                        return Ok(lookup.validate_int(py, input, false)?.map(|v| v.clone_ref(py)));
                     }
                 }
                 Ok(None)
@@ -208,9 +206,9 @@ impl EnumValidateValue for IntEnumValidator {
     fn validate_value<'py, I: Input<'py> + ?Sized>(
         py: Python<'py>,
         input: &I,
-        lookup: &LiteralLookup<PyObject>,
+        lookup: &LiteralLookup<Py<PyAny>>,
         strict: bool,
-    ) -> ValResult<Option<PyObject>> {
+    ) -> ValResult<Option<Py<PyAny>>> {
         Ok(lookup.validate_int(py, input, strict)?.map(|v| v.clone_ref(py)))
     }
 }
@@ -224,9 +222,9 @@ impl EnumValidateValue for StrEnumValidator {
     fn validate_value<'py, I: Input<'py> + ?Sized>(
         py: Python,
         input: &I,
-        lookup: &LiteralLookup<PyObject>,
+        lookup: &LiteralLookup<Py<PyAny>>,
         strict: bool,
-    ) -> ValResult<Option<PyObject>> {
+    ) -> ValResult<Option<Py<PyAny>>> {
         Ok(lookup.validate_str(input, strict)?.map(|v| v.clone_ref(py)))
     }
 }
@@ -240,9 +238,9 @@ impl EnumValidateValue for FloatEnumValidator {
     fn validate_value<'py, I: Input<'py> + ?Sized>(
         py: Python<'py>,
         input: &I,
-        lookup: &LiteralLookup<PyObject>,
+        lookup: &LiteralLookup<Py<PyAny>>,
         strict: bool,
-    ) -> ValResult<Option<PyObject>> {
+    ) -> ValResult<Option<Py<PyAny>>> {
         Ok(lookup.validate_float(py, input, strict)?.map(|v| v.clone_ref(py)))
     }
 }
