@@ -2,22 +2,23 @@ use std::sync::Arc;
 
 use pyo3::exceptions::{PyAssertionError, PyValueError};
 use pyo3::prelude::*;
+use pyo3::pybacked::PyBackedStr;
 use pyo3::types::{PyAny, PyDict, PyString};
-use pyo3::{intern, PyTraverseError, PyVisit};
+use pyo3::{PyTraverseError, PyVisit, intern};
 
+use crate::PydanticUseDefault;
 use crate::errors::{
     ErrorType, PydanticCustomError, PydanticKnownError, PydanticOmit, ToErrorValue, ValError, ValResult,
     ValidationError,
 };
 use crate::input::Input;
 use crate::py_gc::PyGcTraverse;
-use crate::tools::{function_name, safe_repr, SchemaDict};
-use crate::PydanticUseDefault;
+use crate::tools::{SchemaDict, function_name, safe_repr};
 
 use super::generator::InternalValidator;
 use super::{
-    build_validator, BuildValidator, CombinedValidator, DefinitionsBuilder, Extra, InputType, ValidationState,
-    Validator,
+    BuildValidator, CombinedValidator, DefinitionsBuilder, Extra, InputType, ValidationState, Validator,
+    build_validator,
 };
 
 struct FunctionInfo {
@@ -103,9 +104,8 @@ impl FunctionBeforeValidator {
     ) -> ValResult<Py<PyAny>> {
         let r = if self.info_arg {
             let field_name = state
-                .extra()
-                .field_name
-                .clone()
+                .field_name()
+                .cloned()
                 .map(Bound::unbind)
                 .or_else(|| self.field_name.clone());
             let info = ValidationInfo::new(py, state.extra(), &self.config, field_name);
@@ -139,7 +139,7 @@ impl Validator for FunctionBeforeValidator {
         &self,
         py: Python<'py>,
         obj: &Bound<'py, PyAny>,
-        field_name: &str,
+        field_name: &PyBackedStr,
         field_value: &Bound<'py, PyAny>,
         state: &mut ValidationState<'_, 'py>,
     ) -> ValResult<Py<PyAny>> {
@@ -178,9 +178,8 @@ impl FunctionAfterValidator {
         let v = call(input, state)?;
         let r = if self.info_arg {
             let field_name = state
-                .extra()
-                .field_name
-                .clone()
+                .field_name()
+                .cloned()
                 .map(Bound::unbind)
                 .or_else(|| self.field_name.clone());
             let info = ValidationInfo::new(py, state.extra(), &self.config, field_name);
@@ -213,7 +212,7 @@ impl Validator for FunctionAfterValidator {
         &self,
         py: Python<'py>,
         obj: &Bound<'py, PyAny>,
-        field_name: &str,
+        field_name: &PyBackedStr,
         field_value: &Bound<'py, PyAny>,
         state: &mut ValidationState<'_, 'py>,
     ) -> ValResult<Py<PyAny>> {
@@ -273,9 +272,8 @@ impl Validator for FunctionPlainValidator {
     ) -> ValResult<Py<PyAny>> {
         let r = if self.info_arg {
             let field_name = state
-                .extra()
-                .field_name
-                .clone()
+                .field_name()
+                .cloned()
                 .map(Bound::unbind)
                 .or_else(|| self.field_name.clone());
             let info = ValidationInfo::new(py, state.extra(), &self.config, field_name);
@@ -343,9 +341,8 @@ impl FunctionWrapValidator {
     ) -> ValResult<Py<PyAny>> {
         let r = if self.info_arg {
             let field_name = state
-                .extra()
-                .field_name
-                .clone()
+                .field_name()
+                .cloned()
                 .map(Bound::unbind)
                 .or_else(|| self.field_name.clone());
             let info = ValidationInfo::new(py, state.extra(), &self.config, field_name);
@@ -392,7 +389,7 @@ impl Validator for FunctionWrapValidator {
         &self,
         py: Python<'py>,
         obj: &Bound<'py, PyAny>,
-        field_name: &str,
+        field_name: &PyBackedStr,
         field_value: &Bound<'py, PyAny>,
         state: &mut ValidationState<'_, 'py>,
     ) -> ValResult<Py<PyAny>> {
@@ -404,7 +401,7 @@ impl Validator for FunctionWrapValidator {
                 self.hide_input_in_errors,
                 self.validation_error_cause,
             ),
-            updated_field_name: field_name.to_string(),
+            updated_field_name: field_name.clone(),
             updated_field_value: field_value.clone().into(),
         };
         #[allow(clippy::used_underscore_items)]
@@ -451,7 +448,7 @@ impl ValidatorCallable {
 #[pyclass(module = "pydantic_core._pydantic_core")]
 #[derive(Debug)]
 struct AssignmentValidatorCallable {
-    updated_field_name: String,
+    updated_field_name: PyBackedStr,
     updated_field_value: Py<PyAny>,
     validator: InternalValidator,
 }
@@ -469,7 +466,7 @@ impl AssignmentValidatorCallable {
         self.validator.validate_assignment(
             py,
             input_value,
-            self.updated_field_name.as_str(),
+            &self.updated_field_name,
             self.updated_field_value.bind(py),
             outer_location,
         )
@@ -507,12 +504,12 @@ macro_rules! py_err_string {
 pub fn convert_err(py: Python<'_>, err: PyErr, input: impl ToErrorValue) -> ValError {
     if err.is_instance_of::<PyValueError>(py) {
         let error_value = err.value(py);
-        if let Ok(pydantic_value_error) = error_value.extract::<PydanticCustomError>() {
-            pydantic_value_error.into_val_error(input)
-        } else if let Ok(pydantic_error_type) = error_value.extract::<PydanticKnownError>() {
-            pydantic_error_type.into_val_error(input)
-        } else if let Ok(validation_error) = err.value(py).extract::<ValidationError>() {
-            validation_error.into_val_error()
+        if let Ok(pydantic_value_error) = error_value.cast::<PydanticCustomError>() {
+            pydantic_value_error.get().clone().into_val_error(input)
+        } else if let Ok(pydantic_error_type) = error_value.cast::<PydanticKnownError>() {
+            pydantic_error_type.get().clone().into_val_error(input)
+        } else if let Ok(validation_error) = err.value(py).cast::<ValidationError>() {
+            validation_error.get().clone().into_val_error()
         } else {
             py_err_string!(py, err, error_value, ValueError, input)
         }
@@ -536,8 +533,15 @@ pub struct ValidationInfo {
     mode: InputType,
 }
 
+impl_py_gc_traverse!(ValidationInfo {
+    config,
+    context,
+    data,
+    field_name
+});
+
 impl ValidationInfo {
-    fn new(py: Python, extra: &Extra, config: &Py<PyAny>, field_name: Option<Py<PyString>>) -> Self {
+    fn new(py: Python, extra: &Extra<'_, '_>, config: &Py<PyAny>, field_name: Option<Py<PyString>>) -> Self {
         Self {
             config: config.clone_ref(py),
             context: extra.context.map(|ctx| ctx.clone().into()),
@@ -548,11 +552,7 @@ impl ValidationInfo {
     }
 
     fn __traverse__(&self, visit: PyVisit<'_>) -> Result<(), PyTraverseError> {
-        visit.call(&self.config)?;
-        if let Some(context) = &self.context {
-            visit.call(context)?;
-        }
-        Ok(())
+        self.py_gc_traverse(&visit)
     }
 
     fn __clear__(&mut self) {
