@@ -1,3 +1,5 @@
+use std::sync::Arc;
+
 use pyo3::intern;
 use pyo3::prelude::*;
 use pyo3::types::{PyDict, PyType};
@@ -21,8 +23,8 @@ impl BuildValidator for IsSubclassValidator {
     fn build(
         schema: &Bound<'_, PyDict>,
         _config: Option<&Bound<'_, PyDict>>,
-        _definitions: &mut DefinitionsBuilder<CombinedValidator>,
-    ) -> PyResult<CombinedValidator> {
+        _definitions: &mut DefinitionsBuilder<Arc<CombinedValidator>>,
+    ) -> PyResult<Arc<CombinedValidator>> {
         let py = schema.py();
         let class = schema.get_as_req::<Bound<'_, PyType>>(intern!(py, "cls"))?;
 
@@ -31,11 +33,11 @@ impl BuildValidator for IsSubclassValidator {
             None => class.qualname()?.to_string(),
         };
         let name = format!("{}[{class_repr}]", Self::EXPECTED_TYPE);
-        Ok(Self {
+        Ok(CombinedValidator::IsSubclass(Self {
             class: class.into(),
             class_repr,
             name,
-        }
+        })
         .into())
     }
 }
@@ -48,7 +50,7 @@ impl Validator for IsSubclassValidator {
         py: Python<'py>,
         input: &(impl Input<'py> + ?Sized),
         _state: &mut ValidationState<'_, 'py>,
-    ) -> ValResult<PyObject> {
+    ) -> ValResult<Py<PyAny>> {
         let Some(obj) = input.as_python() else {
             let method_name = "issubclass".to_string();
             return Err(ValError::new(
@@ -59,7 +61,7 @@ impl Validator for IsSubclassValidator {
                 input,
             ));
         };
-        match obj.downcast::<PyType>() {
+        match obj.cast::<PyType>() {
             Ok(py_type) if py_type.is_subclass(self.class.bind(py))? => Ok(obj.clone().unbind()),
             _ => Err(ValError::new(
                 ErrorType::IsSubclassOf {
