@@ -6,11 +6,12 @@ use enum_dispatch::enum_dispatch;
 use jiter::{PartialMode, StringCacheMode};
 
 use pyo3::exceptions::{PyTypeError, PyValueError};
+use pyo3::pybacked::PyBackedStr;
 use pyo3::types::{PyAny, PyDict, PyString, PyTuple, PyType};
-use pyo3::{intern, PyTraverseError, PyVisit};
-use pyo3::{prelude::*, IntoPyObjectExt};
+use pyo3::{IntoPyObjectExt, prelude::*};
+use pyo3::{PyTraverseError, PyVisit, intern};
 
-use crate::build_tools::{py_schema_err, py_schema_error_type, ExtraBehavior};
+use crate::build_tools::{ExtraBehavior, py_schema_error_type};
 use crate::definitions::{Definitions, DefinitionsBuilder};
 use crate::errors::{LocItem, ValError, ValResult, ValidationError};
 use crate::input::{Input, InputType, StringMapping};
@@ -56,6 +57,7 @@ mod none;
 mod nullable;
 mod prebuilt;
 mod set;
+mod shared;
 mod string;
 mod time;
 mod timedelta;
@@ -131,11 +133,18 @@ impl_py_gc_traverse!(SchemaValidator {
 #[pymethods]
 impl SchemaValidator {
     #[new]
-    #[pyo3(signature = (schema, config=None))]
-    pub fn py_new(py: Python, schema: &Bound<'_, PyAny>, config: Option<&Bound<'_, PyDict>>) -> PyResult<Self> {
-        let mut definitions_builder = DefinitionsBuilder::new();
+    #[pyo3(signature = (schema, config=None, _use_prebuilt=true))]
+    pub fn py_new(
+        py: Python,
+        schema: &Bound<'_, PyAny>,
+        config: Option<&Bound<'_, PyDict>>,
+        _use_prebuilt: bool,
+    ) -> PyResult<Self> {
+        // _use_prebuilt=true by default, but false during rebuilds to avoid stale references
+        // to old validators (see pydantic-core issue #1894)
+        let mut definitions_builder = DefinitionsBuilder::new(_use_prebuilt);
 
-        let validator = build_validator_base(schema, config, &mut definitions_builder)?;
+        let validator = build_validator(schema, config, &mut definitions_builder)?;
         let definitions = definitions_builder.finish()?;
         let py_schema = schema.clone().unbind();
         let py_config = match config {
@@ -324,7 +333,7 @@ impl SchemaValidator {
         &self,
         py: Python,
         obj: Bound<'_, PyAny>,
-        field_name: &str,
+        field_name: PyBackedStr,
         field_value: Bound<'_, PyAny>,
         strict: Option<bool>,
         extra: Option<&Bound<'_, PyString>>,
@@ -343,7 +352,6 @@ impl SchemaValidator {
             strict,
             extra_behavior,
             from_attributes,
-            field_name: Some(PyString::new(py, field_name)),
             context,
             self_instance: None,
             cache_str: self.cache_str,
@@ -352,9 +360,14 @@ impl SchemaValidator {
         };
 
         let guard = &mut RecursionState::default();
-        let mut state = ValidationState::new(extra, guard, false.into());
+        let mut state = ValidationState::new(
+            extra,
+            guard,
+            false.into(),
+            Some(field_name.as_py_str().bind(py).clone()),
+        );
         self.validator
-            .validate_assignment(py, &obj, field_name, &field_value, &mut state)
+            .validate_assignment(py, &obj, &field_name, &field_value, &mut state)
             .map_err(|e| self.prepare_validation_err(py, e, InputType::Python))
     }
 
@@ -371,7 +384,6 @@ impl SchemaValidator {
             strict,
             extra_behavior: None,
             from_attributes: None,
-            field_name: None,
             context,
             self_instance: None,
             cache_str: self.cache_str,
@@ -379,7 +391,7 @@ impl SchemaValidator {
             by_name: None,
         };
         let recursion_guard = &mut RecursionState::default();
-        let mut state = ValidationState::new(extra, recursion_guard, false.into());
+        let mut state = ValidationState::new(extra, recursion_guard, false.into(), None);
         let r = self.validator.default_value(py, None::<i64>, &mut state);
         match r {
             Ok(maybe_default) => match maybe_default {
@@ -391,7 +403,8 @@ impl SchemaValidator {
     }
 
     pub fn __reduce__<'py>(slf: &Bound<'py, Self>) -> PyResult<(Bound<'py, PyType>, Bound<'py, PyTuple>)> {
-        let init_args = (&slf.get().py_schema, &slf.get().py_config).into_pyobject(slf.py())?;
+        // Passing _use_prebuilt=false avoids reusing prebuilt serializers when unpickling
+        let init_args = (&slf.get().py_schema, &slf.get().py_config, false).into_pyobject(slf.py())?;
         Ok((slf.get_type(), init_args))
     }
 
@@ -445,6 +458,7 @@ impl SchemaValidator {
             ),
             &mut recursion_guard,
             allow_partial,
+            None,
         );
         self.validator.validate(py, input, &mut state)
     }
@@ -506,46 +520,15 @@ pub trait BuildValidator: Sized {
     ) -> PyResult<Arc<CombinedValidator>>;
 }
 
-/// Logic to create a particular validator, called in the `validator_match` macro, then in turn by `build_validator`
-fn build_specific_validator<T: BuildValidator>(
-    val_type: &str,
-    schema_dict: &Bound<'_, PyDict>,
-    config: Option<&Bound<'_, PyDict>>,
-    definitions: &mut DefinitionsBuilder<Arc<CombinedValidator>>,
-) -> PyResult<Arc<CombinedValidator>> {
-    T::build(schema_dict, config, definitions)
-        .map_err(|err| py_schema_error_type!("Error building \"{}\" validator:\n  {}", val_type, err))
-}
-
-// macro to build the match statement for validator selection
-macro_rules! validator_match {
-    ($type:ident, $dict:ident, $config:ident, $definitions:ident, $($validator:path,)+) => {
-        match $type {
-            $(
-                <$validator>::EXPECTED_TYPE => build_specific_validator::<$validator>($type, $dict, $config, $definitions),
-            )+
-            "invalid" => return py_schema_err!("Cannot construct schema with `InvalidSchema` member."),
-            _ => return py_schema_err!(r#"Unknown schema type: "{}""#, $type),
-        }
-    };
-}
-
-// Used when creating the base validator instance, to avoid reusing the instance
-// when unpickling:
-pub fn build_validator_base(
-    schema: &Bound<'_, PyAny>,
-    config: Option<&Bound<'_, PyDict>>,
-    definitions: &mut DefinitionsBuilder<Arc<CombinedValidator>>,
-) -> PyResult<Arc<CombinedValidator>> {
-    build_validator_inner(schema, config, definitions, false)
-}
-
 pub fn build_validator(
     schema: &Bound<'_, PyAny>,
     config: Option<&Bound<'_, PyDict>>,
     definitions: &mut DefinitionsBuilder<Arc<CombinedValidator>>,
 ) -> PyResult<Arc<CombinedValidator>> {
-    build_validator_inner(schema, config, definitions, true)
+    // Read use_prebuilt from the definitions builder - this ensures all nested
+    // validators respect the same setting as the top-level build
+    let use_prebuilt = definitions.use_prebuilt();
+    build_validator_inner(schema, config, definitions, use_prebuilt)
 }
 
 fn build_validator_inner(
@@ -554,19 +537,31 @@ fn build_validator_inner(
     definitions: &mut DefinitionsBuilder<Arc<CombinedValidator>>,
     use_prebuilt: bool,
 ) -> PyResult<Arc<CombinedValidator>> {
-    let dict = schema.downcast::<PyDict>()?;
+    let dict = schema.cast::<PyDict>()?;
     let py = schema.py();
     let type_: Bound<'_, PyString> = dict.get_as_req(intern!(py, "type"))?;
     let type_ = type_.to_str()?;
 
-    if use_prebuilt {
-        // if we have a SchemaValidator on the type already, use it
-        if let Ok(Some(prebuilt_validator)) = prebuilt::PrebuiltValidator::try_get_from_schema(type_, dict) {
-            return Ok(Arc::new(prebuilt_validator));
-        }
+    // if we have a SchemaValidator on the type already, use it
+    if use_prebuilt && let Ok(Some(prebuilt_validator)) = prebuilt::PrebuiltValidator::try_get_from_schema(type_, dict)
+    {
+        return Ok(Arc::new(prebuilt_validator));
     }
 
-    validator_match!(
+    // macro to build the match statement for validator selection
+    macro_rules! validator_match {
+        ($type:ident, $dict:ident, $config:ident, $definitions:ident, $($validator:path,)+) => {
+            match $type {
+                $(
+                    <$validator>::EXPECTED_TYPE => <$validator>::build($dict, $config, $definitions),
+                )+
+                "invalid" => return Err(invalid_schema_type()),
+                _ => return Err(unknown_schema_type($type)),
+            }
+        };
+    }
+
+    let result = validator_match!(
         type_,
         dict,
         config,
@@ -661,7 +656,24 @@ fn build_validator_inner(
         definitions::DefinitionRefValidator,
         definitions::DefinitionsValidatorBuilder,
         complex::ComplexValidator,
-    )
+    );
+
+    result.map_err(|e| failed_to_build_validator(type_, e))
+}
+
+#[cold]
+fn failed_to_build_validator(val_type: &str, err: PyErr) -> PyErr {
+    py_schema_error_type!("Error building \"{val_type}\" validator:\n  {err}")
+}
+
+#[cold]
+fn invalid_schema_type() -> PyErr {
+    py_schema_error_type!("Cannot construct schema with `InvalidSchema` member.")
+}
+
+#[cold]
+fn unknown_schema_type(val_type: &str) -> PyErr {
+    py_schema_error_type!("Unknown schema type: \"{val_type}\"")
 }
 
 /// More (mostly immutable) data to pass between validators, should probably be class `Context`,
@@ -681,8 +693,6 @@ pub struct Extra<'a, 'py> {
     pub from_attributes: Option<bool>,
     /// context used in validator functions
     pub context: Option<&'a Bound<'py, PyAny>>,
-    /// The name of the field being validated, if applicable
-    pub field_name: Option<Bound<'py, PyString>>,
     /// This is an instance of the model or dataclass being validated, when validation is performed from `__init__`
     self_instance: Option<&'a Bound<'py, PyAny>>,
     /// Whether to use a cache of short strings to accelerate python string construction
@@ -712,7 +722,6 @@ impl<'a, 'py> Extra<'a, 'py> {
             strict,
             extra_behavior,
             from_attributes,
-            field_name: None,
             context,
             self_instance,
             cache_str,
@@ -730,7 +739,6 @@ impl Extra<'_, '_> {
             strict: Some(true),
             extra_behavior: self.extra_behavior,
             from_attributes: self.from_attributes,
-            field_name: self.field_name.clone(),
             context: self.context,
             self_instance: self.self_instance,
             cache_str: self.cache_str,
@@ -871,7 +879,7 @@ pub trait Validator: Send + Sync + Debug {
         &self,
         _py: Python<'py>,
         _obj: &Bound<'py, PyAny>,
-        _field_name: &str,
+        _field_name: &PyBackedStr,
         _field_value: &Bound<'py, PyAny>,
         _state: &mut ValidationState<'_, 'py>,
     ) -> ValResult<Py<PyAny>> {
