@@ -2,23 +2,23 @@ use std::ptr::null_mut;
 use std::sync::Arc;
 
 use pyo3::exceptions::PyTypeError;
+use pyo3::pybacked::PyBackedStr;
 use pyo3::types::{PyDict, PySet, PyString, PyTuple, PyType};
-use pyo3::{ffi, BoundObject, IntoPyObjectExt};
+use pyo3::{BoundObject, IntoPyObjectExt, ffi};
 use pyo3::{intern, prelude::*};
 
 use super::function::convert_err;
 use super::validation_state::Exactness;
 use super::{
-    build_validator, BuildValidator, CombinedValidator, DefinitionsBuilder, Extra, ValidationState, Validator,
+    BuildValidator, CombinedValidator, DefinitionsBuilder, Extra, ValidationState, Validator, build_validator,
 };
+use crate::PydanticUndefinedType;
 use crate::build_tools::py_schema_err;
 use crate::build_tools::schema_or_config_same;
 use crate::errors::{ErrorType, ErrorTypeDefaults, ValError, ValResult};
-use crate::input::{input_as_python_instance, py_error_on_minusone, Input};
-use crate::tools::{py_err, SchemaDict};
-use crate::PydanticUndefinedType;
+use crate::input::{Input, input_as_python_instance, py_error_on_minusone};
+use crate::tools::{ROOT_FIELD, SchemaDict, py_err, root_field_py_str};
 
-const ROOT_FIELD: &str = "root";
 const DUNDER_DICT: &str = "__dict__";
 const DUNDER_FIELDS_SET_KEY: &str = "__pydantic_fields_set__";
 const DUNDER_MODEL_EXTRA_KEY: &str = "__pydantic_extra__";
@@ -34,11 +34,10 @@ pub(super) enum Revalidate {
 impl Revalidate {
     pub fn from_str(s: Option<&str>) -> PyResult<Self> {
         match s {
-            None => Ok(Self::Never),
             Some("always") => Ok(Self::Always),
-            Some("never") => Ok(Self::Never),
+            Some("never") | None => Ok(Self::Never),
             Some("subclass-instances") => Ok(Self::SubclassInstances),
-            Some(s) => py_schema_err!("Invalid revalidate_instances value: {}", s),
+            Some(s) => py_schema_err!("Invalid revalidate_instances value: {s}"),
         }
     }
 
@@ -101,7 +100,7 @@ impl BuildValidator for ModelValidator {
             frozen: schema.get_as(intern!(py, "frozen"))?.unwrap_or(false),
             custom_init: schema.get_as(intern!(py, "custom_init"))?.unwrap_or(false),
             root_model: schema.get_as(intern!(py, "root_model"))?.unwrap_or(false),
-            undefined: PydanticUndefinedType::new(py).into_any(),
+            undefined: PydanticUndefinedType::get(py).clone_ref(schema.py()).into_any(),
             // Get the class's `__name__`, not using `class.qualname()`
             name,
         })
@@ -156,7 +155,7 @@ impl Validator for ModelValidator {
             if self.revalidate.should_revalidate(py_input, class) || force_revalidate {
                 let fields_set = py_input.getattr(intern!(py, DUNDER_FIELDS_SET_KEY))?;
                 if self.root_model {
-                    let inner_input = py_input.getattr(intern!(py, ROOT_FIELD))?;
+                    let inner_input = py_input.getattr(root_field_py_str(py))?;
                     self.validate_construct(py, &inner_input, Some(&fields_set), state)
                 } else {
                     // get dict here so from_attributes logic doesn't apply
@@ -166,8 +165,8 @@ impl Validator for ModelValidator {
                     let inner_input = if PyAnyMethods::is_none(&model_extra) {
                         dict
                     } else {
-                        let full_model_dict = dict.downcast::<PyDict>()?.copy()?;
-                        full_model_dict.update(model_extra.downcast()?)?;
+                        let full_model_dict = dict.cast::<PyDict>()?.copy()?;
+                        full_model_dict.update(model_extra.cast()?)?;
                         full_model_dict.into_any()
                     };
                     self.validate_construct(py, &inner_input, Some(&fields_set), state)
@@ -186,34 +185,34 @@ impl Validator for ModelValidator {
         &self,
         py: Python<'py>,
         model: &Bound<'py, PyAny>,
-        field_name: &str,
+        field_name: &PyBackedStr,
         field_value: &Bound<'py, PyAny>,
         state: &mut ValidationState<'_, 'py>,
     ) -> ValResult<Py<PyAny>> {
         if self.frozen {
             return Err(ValError::new(ErrorTypeDefaults::FrozenInstance, field_value));
         } else if self.root_model {
-            return if field_name != ROOT_FIELD {
-                Err(ValError::new_with_loc(
+            if field_name != ROOT_FIELD {
+                return Err(ValError::new_with_loc(
                     ErrorType::NoSuchAttribute {
                         attribute: field_name.to_string(),
                         context: None,
                     },
                     field_value,
                     field_name.to_string(),
-                ))
-            } else {
-                let state = &mut state.rebind_extra(|extra| extra.field_name = Some(PyString::new(py, ROOT_FIELD)));
-                let output = self.validator.validate(py, field_value, state)?;
+                ));
+            }
+            let root_field = root_field_py_str(py);
+            let state = &mut state.scoped_set_field_name(Some(root_field.clone()));
+            let output = self.validator.validate(py, field_value, state)?;
 
-                force_setattr(py, model, intern!(py, ROOT_FIELD), output)?;
-                Ok(model.into_py_any(py)?)
-            };
+            force_setattr(py, model, root_field, output)?;
+            return Ok(model.into_py_any(py)?);
         }
-        let old_dict = model.getattr(intern!(py, DUNDER_DICT))?.downcast_into::<PyDict>()?;
+        let old_dict = model.getattr(intern!(py, DUNDER_DICT))?.cast_into::<PyDict>()?;
 
         let input_dict = old_dict.copy()?;
-        if let Ok(old_extra) = model.getattr(intern!(py, DUNDER_MODEL_EXTRA_KEY))?.downcast::<PyDict>() {
+        if let Ok(old_extra) = model.getattr(intern!(py, DUNDER_MODEL_EXTRA_KEY))?.cast::<PyDict>() {
             input_dict.update(old_extra.as_mapping())?;
         }
         input_dict.set_item(field_name, field_value)?;
@@ -229,7 +228,7 @@ impl Validator for ModelValidator {
         ) = output.extract(py)?;
 
         if let Ok(fields_set) = model.getattr(intern!(py, DUNDER_FIELDS_SET_KEY)) {
-            let fields_set = fields_set.downcast::<PySet>()?;
+            let fields_set = fields_set.cast::<PySet>()?;
             for field_name in validated_fields_set {
                 fields_set.add(field_name)?;
             }
@@ -259,16 +258,17 @@ impl ModelValidator {
         let state = &mut state.rebind_extra(|extra| extra.self_instance = None);
 
         if self.root_model {
-            let state = &mut state.rebind_extra(|extra| extra.field_name = Some(PyString::new(py, ROOT_FIELD)));
+            let root_field = root_field_py_str(py);
+            let state = &mut state.scoped_set_field_name(Some(root_field.clone()));
             let output = self.validator.validate(py, input, state)?;
 
             let fields_set = if input.as_python().is_some_and(|py_input| py_input.is(&self.undefined)) {
                 PySet::empty(py)?
             } else {
-                PySet::new(py, [&String::from(ROOT_FIELD)])?
+                PySet::new(py, [root_field])?
             };
             force_setattr(py, self_instance, intern!(py, DUNDER_FIELDS_SET_KEY), &fields_set)?;
-            force_setattr(py, self_instance, intern!(py, ROOT_FIELD), &output)?;
+            force_setattr(py, self_instance, root_field, &output)?;
         } else {
             let output = self.validator.validate(py, input, state)?;
 
@@ -303,17 +303,18 @@ impl ModelValidator {
         let instance;
 
         if self.root_model {
-            let state = &mut state.rebind_extra(|extra| extra.field_name = Some(PyString::new(py, ROOT_FIELD)));
+            let root_field = root_field_py_str(py);
+            let state = &mut state.scoped_set_field_name(Some(root_field.clone()));
             let output = self.validator.validate(py, input, state)?;
             instance = create_class(self.class.bind(py))?;
 
             let fields_set = if input.as_python().is_some_and(|py_input| py_input.is(&self.undefined)) {
                 PySet::empty(py)?
             } else {
-                PySet::new(py, [&String::from(ROOT_FIELD)])?
+                PySet::new(py, [root_field])?
             };
             force_setattr(py, &instance, intern!(py, DUNDER_FIELDS_SET_KEY), &fields_set)?;
-            force_setattr(py, &instance, intern!(py, ROOT_FIELD), output)?;
+            force_setattr(py, &instance, root_field, output)?;
         } else {
             let output = self.validator.validate(py, input, state)?;
             instance = create_class(self.class.bind(py))?;
@@ -331,7 +332,7 @@ impl ModelValidator {
         py: Python<'py>,
         instance: Bound<'_, PyAny>,
         input: &(impl Input<'py> + ?Sized),
-        extra: &Extra,
+        extra: &Extra<'_, 'py>,
     ) -> ValResult<Py<PyAny>> {
         if let Some(ref post_init) = self.post_init {
             instance
