@@ -2,28 +2,29 @@ use std::sync::Arc;
 
 use pyo3::intern;
 use pyo3::prelude::*;
-use pyo3::types::{PyDict, PyString, PyType};
+use pyo3::pybacked::PyBackedStr;
+use pyo3::types::{PyDict, PyType};
 
 use crate::build_tools::py_schema_err;
-use crate::build_tools::{is_strict, schema_or_config, ExtraBehavior};
+use crate::build_tools::{ExtraBehavior, is_strict, schema_or_config};
 use crate::errors::LocItem;
 use crate::errors::{ErrorTypeDefaults, ValError, ValLineError, ValResult};
 use crate::input::BorrowInput;
 use crate::input::ConsumeIterator;
 use crate::input::ValidationMatch;
 use crate::input::{Input, ValidatedDict};
-use crate::lookup_key::LookupKeyCollection;
+use crate::lookup_key::LookupPathCollection;
+use crate::lookup_key::LookupType;
 use crate::tools::SchemaDict;
 use ahash::AHashSet;
 use jiter::PartialMode;
 
-use super::{build_validator, BuildValidator, CombinedValidator, DefinitionsBuilder, ValidationState, Validator};
+use super::{BuildValidator, CombinedValidator, DefinitionsBuilder, ValidationState, Validator, build_validator};
 
 #[derive(Debug)]
 struct TypedDictField {
-    name: String,
-    lookup_key_collection: LookupKeyCollection,
-    name_py: Py<PyString>,
+    name: PyBackedStr,
+    lookup_path_collection: LookupPathCollection,
     required: bool,
     validator: Arc<CombinedValidator>,
 }
@@ -81,54 +82,44 @@ impl BuildValidator for TypedDictValidator {
         };
 
         for (key, value) in fields_dict {
-            let field_info = value.downcast::<PyDict>()?;
-            let field_name_py = key.downcast_into::<PyString>()?;
-            let field_name = field_name_py.to_str()?;
+            let field_info = value.cast::<PyDict>()?;
+            let name: PyBackedStr = key.extract()?;
 
             let schema = field_info.get_as_req(intern!(py, "schema"))?;
 
             let validator = match build_validator(&schema, config, definitions) {
                 Ok(v) => v,
-                Err(err) => return py_schema_err!("Field \"{}\":\n  {}", field_name, err),
+                Err(err) => return py_schema_err!("Field \"{name}\":\n  {err}"),
             };
 
             let required = match field_info.get_as::<bool>(intern!(py, "required"))? {
                 Some(required) => {
-                    if required {
-                        if let CombinedValidator::WithDefault(ref val) = validator.as_ref() {
-                            if val.has_default() {
-                                return py_schema_err!(
-                                    "Field '{}': a required field cannot have a default value",
-                                    field_name
-                                );
-                            }
-                        }
+                    if required
+                        && let CombinedValidator::WithDefault(val) = validator.as_ref()
+                        && val.has_default()
+                    {
+                        return py_schema_err!("Field '{name}': a required field cannot have a default value");
                     }
                     required
                 }
                 None => total,
             };
 
-            if required {
-                if let CombinedValidator::WithDefault(ref val) = validator.as_ref() {
-                    if val.omit_on_error() {
-                        return py_schema_err!(
-                            "Field '{}': 'on_error = omit' cannot be set for required fields",
-                            field_name
-                        );
-                    }
-                }
+            if required
+                && let CombinedValidator::WithDefault(val) = validator.as_ref()
+                && val.omit_on_error()
+            {
+                return py_schema_err!("Field '{name}': 'on_error = omit' cannot be set for required fields");
             }
 
-            let validation_alias = field_info.get_item(intern!(py, "validation_alias"))?;
-            let lookup_key_collection = LookupKeyCollection::new(py, validation_alias, field_name)?;
+            let validation_alias = field_info.get_as(intern!(py, "validation_alias"))?;
+            let lookup_path_collection = LookupPathCollection::new(validation_alias, name.clone())?;
 
             fields.push(TypedDictField {
-                name: field_name.to_string(),
-                lookup_key_collection,
-                name_py: field_name_py.into(),
-                validator,
+                name,
+                lookup_path_collection,
                 required,
+                validator,
             });
         }
         Ok(CombinedValidator::TypedDict(Self {
@@ -173,6 +164,7 @@ impl Validator for TypedDictValidator {
 
         let validate_by_alias = state.validate_by_alias_or(self.validate_by_alias);
         let validate_by_name = state.validate_by_name_or(self.validate_by_name);
+        let lookup_type = LookupType::from_bools(validate_by_alias, validate_by_name)?;
 
         // we only care about which keys have been used if we're iterating over the object for extra after
         // the first pass
@@ -185,27 +177,30 @@ impl Validator for TypedDictValidator {
 
         {
             let state = &mut state.rebind_extra(|extra| extra.data = Some(output_dict.clone()));
+            let state = &mut state.scoped_set(|state| &mut state.has_field_error, false);
 
             let mut fields_set_count: usize = 0;
 
             for field in &self.fields {
-                let lookup_key = field
-                    .lookup_key_collection
-                    .select(validate_by_alias, validate_by_name)?;
-                let op_key_value = match dict.get_item(lookup_key) {
-                    Ok(v) => v,
-                    Err(ValError::LineErrors(line_errors)) => {
-                        let field_loc: LocItem = field.name.clone().into();
-                        if partial_last_key.as_ref() == Some(&field_loc) {
-                            for err in line_errors {
-                                errors.push(err.with_outer_location(field_loc.clone()));
+                if let Some((lookup_path, lookup_result)) = field
+                    .lookup_path_collection
+                    .lookup_paths(lookup_type)
+                    .find_map(|path| Some((path, dict.get_item(path).transpose()?)))
+                {
+                    let value = match lookup_result {
+                        Ok(v) => v,
+                        Err(ValError::LineErrors(line_errors)) => {
+                            let field_loc: LocItem = field.name.clone().into();
+                            if partial_last_key.as_ref() == Some(&field_loc) {
+                                for err in line_errors {
+                                    errors.push(err.with_outer_location(field_loc.clone()));
+                                }
                             }
+                            continue;
                         }
-                        continue;
-                    }
-                    Err(err) => return Err(err),
-                };
-                if let Some((lookup_path, value)) = op_key_value {
+                        Err(err) => return Err(err),
+                    };
+
                     if let Some(ref mut used_keys) = used_keys {
                         // key is "used" whether or not validation passes, since we want to skip this key in
                         // extra logic either way
@@ -221,45 +216,53 @@ impl Validator for TypedDictValidator {
                         true => allow_partial,
                         false => false.into(),
                     };
-                    let state =
-                        &mut state.rebind_extra(|extra| extra.field_name = Some(field.name_py.bind(py).clone()));
+
+                    // FIXME: for model and dataclass, `default_value` is called with field name set in extra, does that matter?
+                    let state = &mut state.scoped_set_field_name(Some(field.name.as_py_str().bind(py).clone()));
 
                     match field.validator.validate(py, value.borrow_input(), state) {
                         Ok(value) => {
-                            output_dict.set_item(&field.name_py, value)?;
+                            output_dict.set_item(&field.name, value)?;
                             fields_set_count += 1;
                         }
-                        Err(ValError::Omit) => continue,
-                        Err(ValError::LineErrors(line_errors)) => {
-                            if !is_last_partial || field.required {
-                                for err in line_errors {
-                                    errors.push(lookup_path.apply_error_loc(err, self.loc_by_alias, &field.name));
+                        Err(e) => {
+                            state.has_field_error = true;
+                            match e {
+                                ValError::Omit => {}
+                                ValError::LineErrors(line_errors) => {
+                                    if !is_last_partial || field.required {
+                                        for err in line_errors {
+                                            errors.push(lookup_path.apply_error_loc(
+                                                err,
+                                                self.loc_by_alias,
+                                                &field.name,
+                                            ));
+                                        }
+                                    }
                                 }
+                                err => return Err(err),
                             }
                         }
-                        Err(err) => return Err(err),
                     }
                     continue;
                 }
 
-                match field.validator.default_value(py, Some(field.name.as_str()), state) {
+                match field.validator.default_value(py, Some(field.name.clone()), state) {
                     Ok(Some(value)) => {
                         // Default value exists, and passed validation if required
-                        output_dict.set_item(&field.name_py, value)?;
+                        output_dict.set_item(&field.name, value)?;
                     }
                     Ok(None) => {
                         // This means there was no default value
                         if field.required {
-                            errors.push(lookup_key.error(
-                                ErrorTypeDefaults::Missing,
-                                input,
-                                self.loc_by_alias,
-                                &field.name,
-                            ));
+                            let error_type = ErrorTypeDefaults::Missing;
+                            let error_loc = field.lookup_path_collection.error_loc(lookup_type, self.loc_by_alias);
+                            errors.push(ValLineError::new_with_full_loc(error_type, input, error_loc));
                         }
                     }
                     Err(ValError::Omit) => {}
                     Err(ValError::LineErrors(line_errors)) => {
+                        state.has_field_error = true;
                         for err in line_errors {
                             // Note: this will always use the field name even if there is an alias
                             // However, we don't mind so much because this error can only happen if the
