@@ -1,18 +1,19 @@
 use std::str::from_utf8;
+use std::sync::Arc;
 
 use pyo3::intern;
 use pyo3::prelude::*;
-use pyo3::sync::GILOnceCell;
+use pyo3::sync::PyOnceLock;
 use pyo3::types::{PyDict, PyType};
 use uuid::Uuid;
 use uuid::Variant;
 
 use crate::build_tools::is_strict;
 use crate::errors::{ErrorType, ErrorTypeDefaults, ValError, ValResult};
-use crate::input::input_as_python_instance;
 use crate::input::Input;
 use crate::input::InputType;
 use crate::input::ValidationMatch;
+use crate::input::input_as_python_instance;
 use crate::serializers::BytesMode;
 use crate::tools::SchemaDict;
 
@@ -24,16 +25,10 @@ use super::{BuildValidator, CombinedValidator, DefinitionsBuilder, Exactness, Va
 const UUID_INT: &str = "int";
 const UUID_IS_SAFE: &str = "is_safe";
 
-static UUID_TYPE: GILOnceCell<Py<PyType>> = GILOnceCell::new();
-
-fn import_type(py: Python, module: &str, attr: &str) -> PyResult<Py<PyType>> {
-    py.import(module)?.getattr(attr)?.extract()
-}
+static UUID_TYPE: PyOnceLock<Py<PyType>> = PyOnceLock::new();
 
 fn get_uuid_type(py: Python<'_>) -> PyResult<&Bound<'_, PyType>> {
-    Ok(UUID_TYPE
-        .get_or_init(py, || import_type(py, "uuid", "UUID").unwrap())
-        .bind(py))
+    UUID_TYPE.import(py, "uuid", "UUID")
 }
 
 #[derive(Debug, Clone, Copy)]
@@ -80,15 +75,15 @@ impl BuildValidator for UuidValidator {
     fn build(
         schema: &Bound<'_, PyDict>,
         config: Option<&Bound<'_, PyDict>>,
-        _definitions: &mut DefinitionsBuilder<CombinedValidator>,
-    ) -> PyResult<CombinedValidator> {
+        _definitions: &mut DefinitionsBuilder<Arc<CombinedValidator>>,
+    ) -> PyResult<Arc<CombinedValidator>> {
         let py = schema.py();
         // Note(lig): let's keep this conversion through the Version enum just for the sake of validation
         let version = schema.get_as::<u8>(intern!(py, "version"))?.map(Version::from);
-        Ok(Self {
+        Ok(CombinedValidator::Uuid(Self {
             strict: is_strict(schema, config)?,
             version: version.map(usize::from),
-        }
+        })
         .into())
     }
 }
@@ -101,7 +96,7 @@ impl Validator for UuidValidator {
         py: Python<'py>,
         input: &(impl Input<'py> + ?Sized),
         state: &mut ValidationState<'_, 'py>,
-    ) -> ValResult<PyObject> {
+    ) -> ValResult<Py<PyAny>> {
         let class = get_uuid_type(py)?;
         if let Some(py_input) = input_as_python_instance(input, class) {
             if let Some(expected_version) = self.version {
@@ -142,16 +137,16 @@ impl Validator for UuidValidator {
             // This block checks if the UUID version matches the expected version and
             // if the UUID variant conforms to RFC 9562 (superseding RFC 4122).
             // When dealing with Python inputs, UUIDs must adhere to RFC 9562 standards.
-            if let Some(expected_version) = self.version {
-                if uuid.get_version_num() != expected_version || uuid.get_variant() != Variant::RFC4122 {
-                    return Err(ValError::new(
-                        ErrorType::UuidVersion {
-                            expected_version,
-                            context: None,
-                        },
-                        input,
-                    ));
-                }
+            if let Some(expected_version) = self.version
+                && (uuid.get_version_num() != expected_version || uuid.get_variant() != Variant::RFC4122)
+            {
+                return Err(ValError::new(
+                    ErrorType::UuidVersion {
+                        expected_version,
+                        context: None,
+                    },
+                    input,
+                ));
             }
             self.create_py_uuid(class, &uuid)
         }
@@ -184,13 +179,13 @@ impl UuidValidator {
                     .map_err(|_| ValError::new(ErrorTypeDefaults::UuidType, input))?
                     .into_inner();
                 let bytes_slice = either_bytes.as_slice();
-                'parse: {
-                    // Try parsing as utf8, but don't care if it fails
-                    if let Ok(utf8_str) = from_utf8(bytes_slice) {
-                        if let Ok(uuid) = Uuid::parse_str(utf8_str) {
-                            break 'parse uuid;
-                        }
-                    }
+
+                // Try parsing as utf8, if it fails fall back to bytes
+                if let Ok(utf8_str) = from_utf8(bytes_slice)
+                    && let Ok(uuid) = Uuid::parse_str(utf8_str)
+                {
+                    uuid
+                } else {
                     Uuid::from_slice(bytes_slice).map_err(|e| {
                         ValError::new(
                             ErrorType::UuidParsing {
@@ -204,17 +199,16 @@ impl UuidValidator {
             }
         };
 
-        if let Some(expected_version) = self.version {
-            let v1 = uuid.get_version_num();
-            if v1 != expected_version {
-                return Err(ValError::new(
-                    ErrorType::UuidVersion {
-                        expected_version,
-                        context: None,
-                    },
-                    input,
-                ));
-            }
+        if let Some(expected_version) = self.version
+            && uuid.get_version_num() != expected_version
+        {
+            return Err(ValError::new(
+                ErrorType::UuidVersion {
+                    expected_version,
+                    context: None,
+                },
+                input,
+            ));
         }
         Ok(uuid)
     }
@@ -226,16 +220,19 @@ impl UuidValidator {
     ///
     /// This implementation does not use the Python `__init__` function to speed up the process,
     /// as the `__init__` function in the Python `uuid` module performs extensive checks.
-    fn create_py_uuid(&self, py_type: &Bound<'_, PyType>, uuid: &Uuid) -> ValResult<PyObject> {
+    fn create_py_uuid(&self, py_type: &Bound<'_, PyType>, uuid: &Uuid) -> ValResult<Py<PyAny>> {
+        static UUID_SAFE_UNKNOWN: PyOnceLock<Py<PyAny>> = PyOnceLock::new();
         let py = py_type.py();
-        let dc = create_class(py_type)?;
-        let int = uuid.as_u128();
-        let safe = py
-            .import(intern!(py, "uuid"))?
-            .getattr(intern!(py, "SafeUUID"))?
-            .get_item("safe")?;
-        force_setattr(py, &dc, intern!(py, UUID_INT), int)?;
-        force_setattr(py, &dc, intern!(py, UUID_IS_SAFE), safe)?;
-        Ok(dc.into())
+        let safe_unknown = UUID_SAFE_UNKNOWN.get_or_try_init(py, || {
+            py.import("uuid")?
+                .getattr("SafeUUID")?
+                .get_item("unknown")
+                .map(Bound::unbind)
+        })?;
+
+        let uuid_instance = create_class(py_type)?;
+        force_setattr(py, &uuid_instance, intern!(py, UUID_INT), uuid.as_u128())?;
+        force_setattr(py, &uuid_instance, intern!(py, UUID_IS_SAFE), safe_unknown)?;
+        Ok(uuid_instance.into())
     }
 }
