@@ -1,3 +1,5 @@
+use std::ops::{Deref, DerefMut};
+
 use pyo3::prelude::*;
 use pyo3::types::PyString;
 
@@ -26,17 +28,33 @@ pub struct ValidationState<'a, 'py> {
     pub fields_set_count: Option<usize>,
     // True if `allow_partial=true` and we're validating the last element of a sequence or mapping.
     pub allow_partial: PartialMode,
+    // Whether at least one field had a validation error. This is used in the context of structured types
+    // (models, dataclasses, etc), where we need to know if a validation error occurred before calling
+    // a default factory that takes the validated data.
+    //
+    // TODO: this should probably be moved directly into the structured types which need it, but that
+    // requires some refactoring to make them have knowledge of default (factories).
+    pub has_field_error: bool,
+    /// The name of the field being validated, if applicable
+    field_name: Option<Bound<'py, PyString>>,
     // deliberately make Extra readonly
     extra: Extra<'a, 'py>,
 }
 
 impl<'a, 'py> ValidationState<'a, 'py> {
-    pub fn new(extra: Extra<'a, 'py>, recursion_guard: &'a mut RecursionState, allow_partial: PartialMode) -> Self {
+    pub fn new(
+        extra: Extra<'a, 'py>,
+        recursion_guard: &'a mut RecursionState,
+        allow_partial: PartialMode,
+        field_name: Option<Bound<'py, PyString>>,
+    ) -> Self {
         Self {
             recursion_guard, // Don't care about exactness unless doing union validation
             exactness: None,
             fields_set_count: None,
             allow_partial,
+            has_field_error: false,
+            field_name,
             extra,
         }
     }
@@ -53,11 +71,43 @@ impl<'a, 'py> ValidationState<'a, 'py> {
         ValidationStateWithReboundExtra { state: self, old_extra }
     }
 
+    /// Temporarily rebinds a field of the state by calling `projector` to get a mutable reference to the field,
+    /// and setting that field to `value`.
+    ///
+    /// When `ScopedSetState` drops, the field is restored to its original value.
+    pub fn scoped_set<'state, P, T>(
+        &'state mut self,
+        projector: P,
+        new_value: T,
+    ) -> ScopedSetState<'state, 'a, 'py, P, T>
+    where
+        P: for<'p> Fn(&'p mut ValidationState<'a, 'py>) -> &'p mut T,
+    {
+        let value = std::mem::replace((projector)(self), new_value);
+        ScopedSetState {
+            state: self,
+            projector,
+            value,
+        }
+    }
+
+    /// Set the field name within state for the given scope.
+    pub fn scoped_set_field_name(
+        &mut self,
+        new_value: Option<Bound<'py, PyString>>,
+    ) -> ScopedFieldNameState<'_, 'a, 'py> {
+        self.scoped_set(Self::field_name_mut, new_value)
+    }
+
+    pub fn field_name(&self) -> Option<&Bound<'py, PyString>> {
+        self.field_name.as_ref()
+    }
+
     pub fn extra(&self) -> &'_ Extra<'a, 'py> {
         &self.extra
     }
 
-    pub fn enumerate_last_partial<I>(&self, iter: impl Iterator<Item = I>) -> impl Iterator<Item = (usize, bool, I)> {
+    pub fn enumerate_last_partial<I: Iterator>(&self, iter: I) -> EnumerateLastPartial<I> {
         EnumerateLastPartial::new(iter, self.allow_partial)
     }
 
@@ -105,6 +155,10 @@ impl<'a, 'py> ValidationState<'a, 'py> {
     pub fn maybe_cached_str(&self, py: Python<'py>, s: &str) -> Bound<'py, PyString> {
         new_py_string(py, s, self.extra.cache_str)
     }
+
+    fn field_name_mut(&mut self) -> &mut Option<Bound<'py, PyString>> {
+        &mut self.field_name
+    }
 }
 
 impl ContainsRecursionState for ValidationState<'_, '_> {
@@ -171,3 +225,52 @@ impl<I: Iterator> Iterator for EnumerateLastPartial<I> {
         self.iter.size_hint()
     }
 }
+
+pub struct ScopedSetState<'scope, 'a, 'py, P, T>
+where
+    P: for<'p> Fn(&'p mut ValidationState<'a, 'py>) -> &'p mut T,
+{
+    /// The state which has been set for the scope.
+    state: &'scope mut ValidationState<'a, 'py>,
+    /// A function that projects from the state to the field that has been set.
+    projector: P,
+    /// The previous value of the field that has been set.
+    value: T,
+}
+
+impl<'a, 'py, P, T> Drop for ScopedSetState<'_, 'a, 'py, P, T>
+where
+    P: for<'drop> Fn(&'drop mut ValidationState<'a, 'py>) -> &'drop mut T,
+{
+    fn drop(&mut self) {
+        std::mem::swap((self.projector)(self.state), &mut self.value);
+    }
+}
+
+impl<'a, 'py, P, T> Deref for ScopedSetState<'_, 'a, 'py, P, T>
+where
+    P: for<'p> Fn(&'p mut ValidationState<'a, 'py>) -> &'p mut T,
+{
+    type Target = ValidationState<'a, 'py>;
+
+    fn deref(&self) -> &Self::Target {
+        self.state
+    }
+}
+
+impl<'a, 'py, P, T> DerefMut for ScopedSetState<'_, 'a, 'py, P, T>
+where
+    P: for<'p> Fn(&'p mut ValidationState<'a, 'py>) -> &'p mut T,
+{
+    fn deref_mut(&mut self) -> &mut ValidationState<'a, 'py> {
+        self.state
+    }
+}
+
+type ScopedFieldNameState<'scope, 'a, 'py> = ScopedSetState<
+    'scope,
+    'a,
+    'py,
+    for<'s> fn(&'s mut ValidationState<'a, 'py>) -> &'s mut Option<Bound<'py, PyString>>,
+    Option<Bound<'py, PyString>>,
+>;
