@@ -1,3 +1,4 @@
+import os
 import platform
 import re
 import sys
@@ -47,7 +48,7 @@ def test_args_kwargs():
         [
             ArgsKwargs((1, 'a', True), {'x': 1}),
             Err(
-                '',
+                '1 validation error for arguments',
                 [
                     {
                         'type': 'unexpected_keyword_argument',
@@ -61,7 +62,7 @@ def test_args_kwargs():
         [
             [1],
             Err(
-                '',
+                '2 validation errors for arguments',
                 [
                     {
                         'type': 'missing_positional_only_argument',
@@ -81,7 +82,7 @@ def test_args_kwargs():
         [
             [1, 'a', True, 4],
             Err(
-                '',
+                '1 validation error for arguments',
                 [
                     {
                         'type': 'unexpected_positional_argument',
@@ -95,7 +96,7 @@ def test_args_kwargs():
         [
             [1, 'a', True, 4, 5],
             Err(
-                '',
+                '2 validation errors for arguments',
                 [
                     {
                         'type': 'unexpected_positional_argument',
@@ -115,7 +116,7 @@ def test_args_kwargs():
         [
             ('x', 'a', 'wrong'),
             Err(
-                '',
+                '2 validation errors for arguments',
                 [
                     {
                         'type': 'int_parsing',
@@ -221,7 +222,7 @@ def test_positional_args(py_and_json: PyAndJson, input_value, expected):
         [
             ArgsKwargs((), {'a': 'x', 'b': 'a', 'c': 'wrong'}),
             Err(
-                '',
+                '2 validation errors for arguments',
                 [
                     {
                         'type': 'int_parsing',
@@ -241,7 +242,7 @@ def test_positional_args(py_and_json: PyAndJson, input_value, expected):
         [
             ArgsKwargs(()),
             Err(
-                '',
+                '3 validation errors for arguments',
                 [
                     {
                         'type': 'missing_keyword_only_argument',
@@ -1169,8 +1170,12 @@ def test_error_display(pydantic_version):
         '1 validation error for arguments\n'
         'b\n'
         '  Missing required argument [type=missing_argument, '
-        "input_value=ArgsKwargs((), {'a': 1}), input_type=ArgsKwargs]\n"
-        f'    For further information visit https://errors.pydantic.dev/{pydantic_version}/v/missing_argument'
+        "input_value=ArgsKwargs((), {'a': 1}), input_type=ArgsKwargs]"
+        + (
+            f'\n    For further information visit https://errors.pydantic.dev/{pydantic_version}/v/missing_argument'
+            if os.environ.get('PYDANTIC_ERRORS_INCLUDE_URL') != 'false'
+            else ''
+        )
     )
     # insert_assert(exc_info.value.json(include_url=False))
     assert exc_info.value.json(include_url=False) == (
