@@ -1,8 +1,10 @@
+import os
 import re
 
 import pytest
 
 from pydantic_core import SchemaValidator, ValidationError, core_schema
+from pydantic_core import core_schema as cs
 
 from ..conftest import Err, PyAndJson, plain_repr
 
@@ -54,7 +56,7 @@ def test_bool_strict(py_and_json: PyAndJson):
 
 
 def test_bool_error(pydantic_version):
-    v = SchemaValidator({'type': 'bool'})
+    v = SchemaValidator(cs.bool_schema())
 
     with pytest.raises(ValidationError) as exc_info:
         v.validate_python('wrong')
@@ -62,8 +64,12 @@ def test_bool_error(pydantic_version):
     assert str(exc_info.value) == (
         '1 validation error for bool\n'
         '  Input should be a valid boolean, '
-        "unable to interpret input [type=bool_parsing, input_value='wrong', input_type=str]\n"
-        f"    For further information visit https://errors.pydantic.dev/{pydantic_version}/v/bool_parsing"
+        "unable to interpret input [type=bool_parsing, input_value='wrong', input_type=str]"
+        + (
+            f'\n    For further information visit https://errors.pydantic.dev/{pydantic_version}/v/bool_parsing'
+            if os.environ.get('PYDANTIC_ERRORS_INCLUDE_URL', '1') != 'false'
+            else ''
+        )
     )
     assert exc_info.value.errors(include_url=False) == [
         {
@@ -76,10 +82,16 @@ def test_bool_error(pydantic_version):
 
 
 def test_bool_repr():
-    v = SchemaValidator({'type': 'bool'})
-    assert plain_repr(v) == 'SchemaValidator(title="bool",validator=Bool(BoolValidator{strict:false}),definitions=[])'
-    v = SchemaValidator({'type': 'bool', 'strict': True})
-    assert plain_repr(v) == 'SchemaValidator(title="bool",validator=Bool(BoolValidator{strict:true}),definitions=[])'
+    v = SchemaValidator(cs.bool_schema())
+    assert (
+        plain_repr(v)
+        == 'SchemaValidator(title="bool",validator=Bool(BoolValidator{strict:false}),definitions=[],cache_strings=True)'
+    )
+    v = SchemaValidator(cs.bool_schema(strict=True))
+    assert (
+        plain_repr(v)
+        == 'SchemaValidator(title="bool",validator=Bool(BoolValidator{strict:true}),definitions=[],cache_strings=True)'
+    )
 
 
 def test_bool_key(py_and_json: PyAndJson):
@@ -88,7 +100,8 @@ def test_bool_key(py_and_json: PyAndJson):
     assert v.validate_test({'true': 1, 'off': 2}) == {True: 1, False: 2}
     assert v.validate_test({'true': 1, 'off': 2}, strict=False) == {True: 1, False: 2}
     with pytest.raises(ValidationError, match='Input should be a valid boolean'):
-        v.validate_test({'true': 1, 'off': 2}, strict=True)
+        v.validate_python({'true': 1, 'off': 2}, strict=True)
+    assert v.validate_json('{"true": 1, "off": 2}', strict=True) == {True: 1, False: 2}
 
 
 def test_validate_assignment_not_supported() -> None:
