@@ -9,7 +9,8 @@ from typing import Any
 import pytest
 from dirty_equals import FunctionCheck, IsStr
 
-from pydantic_core import SchemaValidator, ValidationError, core_schema
+from pydantic_core import SchemaError, SchemaValidator, ValidationError
+from pydantic_core import core_schema as cs
 
 from ..conftest import Err, PyAndJson, plain_repr
 
@@ -18,6 +19,23 @@ class DecimalSubclass(Decimal):
     pass
 
 
+# Note: there's another constraint validation (allow_inf_nan=True cannot be used with max_digits or decimal_places).
+# but it is tested in Pydantic:
+@pytest.mark.parametrize(
+    'constraint',
+    ['multiple_of', 'le', 'lt', 'ge', 'gt'],
+)
+def test_constraints_schema_validation_error(constraint: str) -> None:
+    with pytest.raises(SchemaError, match=f"'{constraint}' must be coercible to a Decimal instance"):
+        SchemaValidator(cs.decimal_schema(**{constraint: 'bad_value'}))
+
+
+def test_constraints_schema_validation() -> None:
+    val = SchemaValidator(cs.decimal_schema(gt='1'))
+    with pytest.raises(ValidationError):
+        val.validate_python('0')
+
+
 @pytest.mark.parametrize(
     'input_value,expected',
     [
@@ -74,6 +92,33 @@ def test_decimal(py_and_json: PyAndJson, input_value, expected):
         assert isinstance(output, Decimal)
 
 
+@pytest.mark.parametrize(
+    'input_value,expected',
+    [
+        ((0, (1, 4, 1, 4), -3), Decimal('1.414')),
+        ((0, (1, 2, 3), 0), Decimal('123')),
+        ((0, (1, 2, 3), 2), Decimal('12300')),
+        ((0, (1, 2, 3), -2), Decimal('1.23')),
+        ((1, (1, 4, 1, 4), -3), Decimal('-1.414')),
+        ((1, (1, 2, 3), 0), Decimal('-123')),
+        ((1, (1, 2, 3), 2), Decimal('-12300')),
+        ((1, (1, 2, 3), -2), Decimal('-1.23')),
+        ((0, (0,), 0), Decimal('0')),
+        ((0, (5,), -1), Decimal('0.5')),
+        ((1, (5,), -1), Decimal('-0.5')),
+        ((0, (1, 0, 0), -2), Decimal('1.00')),
+        ((0, (9, 9, 9), 3), Decimal('999000')),
+    ],
+    ids=repr,
+)
+def test_decimal_three_tuple_constructor(py_and_json: PyAndJson, input_value, expected):
+    """Test that Decimal can be constructed from a three-tuple (sign, digits, exponent)."""
+    v = py_and_json(cs.decimal_schema())
+    output = v.validate_test(input_value)
+    assert output == expected
+    assert isinstance(output, Decimal)
+
+
 @pytest.mark.parametrize(
     'input_value,expected',
     [
@@ -90,7 +135,7 @@ def test_decimal(py_and_json: PyAndJson, input_value, expected):
     ids=repr,
 )
 def test_decimal_strict_py(input_value, expected):
-    v = SchemaValidator({'type': 'decimal', 'strict': True})
+    v = SchemaValidator(cs.decimal_schema(strict=True))
     if isinstance(expected, Err):
         with pytest.raises(ValidationError, match=re.escape(expected.message)):
             v.validate_python(input_value)
@@ -120,7 +165,7 @@ def test_decimal_strict_py(input_value, expected):
     ids=repr,
 )
 def test_decimal_strict_json(input_value, expected):
-    v = SchemaValidator({'type': 'decimal', 'strict': True})
+    v = SchemaValidator(cs.decimal_schema(strict=True))
     if isinstance(expected, Err):
         with pytest.raises(ValidationError, match=re.escape(expected.message)):
             v.validate_json(json.dumps(input_value))
@@ -224,9 +269,7 @@ def test_decimal_multiple_of(py_and_json: PyAndJson, multiple_of: float, input_v
 
 
 def test_union_decimal_py():
-    v = SchemaValidator(
-        {'type': 'union', 'choices': [{'type': 'decimal', 'strict': True}, {'type': 'decimal', 'multiple_of': 7}]}
-    )
+    v = SchemaValidator(cs.union_schema(choices=[cs.decimal_schema(strict=True), cs.decimal_schema(multiple_of=7)]))
     assert v.validate_python('14') == 14
     assert v.validate_python(Decimal(5)) == 5
     with pytest.raises(ValidationError) as exc_info:
@@ -250,9 +293,7 @@ def test_union_decimal_py():
 
 
 def test_union_decimal_json():
-    v = SchemaValidator(
-        {'type': 'union', 'choices': [{'type': 'decimal', 'strict': True}, {'type': 'decimal', 'multiple_of': 7}]}
-    )
+    v = SchemaValidator(cs.union_schema(choices=[cs.decimal_schema(strict=True), cs.decimal_schema(multiple_of=7)]))
     assert v.validate_json(json.dumps('14')) == 14
     assert v.validate_json(json.dumps('5')) == 5
 
@@ -275,21 +316,21 @@ def test_union_decimal_simple(py_and_json: PyAndJson):
 
 
 def test_decimal_repr():
-    v = SchemaValidator({'type': 'decimal'})
+    v = SchemaValidator(cs.decimal_schema())
     assert plain_repr(v).startswith(
         'SchemaValidator(title="decimal",validator=Decimal(DecimalValidator{strict:false,allow_inf_nan:false'
     )
-    v = SchemaValidator({'type': 'decimal', 'strict': True})
+    v = SchemaValidator(cs.decimal_schema(strict=True))
     assert plain_repr(v).startswith(
         'SchemaValidator(title="decimal",validator=Decimal(DecimalValidator{strict:true,allow_inf_nan:false'
     )
-    v = SchemaValidator({'type': 'decimal', 'multiple_of': 7})
+    v = SchemaValidator(cs.decimal_schema(multiple_of=7))
     assert plain_repr(v).startswith('SchemaValidator(title="decimal",validator=Decimal(')
 
 
 @pytest.mark.parametrize('input_value,expected', [(Decimal('1.23'), Decimal('1.23')), (Decimal('1'), Decimal('1.0'))])
 def test_decimal_not_json(input_value, expected):
-    v = SchemaValidator({'type': 'decimal'})
+    v = SchemaValidator(cs.decimal_schema())
     if isinstance(expected, Err):
         with pytest.raises(ValidationError, match=re.escape(expected.message)):
             v.validate_python(input_value)
@@ -383,7 +424,7 @@ def test_non_finite_json_values(py_and_json: PyAndJson, input_value, allow_inf_n
     ],
 )
 def test_non_finite_decimal_values(strict, input_value, allow_inf_nan, expected):
-    v = SchemaValidator({'type': 'decimal', 'allow_inf_nan': allow_inf_nan, 'strict': strict})
+    v = SchemaValidator(cs.decimal_schema(allow_inf_nan=allow_inf_nan, strict=strict))
     if isinstance(expected, Err):
         with pytest.raises(ValidationError, match=re.escape(expected.message)):
             v.validate_python(input_value)
@@ -419,7 +460,7 @@ def test_non_finite_decimal_values(strict, input_value, allow_inf_nan, expected)
     ],
 )
 def test_non_finite_constrained_decimal_values(input_value, allow_inf_nan, expected):
-    v = SchemaValidator({'type': 'decimal', 'allow_inf_nan': allow_inf_nan, 'gt': 0})
+    v = SchemaValidator(cs.decimal_schema(allow_inf_nan=allow_inf_nan, gt=0))
     if isinstance(expected, Err):
         with pytest.raises(ValidationError, match=re.escape(expected.message)):
             v.validate_python(input_value)
@@ -457,12 +498,12 @@ def test_non_finite_constrained_decimal_values(input_value, allow_inf_nan, expec
     ],
 )
 def test_validate_scientific_notation_from_json(input_value, expected):
-    v = SchemaValidator({'type': 'decimal'})
+    v = SchemaValidator(cs.decimal_schema())
     assert v.validate_json(input_value) == expected
 
 
 def test_validate_max_digits_and_decimal_places() -> None:
-    v = SchemaValidator({'type': 'decimal', 'max_digits': 5, 'decimal_places': 2})
+    v = SchemaValidator(cs.decimal_schema(max_digits=5, decimal_places=2))
 
     # valid inputs
     assert v.validate_json('1.23') == Decimal('1.23')
@@ -481,7 +522,7 @@ def test_validate_max_digits_and_decimal_places() -> None:
 
 
 def test_validate_max_digits_and_decimal_places_edge_case() -> None:
-    v = SchemaValidator({'type': 'decimal', 'max_digits': 34, 'decimal_places': 18})
+    v = SchemaValidator(cs.decimal_schema(max_digits=34, decimal_places=18))
 
     # valid inputs
     assert v.validate_python(Decimal('9999999999999999.999999999999999999')) == Decimal(
@@ -490,20 +531,20 @@ def test_validate_max_digits_and_decimal_places_edge_case() -> None:
 
 
 def test_str_validation_w_strict() -> None:
-    s = SchemaValidator(core_schema.decimal_schema(strict=True))
+    s = SchemaValidator(cs.decimal_schema(strict=True))
 
     with pytest.raises(ValidationError):
         assert s.validate_python('1.23')
 
 
 def test_str_validation_w_lax() -> None:
-    s = SchemaValidator(core_schema.decimal_schema(strict=False))
+    s = SchemaValidator(cs.decimal_schema(strict=False))
 
     assert s.validate_python('1.23') == Decimal('1.23')
 
 
 def test_union_with_str_prefers_str() -> None:
-    s = SchemaValidator(core_schema.union_schema([core_schema.decimal_schema(), core_schema.str_schema()]))
+    s = SchemaValidator(cs.union_schema([cs.decimal_schema(), cs.str_schema()]))
 
     assert s.validate_python('1.23') == '1.23'
     assert s.validate_python(1.23) == Decimal('1.23')
