@@ -1,4 +1,5 @@
 import math
+import os
 import re
 import sys
 from collections.abc import Mapping
@@ -9,7 +10,8 @@ from typing import Any, Union
 import pytest
 from dirty_equals import FunctionCheck, HasRepr, IsStr
 
-from pydantic_core import CoreConfig, SchemaError, SchemaValidator, ValidationError, core_schema, validate_core_schema
+from pydantic_core import CoreConfig, SchemaError, SchemaValidator, ValidationError, core_schema
+from pydantic_core.core_schema import ExtraBehavior
 
 from ..conftest import Err, PyAndJson
 
@@ -116,11 +118,15 @@ def test_missing_error(pydantic_version):
         v.validate_python({'field_a': b'abc'})
     assert (
         str(exc_info.value)
-        == f"""\
+        == """\
 1 validation error for model-fields
 field_b
-  Field required [type=missing, input_value={{'field_a': b'abc'}}, input_type=dict]
-    For further information visit https://errors.pydantic.dev/{pydantic_version}/v/missing"""
+  Field required [type=missing, input_value={'field_a': b'abc'}, input_type=dict]"""
+        + (
+            f'\n    For further information visit https://errors.pydantic.dev/{pydantic_version}/v/missing'
+            if os.environ.get('PYDANTIC_ERRORS_INCLUDE_URL', '1') != 'false'
+            else ''
+        )
     )
 
 
@@ -213,6 +219,13 @@ def test_allow_extra_invalid():
             )
         )
 
+    with pytest.raises(SchemaError, match='extras_keys_schema can only be used if extra_behavior=allow'):
+        SchemaValidator(
+            schema=core_schema.model_fields_schema(
+                fields={}, extras_keys_schema=core_schema.int_schema(), extra_behavior='ignore'
+            )
+        )
+
 
 def test_allow_extra_wrong():
     with pytest.raises(SchemaError, match='Invalid extra_behavior: `wrong`'):
@@ -221,6 +234,12 @@ def test_allow_extra_wrong():
         )
 
 
+def test_allow_extra_fn_override_wrong():
+    v = SchemaValidator(schema=core_schema.model_fields_schema(fields={}))
+    with pytest.raises(ValueError, match='Invalid extra_behavior: `wrong`'):
+        v.validate_python({}, extra='wrong')
+
+
 def test_str_config():
     v = SchemaValidator(
         core_schema.model_fields_schema(fields={'field_a': core_schema.model_field(schema=core_schema.str_schema())}),
@@ -423,11 +442,6 @@ def test_json_error():
     ]
 
 
-def test_missing_schema_key():
-    with pytest.raises(SchemaError, match='model-fields.fields.x.schema\n  Field required'):
-        validate_core_schema({'type': 'model-fields', 'fields': {'x': {'type': 'str'}}})
-
-
 def test_fields_required_by_default():
     """By default all fields should be required"""
     v = SchemaValidator(
@@ -512,7 +526,21 @@ def test_alias_allow_pop(py_and_json: PyAndJson):
     )
     assert v.validate_test({'FieldA': '123'}) == ({'field_a': 123}, None, {'field_a'})
     assert v.validate_test({'field_a': '123'}) == ({'field_a': 123}, None, {'field_a'})
+
+    # alias always wins if both are present
     assert v.validate_test({'FieldA': '1', 'field_a': '2'}) == ({'field_a': 1}, None, {'field_a'})
+    assert v.validate_test({'field_a': '1', 'FieldA': '2'}) == ({'field_a': 2}, None, {'field_a'})
+
+    # even invalid values are ignored if alias is present
+    assert v.validate_test({'FieldA': '1', 'field_a': 'q'}) == ({'field_a': 1}, None, {'field_a'})
+    assert v.validate_test({'field_a': 'q', 'FieldA': '2'}) == ({'field_a': 2}, None, {'field_a'})
+
+    # but if the alias is invalid, those errors are raised
+    with pytest.raises(ValidationError, match=r'FieldA\n +Input should be a valid integer.+\[type=int_parsing,'):
+        assert v.validate_test({'FieldA': 'q', 'field_a': '2'}) == ({'field_a': 1}, None, {'field_a'})
+    with pytest.raises(ValidationError, match=r'FieldA\n +Input should be a valid integer.+\[type=int_parsing,'):
+        assert v.validate_test({'field_a': 'q', 'FieldA': 'q'}) == ({'field_a': 2}, None, {'field_a'})
+
     with pytest.raises(ValidationError, match=r'FieldA\n +Field required \[type=missing,'):
         assert v.validate_test({'foobar': '123'})
 
@@ -679,7 +707,8 @@ def test_aliases_debug():
     )
     print(repr(v))
     assert repr(v).startswith('SchemaValidator(title="model-fields", validator=ModelFields(')
-    assert 'PathChoices(' in repr(v)
+    # check that aliases with non-empty "rest" are present, i.e. non-trivial paths
+    assert 'rest: [\n' in repr(v)
 
 
 def get_int_key():
@@ -732,11 +761,9 @@ def test_paths_allow_by_name(py_and_json: PyAndJson, input_value):
 @pytest.mark.parametrize(
     'alias_schema,error',
     [
-        ({'validation_alias': ['foo', ['bar']]}, 'Input should be a valid string'),
         ({'validation_alias': []}, 'Lookup paths should have at least one element'),
         ({'validation_alias': [[]]}, 'Each alias path should have at least one element'),
-        ({'validation_alias': [123]}, "TypeError: 'int' object cannot be converted to 'PyList'"),
-        ({'validation_alias': [[[]]]}, 'Input should be a valid string'),
+        ({'validation_alias': [123]}, "TypeError: 'int' object is not an instance of 'list'"),
         ({'validation_alias': [[1, 'foo']]}, 'TypeError: The first item in an alias path should be a string'),
     ],
     ids=repr,
@@ -744,12 +771,10 @@ def test_paths_allow_by_name(py_and_json: PyAndJson, input_value):
 def test_alias_build_error(alias_schema, error):
     with pytest.raises(SchemaError, match=error):
         SchemaValidator(
-            schema=validate_core_schema(
-                {
-                    'type': 'model-fields',
-                    'fields': {'field_a': {'type': 'model-field', 'schema': {'type': 'int'}, **alias_schema}},
-                }
-            )
+            schema={
+                'type': 'model-fields',
+                'fields': {'field_a': {'type': 'model-field', 'schema': {'type': 'int'}, **alias_schema}},
+            }
         )
 
 
@@ -1471,20 +1496,6 @@ def test_bad_default_factory(default_factory, error_message):
 
 
 class TestOnError:
-    def test_on_error_bad_name(self):
-        with pytest.raises(SchemaError, match="Input should be 'raise', 'omit' or 'default'"):
-            validate_core_schema(
-                {
-                    'type': 'model-fields',
-                    'fields': {
-                        'x': {
-                            'type': 'model-field',
-                            'schema': {'type': 'default', 'schema': {'type': 'str'}, 'on_error': 'rais'},
-                        }
-                    },
-                }
-            )
-
     def test_on_error_bad_default(self):
         with pytest.raises(SchemaError, match="'on_error = default' requires a `default` or `default_factory`"):
             SchemaValidator(
@@ -1667,6 +1678,9 @@ def test_extra_behavior_allow(
     assert new_fields_set == {'not_f'}
 
 
+# We can't test the extra parameter of the validate_* functions above, since the
+# extras_schema parameter isn't valid unless the models are configured with extra='allow'.
+# Test the validate_* extra parameter separately instead:
 @pytest.mark.parametrize(
     'config,schema_extra_behavior_kw',
     [
@@ -1674,10 +1688,16 @@ def test_extra_behavior_allow(
         (core_schema.CoreConfig(extra_fields_behavior='forbid'), {'extra_behavior': None}),
         (core_schema.CoreConfig(), {'extra_behavior': 'forbid'}),
         (None, {'extra_behavior': 'forbid'}),
-        (core_schema.CoreConfig(extra_fields_behavior='allow'), {'extra_behavior': 'forbid'}),
+        (core_schema.CoreConfig(extra_fields_behavior='ignore'), {'extra_behavior': 'forbid'}),
+        (core_schema.CoreConfig(), {}),
+        (core_schema.CoreConfig(), {'extra_behavior': None}),
+        (None, {'extra_behavior': None}),
     ],
 )
-def test_extra_behavior_forbid(config: Union[core_schema.CoreConfig, None], schema_extra_behavior_kw: dict[str, Any]):
+def test_extra_behavior_allow_with_validate_fn_override(
+    config: Union[core_schema.CoreConfig, None],
+    schema_extra_behavior_kw: dict[str, Any],
+):
     v = SchemaValidator(
         core_schema.model_fields_schema(
             {'f': core_schema.model_field(core_schema.str_schema())}, **schema_extra_behavior_kw
@@ -1685,21 +1705,65 @@ def test_extra_behavior_forbid(config: Union[core_schema.CoreConfig, None], sche
         config=config,
     )
 
-    m, model_extra, fields_set = v.validate_python({'f': 'x'})
+    m, model_extra, fields_set = v.validate_python({'f': 'x', 'extra_field': '123'}, extra='allow')
+    assert m == {'f': 'x'}
+    assert model_extra == {'extra_field': '123'}
+    assert fields_set == {'f', 'extra_field'}
+
+    v.validate_assignment(m, 'f', 'y', extra='allow')
+    assert m == {'f': 'y'}
+
+    new_m, new_model_extra, new_fields_set = v.validate_assignment({**m, **model_extra}, 'not_f', '123', extra='allow')
+    assert new_m == {'f': 'y'}
+    assert new_model_extra == {'extra_field': '123', 'not_f': '123'}
+    assert new_fields_set == {'not_f'}
+
+
+@pytest.mark.parametrize(
+    'config,schema_extra_behavior_kw,validate_fn_extra_kw',
+    [
+        (core_schema.CoreConfig(extra_fields_behavior='forbid'), {}, None),
+        (core_schema.CoreConfig(extra_fields_behavior='forbid'), {'extra_behavior': None}, None),
+        (core_schema.CoreConfig(), {'extra_behavior': 'forbid'}, None),
+        (None, {'extra_behavior': 'forbid'}, None),
+        (core_schema.CoreConfig(extra_fields_behavior='allow'), {'extra_behavior': 'forbid'}, None),
+        (core_schema.CoreConfig(extra_fields_behavior='ignore'), {}, 'forbid'),
+        (core_schema.CoreConfig(extra_fields_behavior='ignore'), {'extra_behavior': None}, 'forbid'),
+        (core_schema.CoreConfig(), {'extra_behavior': 'ignore'}, 'forbid'),
+        (None, {'extra_behavior': 'ignore'}, 'forbid'),
+        (core_schema.CoreConfig(extra_fields_behavior='allow'), {'extra_behavior': 'ignore'}, 'forbid'),
+        (core_schema.CoreConfig(), {}, 'forbid'),
+        (core_schema.CoreConfig(), {'extra_behavior': None}, 'forbid'),
+        (None, {'extra_behavior': None}, 'forbid'),
+    ],
+)
+def test_extra_behavior_forbid(
+    config: Union[core_schema.CoreConfig, None],
+    schema_extra_behavior_kw: dict[str, Any],
+    validate_fn_extra_kw: Union[ExtraBehavior, None],
+):
+    v = SchemaValidator(
+        core_schema.model_fields_schema(
+            {'f': core_schema.model_field(core_schema.str_schema())}, **schema_extra_behavior_kw
+        ),
+        config=config,
+    )
+
+    m, model_extra, fields_set = v.validate_python({'f': 'x'}, extra=validate_fn_extra_kw)
     assert m == {'f': 'x'}
     assert fields_set == {'f'}
 
     with pytest.raises(ValidationError) as exc_info:
-        v.validate_python({'f': 'x', 'extra_field': 123})
+        v.validate_python({'f': 'x', 'extra_field': 123}, extra=validate_fn_extra_kw)
     assert exc_info.value.errors(include_url=False) == [
         {'type': 'extra_forbidden', 'loc': ('extra_field',), 'msg': 'Extra inputs are not permitted', 'input': 123}
     ]
 
-    v.validate_assignment(m, 'f', 'y')
+    v.validate_assignment(m, 'f', 'y', extra=validate_fn_extra_kw)
     assert m['f'] == 'y'
 
     with pytest.raises(ValidationError) as exc_info:
-        v.validate_assignment(m, 'not_f', 'xyz')
+        v.validate_assignment(m, 'not_f', 'xyz', extra=validate_fn_extra_kw)
     assert exc_info.value.errors(include_url=False) == [
         {
             'type': 'no_such_attribute',
@@ -1713,18 +1777,26 @@ def test_extra_behavior_forbid(config: Union[core_schema.CoreConfig, None], sche
 
 
 @pytest.mark.parametrize(
-    'config,schema_extra_behavior_kw',
+    'config,schema_extra_behavior_kw,validate_fn_extra_kw',
     [
-        (core_schema.CoreConfig(extra_fields_behavior='ignore'), {}),
-        (core_schema.CoreConfig(), {'extra_behavior': 'ignore'}),
-        (None, {'extra_behavior': 'ignore'}),
-        (core_schema.CoreConfig(extra_fields_behavior='forbid'), {'extra_behavior': 'ignore'}),
-        (core_schema.CoreConfig(), {}),
-        (core_schema.CoreConfig(), {'extra_behavior': None}),
-        (None, {'extra_behavior': None}),
+        (core_schema.CoreConfig(extra_fields_behavior='ignore'), {}, None),
+        (core_schema.CoreConfig(), {'extra_behavior': 'ignore'}, None),
+        (None, {'extra_behavior': 'ignore'}, None),
+        (core_schema.CoreConfig(extra_fields_behavior='forbid'), {'extra_behavior': 'ignore'}, None),
+        (core_schema.CoreConfig(), {}, None),
+        (core_schema.CoreConfig(), {'extra_behavior': None}, None),
+        (None, {'extra_behavior': None}, None),
+        (core_schema.CoreConfig(extra_fields_behavior='allow'), {}, 'ignore'),
+        (core_schema.CoreConfig(), {'extra_behavior': 'allow'}, 'ignore'),
+        (None, {'extra_behavior': 'allow'}, 'ignore'),
+        (core_schema.CoreConfig(extra_fields_behavior='forbid'), {'extra_behavior': 'allow'}, 'ignore'),
     ],
 )
-def test_extra_behavior_ignore(config: Union[core_schema.CoreConfig, None], schema_extra_behavior_kw: dict[str, Any]):
+def test_extra_behavior_ignore(
+    config: Union[core_schema.CoreConfig, None],
+    schema_extra_behavior_kw: dict[str, Any],
+    validate_fn_extra_kw: Union[ExtraBehavior, None],
+):
     v = SchemaValidator(
         core_schema.model_fields_schema(
             {'f': core_schema.model_field(core_schema.str_schema())}, **schema_extra_behavior_kw
@@ -1732,12 +1804,12 @@ def test_extra_behavior_ignore(config: Union[core_schema.CoreConfig, None], sche
         config=config,
     )
 
-    m, model_extra, fields_set = v.validate_python({'f': 'x', 'extra_field': 123})
+    m, model_extra, fields_set = v.validate_python({'f': 'x', 'extra_field': 123}, extra=validate_fn_extra_kw)
     assert m == {'f': 'x'}
     assert model_extra is None
     assert fields_set == {'f'}
 
-    v.validate_assignment(m, 'f', 'y')
+    v.validate_assignment(m, 'f', 'y', extra=validate_fn_extra_kw)
     assert m['f'] == 'y'
 
     # even if we ignore extra attributes during initialization / validation
@@ -1745,7 +1817,7 @@ def test_extra_behavior_ignore(config: Union[core_schema.CoreConfig, None], sche
     # instead if extra='ignore' was set (or nothing was set since that's the default)
     # we treat it as if it were extra='forbid'
     with pytest.raises(ValidationError) as exc_info:
-        v.validate_assignment(m, 'not_f', 'xyz')
+        v.validate_assignment(m, 'not_f', 'xyz', extra=validate_fn_extra_kw)
     assert exc_info.value.errors(include_url=False) == [
         {
             'type': 'no_such_attribute',
@@ -1758,6 +1830,24 @@ def test_extra_behavior_ignore(config: Union[core_schema.CoreConfig, None], sche
     assert 'not_f' not in m
 
 
+def test_extra_behavior_allow_keys_validation() -> None:
+    v = SchemaValidator(
+        core_schema.model_fields_schema(
+            {}, extra_behavior='allow', extras_keys_schema=core_schema.str_schema(max_length=3)
+        )
+    )
+
+    m, model_extra, fields_set = v.validate_python({'ext': 123})
+    assert m == {}
+    assert model_extra == {'ext': 123}
+    assert fields_set == {'ext'}
+
+    with pytest.raises(ValidationError) as exc_info:
+        v.validate_python({'extra_too_long': 123})
+
+    assert exc_info.value.errors()[0]['type'] == 'string_too_long'
+
+
 @pytest.mark.parametrize('config_by_alias', [None, True, False])
 @pytest.mark.parametrize('config_by_name', [None, True, False])
 @pytest.mark.parametrize('runtime_by_alias', [None, True, False])
