@@ -1,3 +1,4 @@
+import platform
 import re
 import sys
 from decimal import Decimal
@@ -6,7 +7,7 @@ from typing import Any, Union
 
 import pytest
 
-from pydantic_core import SchemaError, SchemaValidator, ValidationError, core_schema
+from pydantic_core import CoreConfig, SchemaError, SchemaValidator, ValidationError, core_schema
 
 from ..conftest import Err, PyAndJson, plain_repr
 
@@ -56,7 +57,7 @@ def test_str(py_and_json: PyAndJson, input_value, expected):
     ],
 )
 def test_str_not_json(input_value, expected):
-    v = SchemaValidator({'type': 'str'})
+    v = SchemaValidator(core_schema.str_schema())
     if isinstance(expected, Err):
         with pytest.raises(ValidationError, match=re.escape(expected.message)):
             v.validate_python(input_value)
@@ -112,7 +113,7 @@ def test_constrained_str(py_and_json: PyAndJson, kwargs: dict[str, Any], input_v
     ],
 )
 def test_constrained_str_py_only(kwargs: dict[str, Any], input_value, expected):
-    v = SchemaValidator({'type': 'str', **kwargs})
+    v = SchemaValidator(core_schema.str_schema(**kwargs))
     if isinstance(expected, Err):
         with pytest.raises(ValidationError, match=re.escape(expected.message)):
             v.validate_python(input_value)
@@ -123,7 +124,7 @@ def test_constrained_str_py_only(kwargs: dict[str, Any], input_value, expected):
 def test_unicode_error():
     # `.to_str()` Returns a `UnicodeEncodeError` if the input is not valid unicode (containing unpaired surrogates).
     # https://github.com/PyO3/pyo3/blob/6503128442b8f3e767c663a6a8d96376d7fb603d/src/types/string.rs#L477
-    v = SchemaValidator({'type': 'str', 'min_length': 1})
+    v = SchemaValidator(core_schema.str_schema(min_length=1))
     assert v.validate_python('🐈 Hello') == '🐈 Hello'
 
     with pytest.raises(ValidationError) as exc_info:
@@ -154,7 +155,7 @@ def test_unicode_error():
     ],
 )
 def test_str_constrained(data: str, max_length: int, error: Union[re.Pattern, None]):
-    v = SchemaValidator({'type': 'str', 'max_length': max_length})
+    v = SchemaValidator(core_schema.str_schema(max_length=max_length))
     if error is None:
         assert v.validate_python(data) == data
     else:
@@ -163,7 +164,7 @@ def test_str_constrained(data: str, max_length: int, error: Union[re.Pattern, No
 
 
 def test_str_constrained_config():
-    v = SchemaValidator({'type': 'str'}, {'str_max_length': 5})
+    v = SchemaValidator(core_schema.str_schema(), config=CoreConfig(str_max_length=5))
     assert v.validate_python('test') == 'test'
 
     with pytest.raises(ValidationError, match='String should have at most 5 characters'):
@@ -172,11 +173,15 @@ def test_str_constrained_config():
 
 @pytest.mark.parametrize('engine', [None, 'rust-regex', 'python-re'])
 def test_invalid_regex(engine):
+    if platform.python_implementation() == 'PyPy' and sys.version_info[:2] == (3, 11):
+        # pypy 3.11 type formatting
+        pytest.xfail()
+
     # TODO uncomment and fix once #150 is done
     # with pytest.raises(SchemaError) as exc_info:
     #     SchemaValidator({'type': 'str', 'pattern': 123})
     # assert exc_info.value.args[0] == (
-    #     'Error building "str" validator:\n  TypeError: \'int\' object cannot be converted to \'PyString\''
+    #     'Error building "str" validator:\n  TypeError: \'int\' object is not an instance of \'str\''
     # )
     with pytest.raises(SchemaError) as exc_info:
         SchemaValidator(core_schema.str_schema(pattern='(abc', regex_engine=engine))
@@ -209,7 +214,9 @@ def test_regex_error(engine):
 
 
 def test_default_validator():
-    v = SchemaValidator(core_schema.str_schema(strict=True, to_lower=False), {'str_strip_whitespace': False})
+    v = SchemaValidator(
+        core_schema.str_schema(strict=True, to_lower=False), config=CoreConfig(str_strip_whitespace=False)
+    )
     assert (
         plain_repr(v)
         == 'SchemaValidator(title="str",validator=Str(StrValidator{strict:true,coerce_numbers_to_str:false}),definitions=[],cache_strings=True)'
@@ -286,14 +293,14 @@ def test_subclass_preserved() -> None:
 def test_coerce_numbers_to_str_with_invalid_unicode_character(string) -> None:
     config = core_schema.CoreConfig(coerce_numbers_to_str=True)
 
-    v = SchemaValidator(core_schema.str_schema(strict=string), config)
+    v = SchemaValidator(core_schema.str_schema(strict=string), config=config)
     assert v.validate_python('\ud835') == '\ud835'
 
 
 def test_coerce_numbers_to_str_disabled_in_strict_mode() -> None:
     config = core_schema.CoreConfig(coerce_numbers_to_str=True)
 
-    v = SchemaValidator(core_schema.str_schema(strict=True), config)
+    v = SchemaValidator(core_schema.str_schema(strict=True), config=config)
     with pytest.raises(ValidationError):
         v.validate_python(42)
     with pytest.raises(ValidationError):
@@ -303,7 +310,7 @@ def test_coerce_numbers_to_str_disabled_in_strict_mode() -> None:
 def test_coerce_numbers_to_str_raises_for_bool() -> None:
     config = core_schema.CoreConfig(coerce_numbers_to_str=True)
 
-    v = SchemaValidator(core_schema.str_schema(), config)
+    v = SchemaValidator(core_schema.str_schema(), config=config)
     with pytest.raises(ValidationError):
         v.validate_python(True)
     with pytest.raises(ValidationError):
@@ -321,7 +328,7 @@ def test_coerce_numbers_to_str_raises_for_bool() -> None:
 def test_coerce_numbers_to_str(number: Number, expected_str: str) -> None:
     config = core_schema.CoreConfig(coerce_numbers_to_str=True)
 
-    v = SchemaValidator(core_schema.str_schema(), config)
+    v = SchemaValidator(core_schema.str_schema(), config=config)
     assert v.validate_python(number) == expected_str
 
 
@@ -336,11 +343,14 @@ def test_coerce_numbers_to_str(number: Number, expected_str: str) -> None:
 def test_coerce_numbers_to_str_from_json(number: str, expected_str: str) -> None:
     config = core_schema.CoreConfig(coerce_numbers_to_str=True)
 
-    v = SchemaValidator(core_schema.str_schema(), config)
+    v = SchemaValidator(core_schema.str_schema(), config=config)
     assert v.validate_json(number) == expected_str
 
 
 @pytest.mark.parametrize('mode', (None, 'schema', 'config'))
+@pytest.mark.xfail(
+    platform.python_implementation() == 'PyPy' and sys.version_info[:2] == (3, 11), reason='pypy 3.11 type formatting'
+)
 def test_backtracking_regex_rust_unsupported(mode) -> None:
     pattern = r'r(#*)".*?"\1'
 
@@ -351,7 +361,9 @@ def test_backtracking_regex_rust_unsupported(mode) -> None:
         elif mode == 'schema':
             SchemaValidator(core_schema.str_schema(pattern=pattern, regex_engine='rust-regex'))
         elif mode == 'config':
-            SchemaValidator(core_schema.str_schema(pattern=pattern), core_schema.CoreConfig(regex_engine='rust-regex'))
+            SchemaValidator(
+                schema=core_schema.str_schema(pattern=pattern), config=core_schema.CoreConfig(regex_engine='rust-regex')
+            )
 
     assert exc_info.value.args[0] == (
         'Error building "str" validator:\n'
@@ -369,7 +381,9 @@ def test_backtracking_regex_python(mode) -> None:
     if mode == 'schema':
         v = SchemaValidator(core_schema.str_schema(pattern=pattern, regex_engine='python-re'))
     elif mode == 'config':
-        v = SchemaValidator(core_schema.str_schema(pattern=pattern), core_schema.CoreConfig(regex_engine='python-re'))
+        v = SchemaValidator(
+            schema=core_schema.str_schema(pattern=pattern), config=core_schema.CoreConfig(regex_engine='python-re')
+        )
     assert v.validate_python('r""') == 'r""'
     assert v.validate_python('r#""#') == 'r#""#'
     with pytest.raises(ValidationError):
